// Modulo-Operations for unsigned int
// written by Thorsten Reinecke (12.03.1999)
// last change: 17.04.1999

// the following functions are necessary because a*b%m will return false
// results if a*b overflows unsigned-int-range.
// these routines are okay for unsigned-int-range/2

#include <iostream.h>


#ifndef ASM386

unsigned int mulmod(register unsigned int a, register unsigned int b, const unsigned int m)
{
  // assumes a<m, b<m !!
  // hint: multiplication runs faster if b<=a
  // returns (a*b) mod m
  //a%=m; b%=m; // (only necessary if a>=m or b>=m)
  register unsigned int x = 0;
  if (b&1) x=a;
  while (b>>=1)
   {
    a<<=1; if (a>=m) a-=m;
    if (b&1) { x+=a; if (x>=m) x-=m; }
   }
  return x;
}

unsigned int squaremod(register unsigned int a, const unsigned int m)
{
  // assumes a<m !!
  // returns (a^2) mod m
  //a%=m; // only necessary if a>=m
  register unsigned int b=a;
  register unsigned int x=0;
  if (b&1) x=a;
  while (b>>=1)
   {
    a<<=1; if (a>=m) a-=m;
    if (b&1) { x+=a; if (x>=m) x-=m; }
   }
  return x;
}

#else

// i386 specific assembler code
// runs much faster :)

inline unsigned int mulmod(const unsigned int a, const unsigned int b, const unsigned int m)
{
  // returns (a*b) mod m
  unsigned int x;
  asm ("movl %1,%%eax; mull %2; divl %3" : "=d" (x) : "m" (a), "m" (b), "m" (m) : "eax");
  return x;
}

inline unsigned int squaremod(const unsigned int a, const unsigned int m)
{
  // returns (a^2) mod m
  unsigned int x;
  asm ("movl %1,%%eax; mull %%eax; divl %2" : "=d" (x) : "m" (a), "m" (m) : "eax");
  return x;
}

#endif


unsigned int powmod(register unsigned int a, register unsigned int pow, const unsigned int m)
{
  // assumes a<m
  // returns (a^pow) mod m
  unsigned int x = 1;
  if (pow&1) x=a;
  while (pow>>=1)
   {
     a=squaremod(a,m);
     if (pow&1) x=mulmod(x,a,m);
   }
  return x;
}

bool strong_probab_prime(const unsigned int p, const unsigned int base)
{
  // assumes p>base>1 !!
  register unsigned int d=p-1;
  register unsigned int s=0;
  while (!(d&1U)) { d>>=1; ++s; }
  //cout << "d,s: " << d << "," << s << endl;
  if (s==0) return false;
  //cout << "p,d,s: " << p << "," << d << "," << s << endl;
  d=powmod(base,d,p);
  unsigned int x = d;
  while (s && x!=p-1) { x=squaremod(x,p); --s; } 
  //cout << "p,d,s,x: " << p << "," << d << "," << s << "," << x << endl;
  return (d==1 && x!=p-1) || (d!=1 && x==p-1);
}


bool probab_prime(const unsigned int p)
{
 // assumes p>61 !!
 // actually p is prime if 61<p<4.759.123.141 (this has been checked by someone before)
 // problem: mulmod(a,b,n) produces false results if highest
 //          significant bit in unsigned int overflows!
 return strong_probab_prime(p,2)
        && strong_probab_prime(p,7)
        && strong_probab_prime(p,61);
}


bool is_prime(register const int p)
{
  // slow for large p's, but safe for all p's
  if (p==2 || p==3) return true;
  if ((p%2==0) || (p%3==0)) return false;
  register int t = 5;
  while (t*t<=p) 
    {
      if ((p%t)==0) return false;
      t+=2;
      if ((p%t)==0) return false;
      t+=4;
    }
  return true;
}


unsigned int gcd (register unsigned int a, register unsigned int b)
{
  // returns greatest common divisor
  while (a&&b) if (a>b) a%=b; else b%=a;
  if (a) return a; else return b;
}


unsigned int oddgcd(register unsigned int a, register unsigned int b)
{
  // returns greatest (odd) common divisor
  if (a==0) return b;
  if (b==0) return a;
  do
   {
     //cout << a << "," << b << endl;
     while (1&a^1) a>>=1;
     while (1&b^1) b>>=1;
     if (a>b) a-=b;
     if (b>a) b-=a;
   } while (a!=b);
  //cout << a << endl;
  return a;
}


signed int jacobi (int a, unsigned int b)
{
  // returns jacobi(a/b), assumes b odd!

  if (a%b)
   {
     signed int x = 1;
     if (a<0)
      {
        a=-a;
        if ((b-1)&2) x=-x; // (-1/b)=(-1)^((b-1)/2)
      }
     while (a>1 && b>1)
      {
        a%=b; if (a==0) return 0;
        unsigned int twos = 0;
        while ((a&1)==0) { a>>=1; twos++; }
        if (twos&1) // only odd two's in a can change the sign!
         {
          // (2/b)=(-1)^((b^2-1)/8)
          twos=(b&15)*(b&15); // but note that overflow wouldn't hurt here...
          if (twos&8) x=-x;
         }
	// a and b odd natural numbers -> (a/b)=(-1)^( (a-1)/2 * (b-1)/2 ) * (b/a)
        if (2&a&b) x=-x;
        twos=a; a=b; b=twos; // swap loop again...
      }
     return x;
   }
  else return 0;   
}

signed int legendre (signed int a, const unsigned int p)
{
  // returns legendre(a/b), assumes b odd!
  
  // warning: legendre is only defined for odd prime b.
  //          this function works for all odd b: 
  //          it will therefore return jacobi(a/b)
  // remember: undefined results can be runerrors, but
  //           an undefined result can be anything else, too.
  //           So, undefined means, you can't rely on a specific
  //           behaviour.

  //           For debugging reasons one may wish to check for
  //           odd prime input p.
#if 0
  if ( (p%2==0) || !is_prime(p) )
   {
     cerr << "error in legendre: " << p << " is not an odd prime!" << endl;
     exit(1);
   }
#endif
  return jacobi(a,p);
} 

unsigned int invmod(unsigned int x, const unsigned int m)
{
  // returns 1/x mod m (or error if no inverse exists) 
  unsigned int a,b,q,r;
  signed int h,inv,sw;
  inv=1; sw=0;
  a=x%=m; b=m;
  while (b)
   {
     q=a/b; r=a%b;
     a=b; b=r;
     h=inv-sw*q; inv=sw; sw=h;
     //cout << q << "," << r << "," << s0 << "," << s1 << endl; 
   }
  if (a!=1)
   {
     cerr << "WARNING! invmod: 1/" << x << " (mod " << m << ") does not exist" << endl;
     exit(1);
   }
  a = (inv>=0) ? inv : m+inv;
  /*
  if (mulmod(x,a,m)!=1)
   {
     cerr << "error in invmod: buggy?" << endl;
     exit(1);
   }
  */
  return a;
}


#if 0

int main()
{
  unsigned int m=7;
  unsigned int x=5;
  unsigned int y;
  y=invmod(x,m);
  cout << y << " = 1/" << x << " (mod " << m << ")" << endl;
  return 0;
}

#endif
