/* Faktorisierungsprogramm
   written 1997, 1998, 1999, 02/2000 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization",
   Boston; Basel; Stuttgart: Birkhuser, 1985
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch eine
   Variation des multipolynomialen Quadratischen Siebs versucht.
   
   (vgl. Robert D. Silverman: The Multiple Polynomial Quadratic Sieve,
    Mathematics Of Computation, Vol.48, January 1987, p. 329-339)

   Die dort beschriebene Large Prime Variante verwende ich ebenfalls.
   Zustzlich benutze ich aber Large Primes als dynamische Faktoren zur
   dynamischen Erweiterung der (usrprnglich statischen) Faktorbasis, d.h.
   ich siebe zustzlich mit ihnen. Damit das trotz ihrer Gre effizient
   geschehen kann, benutze ich Priority Queues (z.B. Fibonacci-Heaps).
   Diese Methode ist auf eigenem Mist gewachsen; in der von mir studierten
   Literatur konnte ich keinen Hinweis auf hnliche Methoden finden, die
   die Faktorbasis dynamisch erweitern. -- Zustzlich zu diesem Verfahren
   habe ich mich durch ein als "Double Large Prime"-Variante bekanntes
   Verfahren inspirieren lassen. Dieses ist teilweise implementiert und an
   meine dynamische Erweiterung angepat.
   Eine Zyklensuche zur Konstruktion von vollstndig in der Faktorbasis
   aufgehenden Relationen aus Double Large Primes (LP1,LP2 - LP2,LP3 - ...
   ... - LPk,LP1) ist nunmehr ebenfalls implementiert.
   Weitergehende Verbesserungen sind allerdings noch nicht implementiert:
   Eine Absplittung dynamischer Faktoren aus "triple large primes"
   unter Zuhilfenahme von "double large primes"-Zyklen gerade bei der
   Faktorisierung sehr groer Zahlen drfte z.B. ebenfalls sehr ntzlich
   sein, da sie die dynamische Faktorbasis vergrert...

   Das durch die gesiebten Relationen entstehende lineare Gleichungssystem
   wird dabei simultan mit den eingefgten Relationen gelst.
   Um Speicherplatz zu sparen, die Korrektheit bertragener Relationen beim
   verteilten Sieben einfacher berprfen zu knnen sowie triviale Lsungen
   schneller zu entdecken, habe ich die Relationen normiert:
   Jede Relation hat bei mir nun die Form 
    "enthaltene Faktorbasiselemente mit ungeradem Exponenten" * x^2 = 1 (mod n),
   wobei x den bereits ermittelbaren Quadratwurzelanteil darstellt.
   Mehrfachpotenzen der enthaltenen Faktorbasiselemente werden bei
   (ganzzahliger) Halbierung des Exponenten zu x hinzumultipliziert. -- Wenn
   schlielich eine Relation der Form x^2=1 entstanden ist, so kann ein
   (hoffentlich nichttrivialer) Faktor ber gcd(x-1,n) gefunden werden. 

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 95 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
   
   In der aktuellen Version ist jetzt auch eine Implementierung elliptischer
   Kurven zur Ermittlung von Faktoren "gemigter Gre" enthalten. Dies ist
   das einzige Verfahren, mit dem sich momentan auch bei sehr groen Zahlen
   nichttriviale Faktoren entdecken lassen (da seine Laufzeit von der Gre
   des entdeckten Faktors abhngig ist).

   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   Fr ein effizienteres Sieben wre vielleicht noch das "Number-Field-Sieve"
   sinnvoll. (Siebverfahren faktorisieren grundstzlich in Abhngigkeit von
   der Gre der zu faktorisierenden Zahl; sie sind besonders effizient,
   wenn die Zahl lediglich zwei Primteiler hat und diese von der Stelligkeit
   her nicht allzuweit voneinander entfernt liegen.)

   Auerdem wre dann das komplette Programm anzupassen und zu optimieren,
   insbesondere auch die lineare Algebra, die man vielleicht auf den
   Hintergrundspeicher auslagern knnte.

   Ob das simultane Lsen des linearen Gleichungssystems sinnvoll ist, oder
   ob man vielleicht effizientere Methoden (z.B. Wiedemann-Verfahren) nach
   Abschlu der Siebphase anwendet, hngt im wesentlichen vom Speicherbedarf
   des Gleichungssystems ab. Beim verteilten Sieben hat das simultane Lsen
   des Gleichungssystems zumindest den Vorteil, parallel zur Siebphase abzu-
   laufen. 
   [
     Man knnte das Wiedemann-Verfahren ja auch trotz des simultanen
     Lsens einsetzen, um Multi-Large-Primes aufzulsen, in dem man das
     Wiedemann-Verfahren auf die Multi-LargePrime-Faktoren der Relationen
     beschrnkt und die erhaltenen Lsungen in das statische Gleichungssystem
     einfgt. - Analog knnte man das Wiedemann-Verfahren zur Ermittlung
     und Absplittung von dynamische Faktoren aus den Multi-Large-Primes
     verwenden (hnlich der Zyklensuche bei den Double-Large-Primes).
     Aber das ist noch alles "Zukunftsmusik".
   ]
   
   Da das Programm inzwischen schon hufiger erweitert und verndert worden
   ist, mte ich es eigentlich nochmal "sauberschreiben", modularisieren,
   etc. -- Mal sehen, wann ich Zeit und Lust dazu habe... ;)
*/

#ifdef AUTODETECTION
 // autodetection for gcc-compiler
 #warning "autodetection enabled..."
 #ifdef __GNUG__
  #if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 8)
    #ifdef USE_OLD_GCC
     #warning "deaktiviere USE_OLD_GCC"
     #undef USE_OLD_GCC
    #endif
  #else
    #ifndef USE_OLD_GCC
     #warning "aktiviere USE_OLD_GCC"
     #define USE_OLD_GCC
    #endif
    #ifndef USE_FIBHEAP
     /* gcc 2.7.x bentigt Fibonacci-Heap, da priority-queues dort anders
        definiert sind... */
     #warning "aktiviere USE_FIBHEAP"
     #define USE_FIBHEAP
    #endif
  #endif
  #if #cpu (i386)
   #ifndef ASM386
    #warning "aktiviere i386-inline assembler code"
    #define ASM386
   #endif
  #else
   #ifdef ASM386
    #warning "deaktiviere i386-inline assembler code"
    #undef ASM386
   #endif
  #endif
  #if ! defined (unix) && defined(USE_NETWORK)
    #error "Netzwerkuntersttzung nur unter Unix-Systemen" 
  #endif
 #endif
#endif /* autodetection */

#ifdef USE_OLD_GCC
/* gcc 2.7.x bentigt Fibonacci-Heap, da priority-queues dort anders
   definiert sind... */
#ifndef USE_FIBHEAP
#warning "alter gcc bentigt eigentlich USE_FIBHEAP"
//#define USE_FIBHEAP
#endif
#endif 

#ifndef USE_NETWORK
#ifdef IS_SERVER
#error "SERVER und kein USE_NETWORK"
#endif
#else
#ifndef IS_STANDALONE
const int server_port = 12364;
const int server_port_tries = 5;
#ifdef IS_SERVER
int server_socket;
#endif
#include <unix_buffer.H>
#endif
#endif

#ifndef IS_SERVER
#ifndef IS_CLIENT
#undef IS_STANDALONE
#define IS_STANDALONE
#endif
#endif

#ifdef IS_SERVER
#ifdef IS_CLIENT
#error "Illegal Configuration SERVER-CLIENT"
#else
#ifdef IS_STANDALONE
#error "Illegal Configuration SERVER-STANDALONE"
#endif
#endif
#endif

#ifdef IS_CLIENT
#ifdef IS_STANDALONE
#error "Illegal Configuration CLIENT-STANDALONE"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <strstream>
#include <math.h>
#include <gmp.h>
#include "myBitString.H"
#include <list>

#include <set>
struct myintcompare {
  inline bool operator() (const int i1, const int i2) const
   {
     return i1 < i2;
   }
};
typedef set<int,myintcompare> intset;


#include <vector>

#ifdef USE_NETWORK
#include <sys/utsname.h>
#include <signal.h>
#ifdef IS_CLIENT
#include <sys/wait.h>
#include <unistd.h>
#include <pthread.h>
bool Sende_Relationen = true;
int  Sende_Nr = 0;
string file_name;
pid_t childpid = 0;
#endif
#endif

// --------------------------------------------------------------------
// Tfactor ist nunmehr eine Klasse:
enum Factortype { empty, static_prime, single_large_prime, double_large_prime };
class Tfactor
{
 private:
  unsigned int p1,p2; // mit p1<=p2
 public:
  inline Tfactor operator= (const unsigned int x)
   { p1=0; p2=x; return *this; }
  //inline Tfactor operator= (const Tfactor &x)
  // { p1=x.p1; p2=x.p2; return *this }
  bool DLP_get(const mpz_t &n);
  inline int int_value() const
   {
     // liefert int-Wert zurck (oder Fehlermeldung)
     if (p1)
      {
       cerr << "Kann int nicht zwei Werte zuweisen!" << endl;
       exit(1);
       return 0;
      }
     return p2;
   }
  inline int LP1() const { return p1; }
  inline int LP2() const { return p2; }
  inline void set_for_search(const unsigned int x)
   { 
     // invers sortiert einfgen, um nach x suchen zu knnen
     p1=x; p2=0;
   }
  inline void swap() { unsigned int h=p1; p1=p2; p2=h; }
  inline void in_mpz_ablegen(mpz_t &x) const
   {
     if (p1==0) { mpz_set_ui(x,p2); return; }
     if (p2==0) { mpz_set_ui(x,p1); return; }
     mpz_set_ui(x,p1); mpz_mul_ui(x,x,p2);
   }
  inline Factortype Typ() const
   {
     if (p1>p2)
      { cerr << "Tfactor: unsortierte Faktoren!" << endl; }
     if (p1==0 && p2==0) return empty;
     if (p1==0 && p2==1) return static_prime;
     if (p1==0) return single_large_prime;
     return double_large_prime;
   }
  inline bool DLP_enthaelt_Teiler(const unsigned int x) const
   { return p1==x || p2==x; }
  inline bool operator< (const Tfactor &x) const
   {
     if (p1==x.p1) return p2<x.p2;
     else return p1<x.p1;
   }
  inline bool operator== (const Tfactor &x) const
   {
     return (p1==x.p1 && p2==x.p2) || (p1==x.p2 && p2==x.p1);
   }

  friend ostream& operator<< (ostream &ostr, const Tfactor &x);
  friend istream& operator>> (istream &istr, Tfactor &x);
  int operator/ (const unsigned int u) const;
};

int Tfactor::operator/ (const unsigned int u) const
{
  // Diese "Divisionsoperation" liefert den durch u abgespaltenen
  // (also den anderen) Faktor zurck
  if (p1==u) return p2;
  if (p2==u) return p1;
  cerr << "Tfactor: Ungltige Division!" << endl;
  exit(1);
  return 0;
};

ostream& operator<< (ostream &ostr, const Tfactor &x)
 {
   if (x.p1) ostr << x.p1 << "*" << x.p2;
   else ostr << x.p2;
   return ostr;
 };

istream& operator>> (istream &istr, Tfactor &x)
 {
   char s[50];
   istr >> s;
   x.p1=0; x.p2=0;
   int i=0;
   while (s[i]!=0 && s[i]!='*') i++;
   if (s[i]==0) 
    {
      x.p1=0; x.p2=atoi(s);
      //cout << "DLP: ein Wert eingelesen" << endl;
    }
   else 
    if (s[i]=='*')
     {
       s[i]=0;
       x.p1=atoi(s);
       x.p2=atoi(&s[i+1]);
       //cout << "DLP: zwei Werte eingelesen" << endl;
     }
    else
     {
       cerr << "Fehler beim Einlesen DLP!" << endl;
     }
   //cout << "DLP:" << x.p1 << "*" << x.p2 << endl;
   return istr;
 };
// --------------------------------------------------------------------
 

struct TSpecialFactorRelation {
  Tfactor factor;
  streampos fpos;

  inline bool operator() (const TSpecialFactorRelation &t1, const TSpecialFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
};

const char SpecialRelationsFile [] = "special_relations.dat";
const char DynamicRelationsFile [] = "dynamic_relations.dat";
const char StaticRelationsFile []  = "static_relations.dat";
const char RecoveryFile []    = "recovery.dat";

// Streams fr die Speicherung der Relationen und sonstiger Daten
// zur Verfgung stellen
// *************************************************************************
// **** wichtig:
// **** Wenn ostream und istream durch denselben filebuf fest miteinander
// **** gekoppelt sind, darf/kann man sich *nicht* darauf verlassen, da
// **** tellp() und tellg() voneinander unabhngige Funktionen sind!
// **** *Jede* Schreiboperation kann dann *auch* den Wert von tellg()
// **** verndern und *jede* Leseoperation den von tellp()!
// ************************************************************************* 

// Fr die Gleichungszeilen zu den Specialfactors
// streams fr das Datenfile zur Verfgung stellen:
filebuf SpecialRelations_buffer;
ostream SpecialRelations_to_file (&SpecialRelations_buffer);
istream SpecialRelations_from_file (&SpecialRelations_buffer);

// Fr die Gleichungszeilen der dynamischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf DynamicRelations_buffer;
ostream DynamicRelations_to_file (&DynamicRelations_buffer);
istream DynamicRelations_from_file (&DynamicRelations_buffer);

// Fr die Gleichungszeilen der statischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf StaticRelations_buffer;
ostream StaticRelations_to_file (&StaticRelations_buffer);
istream StaticRelations_from_file (&StaticRelations_buffer);

// Fr Recovery
filebuf Recovery_buffer;
ostream Recovery_to_file (&Recovery_buffer);
istream Recovery_from_file (&Recovery_buffer);

#ifdef IS_CLIENT
char   * communication_name;
ofstream communication_stream;
#endif


// Fr das Speichern der gemachten Faktorisierungen
ofstream Factorization_to_file ("factorizations.txt",ios::in|ios::out|ios::app);
/* Bemerkung:
   Auf diesem Stream werden alle mit diesem Programm gemachten Faktorisierungen
   protokolliert. Die zu faktorisierenden Zahlen sowie alle gefundenen Faktoren
   werden angehngt.
   Das zugehrige File sollte im Bedarfsfall gelscht werden, wenn die Daten
   nicht mehr bentigt werden.
*/

struct TDynamicFactorRelation; // "Vorwrtsdeklaration" fr Headerfile
typedef unsigned short int FBsizetype; // Typ der Faktorbasisgre
  // Wenn die Faktorbasis maximal 65535 Elemente haben soll,
  // sollte man "unsigned short int" whlen, um Speicherplatz zu sparen.
  // Wenn Speicherplatz keine Rolle spielt, ist "int" eine gute Wahl (denn
  // dann gibt es u.U. weniger implizite Typkonvertierungen).
  // Hinweis:
  // "FBsizetype" hat *nur* mit der Anzahl der Primzahlen in der statischen
  // Faktorbasis (also den Indices auf die darin enthaltenen Primzahlen),
  // aber *nichts* mit dem Typ (und der Gre) der Primzahlen aus der
  // statischen Faktorbasis selbst zu tun! 

#include "qsieve.H"

const int Specialfactor_Sieb_Threshold  = 1000000000; /* bis hierhin wird auch mit
							 Dynamic-Factoren gesiebt */
mpz_t DLP_Threshold; // Double-Large-Prime-Threshold (wird in main() gesetzt)
const int Dynamicfactor_Cache_Threshold =   10000000; /* bis hierhin werden die Dynamic-Factoren
							 im Speicher gehalten. Falls der Hauptspeicher
							 knapp wird, diesen Wert reduzieren... */

class TDynamicFactorRelation {
  /* Bemerkung:
     Falls factor nicht allzugro ist (im Augenblick ist dies als Obergrenze der
     int-Bereich), dann kann auch mit den Dynamic-Factors gesiebt werden.
     Nachfolgend die dazu erforderlichen Daten...
  */
public:
  int factor; // factor in int-Darstellung
  int SQRT_kN;
  streampos fpos;
  inline bool operator() (const TDynamicFactorRelation &t1, const TDynamicFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
#ifdef LAZY_EVALUATION
  inline const CGleichungszeile* GLcached(void) const
   {
     return NULL; // bei LAZY-Evaluation wird hier nichts gecached!
   }
  inline bool insert_cache(CGleichungszeile* pGL) const
   {
     return false; // bei LAZY-Evakluation wird hier nichts gecached!
   }
#else
private:
  CGleichungszeile *GL; // Zeiger auf GL (oder Null, falls nicht im Speicher)
public:
  inline const CGleichungszeile* GLcached(void) const
   {
     return GL; // GL zurckliefern
   }
  inline bool insert_cache(CGleichungszeile* pGL)
   {
     // Wenn Faktor klein genug ist und die Relation nicht zuviel Speicher
     // schluckt, dann wird die Relation komplett gecached, ansonsten liegt
     // sie nur auf dem File. In jedem Falle mu vor dem Aufruf dieser Methode
     // das Feld factor korrekt belegt sein. Das Feld GL wird beim Aufruf
     // als undefiniert angenommen. Wenn die bergebene Relation gecached wird,
     // dann wird GL auf pGL gesetzt und true zurckgegeben, ansonsten erhlt
     // GL den Wert NULL und es wird false zurckgeliefert.
     // Es bleibt dem Aufrufer berlassen, ob er die zu pGL gehrige Relation
     // zerstrt (falls der Wert nicht gecached wird).

     // momentane Nebenbedingung: es mssen weniger als 50 Faktoren aus der
     // statischen Faktorbasis in der Relation enthalten sein, damit sie
     // gecached wird. (ansonsten wird ein zu hoher Speicherbedarf angenommen)
     if (factor<=Dynamicfactor_Cache_Threshold && pGL->SizeOfRelation()<50)
      {
        // Relation cachen
#if 0 && ! defined(USE_OLD_GCC)
        // keine Ahnung, warum alter g++ hier Fehler rausspuckt?????
        cout << "Fr " << factor << " dynamische Relation gecached." << endl;
#endif
        //char ch; cin >> ch;
        pGL->optisize(); // Speicherbedarf optimieren
        GL=pGL; return true;
      }
     else
      {
        // Relation nicht cachen
        GL=NULL; return false;
      }
   }
#endif
};


struct TSieb_Delta {
  int delta;
  int factor;
  inline bool operator< (const TSieb_Delta &v) const
    {
#ifdef USE_FIBHEAP
      return (delta < v.delta);
#else
      return (delta > v.delta);
#endif
      // delta > v.delta , if queue is used
      // delta < v.delta , if fibonacci-heap is used
    }
};

#ifdef USE_FIBHEAP
#include "fibheap.H"
FibHeap <TSieb_Delta> Sieb_Delta_Heap, Sieb_Delta_QuadrateHeap;
#else
#include <queue>
priority_queue<TSieb_Delta> Sieb_Delta_Heap, Sieb_Delta_QuadrateHeap;
#endif


mpz_t n, /* zu faktorisierende Zahl */
      kN, /* das mit einem gnstigen Vorfaktor multiplizierte n
	     (nachdem n um einfach zu findende Faktoren reduziert wurde) */
      r, /* Quadratwurzel von n (oder bei optimierter Fassung k*n) */
      x,y; /* globale Hilfsvariablen,
	      jede Unterfunktion kann sie nutzen, mu aber davon ausgehen,
	      da sie jederzeit von anderen Funktionen verndert werden! */

int Vorfaktor_von_kN; // der noch zu ermittelnde Vorfaktor fr kN (kN=Vorfaktor_von_kN*n)

const FBsizetype MAX_Primbasis_size = 25000; /* Maximale Gre der Primzahlbasis */ 
/* Hinweis: Je grer die zu faktorisierende Zahl ist, desto grer
   sollte prinzipiell auch die Primzahlbasis sein, da dann die Relationen
   einfacher gefunden werden knnen.
   Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
   den Speicher!
   Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
   Primzahlbasis verlangsamend aus, da fr ein greres Gleichungssystem
   natrlich auch mehr Relationen gesiebt werden mssen.
   Je nach Geschwindigkeit des Rechners, des verfgbaren Haupt- und
   Hintergrundspeichers (und auch dessen Zugriffszeiten) knnen die
   ttschlich zu verwendenden Werte stark variieren. Sie sind deshalb
   experimentell zu ermitteln.
   Die Maximalgre wird durch MAX_Primbasis_size festgelegt und steht
   damit bereits zur Kompilierzeit fest.
   Bei einer Speicherkapazitt von 32 MB (realem) Hauptspeicher
   hat sich 20000 als brauchbarer Wert erwiesen.
*/

FBsizetype Primbasis_size = 5000; // tatschlich zu verwendender Wert
/*  Sinnvolle Werte fr Faktorbasen liegen zwischen 1000 und 50000
    (je nach verfgbarem Speicher und der zu faktorisierenden Zahl).
    Der Wert wird normalerweise spter durch die Prozedur "tune_parameters"
    angepat.
*/

int Primzahlen[MAX_Primbasis_size]; /* fr eine effiziente Probedivision
				       alle Primzahlen der Faktorbasis hier
				       akkumulieren */


int SQRT_kN_der_Primzahlen[MAX_Primbasis_size];
int Delta_der_Primzahlen[MAX_Primbasis_size][2]; /* fr den effizienten Zugriff
						    werden die pro Siebintervall berechneten
						    Werte hier abgelegt */
 
const int FB_max_weitere_Potenzen = 1000; // maximale Anzahl weiterer Primzahlpotenzen, mit denen gesiebt wird
int anz_weitere_Primzahlpotenzen = 0; // tatschliche Anzahl (wird im Programm ermittelt)
int FB_maxQuadrate = 0; /* tatschliche Anzahl der Quadrate,
			   mit denen normal gesiebt wird (wird im Programm ermittelt) */
int Primzahlpotenzen[FB_max_weitere_Potenzen];
int SQRT_kN_der_Primzahlpotenzen[FB_max_weitere_Potenzen];
int Delta_der_Primzahlpotenzen[FB_max_weitere_Potenzen][2];


CGleichungszeile* GLS[MAX_Primbasis_size] = {NULL};
/* Gleichungssystem besteht aus Zeigern auf Gleichungszeilen,
   das mit Null-Pointern initialisiert ist, da anfangs noch
   keine Relationen (=Gleichungszeilen) vorliegen */

CPolynom Polynom; // Objekt zur Verwaltung der Polynomberechnung fr multipolynomiales Sieb



/* Zur Erluterung der Begrifflichkeit von Dynamic-Factors/Special-Factors:
   Ich verwende momentan eine modifizierte Multi-Large-Prime-Variante:
   Dynamic-Factors entsprechen der Large-Prime-Variante. Aber ich siebe mit jedem neu entdeckten
   dynamischen Faktor zustzlich...
   Special-Factors sind zusammengesetzte Restfaktoren (wie sie in der Double-Large-Prime-Variante
   auftreten). Es wird einerseits versucht, Special-Factors durch neu aufgenommene
   Dynamic-Factors zu splitten, um daraus weitere Dynamic-Factors zu gewinnen.
   Andererseits wird eine Zyklensuche auf diesen durchgefhrt, um weitere
   Relationen zu gewinnen.
   In einer spteren Implementierung knnte man die Zyklensuche noch
   auf Hyperkantenzyklen hherer Large-Primes erweitern.
*/
typedef set<TSpecialFactorRelation,TSpecialFactorRelation> TSpecial_Factor_Relations;
TSpecial_Factor_Relations Special_Factor_Relations;
typedef set<TDynamicFactorRelation,TDynamicFactorRelation> TDynamic_Factor_Relations;
TDynamic_Factor_Relations Dynamic_Factor_Relations;


double Factor_Threshold = 3.0; /* Exponent, der grere Faktoren ber die
				  Faktorbasis hinaus erlaubt.
				  Sinnvolle Werte: 1.0 bis 3.0
				  1.0: keine zustzlichen Faktoren auerhalb der Faktorbasis 
				  bis 2.0: Primfaktoren auerhalb der Faktorbasis bis
				  zum Quadrat des grten Faktors der FB
				  ber 2.0: Faktoren bis zu grter Faktor der Faktorbasis
				  hoch Factor_Threshold.
			       */


const int sieben_ab_Primzahlindex = 5; /* ab diesem Primzahlindex wird gesiebt.
					  Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
					  der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
					  beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
					  so einfach) gesiebt werden.
					  Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
					  grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
					  viele Relationen durchs Sieb fallen... */


const int Siebgroesse = 100000;
/* In (physikalischen) Teilintervallen von "Siebgroesse" werden die Relationen
   im (logischen) Siebintervall [-M,M] (siehe auch Deklaration von M) gesiebt.
   Es handelt sich deshalb um eine bereits zur Compilierungszeit feststehenden Konstante. 
   Hinweis: Je grer "Siebgroesse", desto effizienter geht das Aussieben
   der Relationen. Gleichzeitig erhht sich aber auch der Speicherbedarf.
   Der Einflu der Siebgre ist allerdings geringer als der der gewhlten
   Primzahlbasis. Ein Wert um 1 Million drfte daher vollkommen ausreichen.
   Andererseits kann mglicherweise im Cachespeicher des Rechners gesiebt
   werden, wenn die Siebgre klein gewhlt wird. Kleine Siebgren erhhen
   zwar den rechnerischen Overhead, aber dafr ist der Zugriff auf das Sieb
   mglicherweise deutlich schneller. Deshalb sollte man auch eine Siebgre
   um 100000 (bei 512K Cache) ausprobieren.
*/

int M = 1000000; // das zu jedem Polynom gehrige Siebintervall ist [-M,M]
/* M ist ein im Unterschied zu "Siebgroesse" ein logischer Wert, der weniger stark
   von der verwendeten Hardware abhngig ist. Die logische Intervallgre
   ist vor allem abhngig von der Gre der zu faktorisierenden Zahl zu whlen.
   Der hier festgelegte Wert ist also lediglich ein Defaultwert, der von der
   Konfiguration vor Beginn eines konkreten Faktorisierungslaufs (zur Laufzeit)
   modifiziert werden kann (und soll).
   Bemerkung: In multiple-polynomial-quadratic-sieve-Varianten sollte man
   mglichst kleine Siebintervalle whlen, da die relative Trefferquote von
   Relationen mit der Intervallgre abnimmt. (In 10 kleinen Intervallen ist
   die Chance, eine Relation zu finden, hher als in einem 10mal so groen
   Intervall).
   Gegen kleine Intervalle spricht jedoch, da Kosten fr das Wechseln
   der Polynome, Neuinitialisierungen und Wurzelberechnungen anfallen.
*/


int Sieboffset = -M; /* Im allgemeinen wird das fr ein Polynom zu siebende Siebintervall 
			grer sein als die (physikalisch optimale) Siebgre.
			Das Siebintervall ist daher in Teilstcken zu sieben. Hierfr wird der
			Sieboffset bentigt, der den Startpunkt des jeweiligen Teilsiebs angibt. */

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// !!! WICHTIG !!! #define und typedef mssen einander entsprechen !!!
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#define TSIEBELEMENTSIZE 1 /* 1 = signed char, 2 = short signed int, 0 = kein assembler inline code */
// obiges define sorgt bei Verwendung der ASM386-Direktive fr inline-assembler-code in der Siebroutine
typedef signed char TSiebElement; // mu auf jeden Fall vorzeichenbehaftet sein!
/* Je kleiner der Typ von TSiebElement ist, desto 
    a) kleiner ist der fr das Sieb bentigte Speicher,
    b) schneller kann deshalb gesiebt werden.
   Allerdings ist die Granularitt zum Erkennen von Relationen bei
   kleineren Datentypen grber und deshalb schlechter.
   In Frage kommen:
    - signed char (1 Byte)
    - short signed int (2 Byte)
    - signed int (4 Byte)
    - float, double (Fliekommatypen, mehrere (>3) Bytes)
   Die Bytegren knnen zwischen Systemen variieren. Ich habe aber nicht
   geprft, ob das Programm berhaupt auf mit einer Systemeinstellung von
   int < 4 Byte, short int < 2 Byte, int < 4 lauffhig ist.

   Wichtiger Hinweis: Wenn man Fliekommatypen als TSiebElement verwenden will (was ich nicht mehr empfehle),
   dann sind im Programm die "ceil"-Aufrufe fr TSiebElement kontraproduktiv und sollten entfernt werden.
   Aber es gibt keinen vernnftigen Grund mehr, mit Fliekommatypen zu sieben.
*/

#define log_Multiplikator 1 /* Siebwert wird nach damit nach TSiebelement konvertiert */
/*
  Da die Vorkommastellen der natrliche Logarithmen von Primzahlen
  zu hohe Ungenauigkeiten beim Sieben verursachen, werden sie noch
  mit log_Multiplikator multipliziert, um eine hhere Genauigkeit
  zu erreichen.
  Der Multiplikator kann auf 1 gesetzt sein, wenn TSiebElement Fliekommatyp
  ist. Ansonsten sollte man ihn mglichst gro whlen, und zwar so, da
  der Threshold der blicherweise zu faktorisierenden Zahlen noch in den
  Datentyp TSiebElement pat.
  Damit durch Abrundungen keine Relationen durch die Lappen gehen, werden
  aber beim Sieben grundstzlich aufgerundete Werte abgezogen, so da ein
  kleiner Multiplikator hier allenfalls ein zu engmaschiges Sieb erzeugt.
  Die exakten Auswirkungen verschiedener Einstellungen wren allerdings
  noch genauer zu untersuchen.
  Wenn man nicht genau wei, was man tut, sollte man eher TSiebElement
  "hhertypen". Der Multiplikator sollte dann deutlich ber 1 liegen (10 - 1000).
*/

TSiebElement* Sieb; // Sieb (noch zu allokierendes Feld von TSiebelementen)

TSiebElement log_Primzahlen[MAX_Primbasis_size]; /* Hier stehen die konvertierten Logarithmen
						    der Primzahlen aus der Faktorbasis */
TSiebElement log_Primzahl_der_Primzahlpotenzen[FB_max_weitere_Potenzen];
/* fr weitere Primzahlpotenzen stehen hier die konvertierten Logarithmen der Primzahlen */

bool collecting_phase_beendet = false; /* switch, der die "aktive" und "passive" Phase des Siebens
					  bestimmt:
					  Wenn gengend Special-Relations und Dynamic-Relations gesammelt
					  sind, um ausreichend viele Static-Relations zu erhalten, wird
					  das Aufsammeln weiterer Special- und Dynamic-Relations durch
					  das Setzen von collecting_phase_beendet=true unterbunden. */

int Fuellung_GLS = 0; // Fllung des Gleichungssystems mit Relationen
int Erfolgsaufteilung[5] = {0,0,0,0,0}; /* fr statistische Auswertung, inwieweit die dynamische
					   Factorbase fr das Finden der Relationen
					   verantwortlich ist:
					   [0]->gar nicht beteiligt,
					   [1]-> mit einem Faktor beteiligt
					   [2]-> mit zwei Faktoren beteiligt
					   [3]-> mit drei Faktoren beteiligt
					   [4]-> mit mehr als drei Faktoren beteiligt */

int Special_hit = 0;
int Special_to_dynamic_Factor_hit = 0;
int gefundene_DLP_Zyklen = 0;

int biggest_Prime_in_Factorbase;


#include "modulo.cc" // Modulo-Operationen fr unsigned int
                     // als .cc included fr optimierende Compiler


const int mpzbase_f = 36; // mpz-Ein/Ausgabeformat fr Fileoperationen
// Fr lesbare Ein- und Ausgaben wird weiterhin Dezimalsystem verwendet  

ostream& operator << (ostream& ostr, const mpz_t& value)
{ // Ausgabefunktion fr mpz_t - Zahlen (um die hassenswerten expliziten Konvertierungen zu minimieren...)
  char* str = new char [mpz_sizeinbase(value,10)+2]; mpz_get_str(str, 10, value);
  ostr << str << " (" << strlen(str) << ")";
  delete [] str;              
  return ostr;
}

istream& operator >> (istream& istr, mpz_t& value)
{ // Eingabefunktion fr mpz_t - Zahlen
  string s;
  istr >> s;
  mpz_set_str(value, s.c_str(), 10);
  istr >> s; // Sizeinbase berlesen
  return istr;
}

// einige Parameter fr easy-Faktorisierung (defaults)
int rho_Phase   =   150000; // Grundinitialisierung fr Rho-Phase
int phi_Phase1  =   500000; // fr phi-Methode Phase 1
int phi_Phase2  = 10000000; // fr phi-Methode Phase 2 (=isolierter Faktor)
int elcu_Phase1 =    20000; // fr elliptische Kurven Phase 1
int elcu_Phase2 =  2000000; // fr elliptische Kurven Phase 2
int elcu_Kurven =      100; // Anzahl der zu probierenden elliptischen Kurven 

const int probab_prime_checks = 25; // #rounds for mpz_probab_prime

#define DLP_SQUFOF // Schalter fr DLP-Faktorisierungsmethode
#ifndef DLP_SQUFOF
// benutze Pollard-rho zum Faktorisieren der Double-Large-Primes
bool Tfactor::DLP_get(const mpz_t &n)
{
  if (mpz_cmp(n,DLP_Threshold)>0)
   {
     //cout << "DLP_get: Threshold berschritten..." << endl;
     return false;
   }

  int runden=50000; // Anzahl der Pollard-rho-Runden, die maximal versucht werden
  // Hinweis: viele Runden kosten Zeit, wenige Runden senken die Erkennungsrate

  mpz_t x,a,a2;
  mpz_init_set(x,n); mpz_init(a); mpz_init(a2);
  mpz_set_ui(a,1); mpz_set(a2,a);

  p1=0; p2=0; // Vorbesetzung;

  do
    {
     
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_sub(x,a2,a);
      mpz_gcd(x,x,n);
    } while (--runden && mpz_cmp_ui(x,1)==0);
  if (mpz_cmp_ui(x,1)!=0)
    {
      // nun prfen, ob beide Teiler prim sind und klein genug sind
      if (mpz_cmp_ui(x,Specialfactor_Sieb_Threshold)>=0) goto fertig;
      p1=mpz_get_ui(x);
      mpz_divexact(a,n,x);
      if (mpz_cmp_ui(a,Specialfactor_Sieb_Threshold)>=0) goto fertig;
      p2=mpz_get_ui(a);
      // Hinweis: wenn man hier nicht auf Primzahlen prft, dann mu man
      //          es spter bei der Abspaltung eines Faktors tun!
      //          Also machen wir es hier schon...
      if (!probab_prime(p1)) { p1=0; goto fertig; }
      if (!probab_prime(p2)) { p2=0; goto fertig; }
      if (p1>p2) { unsigned int h=p1; p1=p2; p2=h; }
    };
fertig:
  mpz_clear(a); mpz_clear(a2); mpz_clear(x);
  //if (runden==0) cout << "DLP_get: Runden ausgeschpft." << endl;
  cout << "DLP_get: (" << 50000-runden << ") " << p1 << "," << p2 << endl;
  
  //static unsigned int runden_ges = 0;
  //runden_ges+=50000-runden;
  //cout << "DLP_get: bisherige Runden: " << runden_ges << endl;

  return (p1!=0 && p2!=0);
}

#else
// benutze Shank's square forms factorization zum Faktorisieren der Double-Large-Primes
bool Tfactor::DLP_get(const mpz_t &n)
{
  if (mpz_cmp(n,DLP_Threshold)>0)
   {
     //cout << "DLP_get: Threshold berschritten..." << endl;
     return false;
   }

  const int runden=150000; // Anzahl der Runden, die maximal versucht werden
  // Hinweis: viele Runden kosten Zeit, wenige Runden senken die Erkennungsrate

  mpz_t mx,mr;
  mpz_init_set(mx,n); mpz_init(mr);

  p1=0; p2=0; // Vorbesetzung;

  mpz_sqrtrem(mx,mr,n); 
  int sq = mpz_get_ui(mx);
  int d = mpz_get_ui(mr);
  mpz_mul_ui(mx,mx,2); mpz_add_ui(mx,mx,1); mpz_sqrt(mx,mx);
  int sq2sqN = mpz_get_ui(mx);

  int lindex=0;
  const int bound = 60;
  int list[bound];
  bool square_found=false;

  int Q,P2,Q2;
  int r=0, Q0=1, P1=sq, Q1=d;
  int runde1=0, runde2=0;

  if (d==0) { p1=sq; p2=sq; goto fertig; }

  for (runde1=1; runde1<runden; ++runde1)
   {
     //cout << "Runde" << runde1 << endl;
     Q=(sq+P1)/Q1; P2=Q*Q1-P1; Q2=Q0+Q*(P1-P2);
     //cout << " Q:" << Q << " Q1:" << Q1 << " Q2:" << Q2
     //     << " P1:" << P1 << " P2:" << P2 << endl;
     //char ch; cin >> ch;
     int u=Q1; if ((u&1)==0) u>>=1;
     Q0=Q1; Q1=Q2; P1=P2;
     //cout << runde << ": " << u << endl;
     if (u<sq2sqN && u>1)
      {
        //cout << runde1 << ": " << u << endl;
        list[lindex++]=u; 
        if (lindex>=bound)
         {
           cout << "SQUFOF: Liste luft ber..." << endl;
           break;
         }
      }
     if (runde1&1)
      {
        //cout << "Q1:" << Q1 << endl;
        if ((Q1&3)>1) continue; // kein Quadrat
        if ((Q1&7)==5 || (Q1&9)==8) continue; // ebenfalls kein Quadrat
        r=static_cast<int>(sqrt(Q1)); if (Q1!=r*r) continue; // kein Quadrat
        bool fl=false;
        for (int k=0; k<lindex; ++k)
         if (r==list[k]) { fl=true; break; }
        if (fl)
         { 
           // cout << "Square not useful..." << endl;
           continue;
         }
        if (r>1) { square_found=true; break; }
        //cout << "Kein Erfolg gehabt..." << endl;
        goto fertig;
      }
   }

  if (!square_found)
   {
     //cout << "DLP-SQUFOF: Kein Erfolg gehabt..." << endl;
     goto fertig;
   }

  //cout << "Square gefunden in Runde " << runde1 << endl;

  double DQ,DQ0,DQ1,DQ2,DP1,DP2;
  mpz_set_ui(mr,P1); mpz_mul_ui(mr,mr,P1); mpz_sub(mx,n,mr);
  mpz_div_ui(mx,mx,r);
  DQ1=mpz_get_d(mx);
  DQ0=r; DP1=P1;

  for (runde2=1; runde2<runden; ++runde2)
   {
     DQ=floor((sq+DP1)/DQ1); DP2=DQ*DQ1-DP1;
     //cout << DP1 << "," << DP2 << endl;
     DQ2=DQ0+DQ*(DP1-DP2); DQ0=DQ1; DQ1=DQ2;
     if (DP1==DP2)
      {
        if ((DQ0/2)==floor(DQ0/2)) DQ0/=2;
        //cout << "Faktor gefunden..." << endl;
        // nun prfen, ob beide Teiler prim sind und klein genug sind
        if (DQ0>=Specialfactor_Sieb_Threshold) goto fertig;
        p1=static_cast<int>(DQ0);
        if (mpz_div_ui(mx,n,p1)) { cerr << "DLP_get-SQUFOF: Teiler mit Rest??" << endl; exit(1); }
        if (mpz_cmp_ui(mx,Specialfactor_Sieb_Threshold)>=0) goto fertig;
        p2=mpz_get_ui(mx);
        // Hinweis: wenn man hier nicht auf Primzahlen prft, dann mu man
        //          es spter bei der Abspaltung eines Faktors tun!
        //          Also machen wir es hier schon...
        if (!probab_prime(p1)) { p1=0; goto fertig; }
        if (!probab_prime(p2)) { p2=0; goto fertig; }
        if (p1>p2) { unsigned int h=p1; p1=p2; p2=h; }
        goto fertig;
      }
     DP1=DP2;
   }

fertig:
  mpz_clear(mx); mpz_clear(mr);
  //cout << "DLP_get: (" << runde1 << "," << runde2 << ") "
  //     << p1 << "," << p2 << endl;
  return (p1!=0 && p2!=0);
}
#endif


void SpecialFactors_Zyklensuche()
{
  // Wenn man einen Zyklus von Large-Primes durch Double-Large-Primes
  // erzeugen kann, so neutralisieren sich die beteiligten Large-Primes und
  // man erhlt eine vollstndig in der statischen Faktorbasis aufgehende
  // Relation.

  if (Special_Factor_Relations.empty()) return; // es gibt nichts zu finden
  cout << "Starte Zyklensuche auf Special-Factor-Set" << endl;

  TSpecial_Factor_Relations arbeitsmenge;
  cout << "Size: " << Special_Factor_Relations.size() << ", " << arbeitsmenge.size() << endl;

  arbeitsmenge.swap(Special_Factor_Relations);
  // Hinweis: um nicht doppelten Speicherbedarf zu haben, werden die
  // Special-Factor-Relations in die Arbeitsmenge geschoben (und nicht kopiert)
  // es ist dann darauf zu achten, da sie von der Arbeitsmenge auch wieder
  // zurckgeschoben (und nicht nur gelscht) werden...
  cout << "Size: " << Special_Factor_Relations.size() << ", " << arbeitsmenge.size() << endl;

  intset knotenmenge;
  typedef list<TSpecialFactorRelation> Tkantenliste;
  Tkantenliste kantenliste;

  TSpecial_Factor_Relations::iterator pos;
  TSpecialFactorRelation SearchRelation;
  int rechts,links;

  while (!arbeitsmenge.empty())
   {
#ifdef USE_OLD_GCC
     // alte g++-lib kennt kein clear()
     knotenmenge.erase(knotenmenge.begin(),knotenmenge.end());
#else
     knotenmenge.clear(); // wichtig! (wegen Schleifenwiederholungen)
#endif
     pos=arbeitsmenge.begin();
     TSpecialFactorRelation relation = *pos;
     kantenliste.push_back(relation);
     links=relation.factor.LP1();
     rechts=relation.factor.LP2();
     knotenmenge.insert(links);
     knotenmenge.insert(rechts);
     arbeitsmenge.erase(pos); Special_Factor_Relations.insert(relation);
     relation.factor.swap();
     arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
     Special_Factor_Relations.insert(relation);
     relation.factor.swap();

     //cout << "Starte neue Runde der Zyklensuche." << endl;
     while (!kantenliste.empty())
      {
        //for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
        // cout << (*p).factor << "|";
        //cout << endl;
        //cout << "Suche nach " << links << endl;
	//cout << "Arbeitsmenge enthlt: " << endl;
        //for (pos=arbeitsmenge.begin(); pos!=arbeitsmenge.end(); ++pos)
        // cout << (*pos).factor << ",";
	//cout << endl;

        SearchRelation.factor.set_for_search(links);
        pos=arbeitsmenge.lower_bound(SearchRelation);
        if (pos!=arbeitsmenge.end() && (*pos).factor.DLP_enthaelt_Teiler(links) )
         {
           // links kann eine Kante eingefgt werden
           relation = *pos;
           arbeitsmenge.erase(pos); // alten Specialfactor entfernen
           Special_Factor_Relations.insert(relation);
           relation.factor.swap();
           arbeitsmenge.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
           Special_Factor_Relations.insert(relation);
           relation.factor.swap();
           links=relation.factor/links;
           if (knotenmenge.find(links)!=knotenmenge.end())
            { 
              // Zyklus gefunden!
	      ++gefundene_DLP_Zyklen;
	      cout << "DLP-Zyklus gefunden... " << endl;
              // nun noch: 
              // Zyklus auswerten und erhaltene Relation ins Gleichungssystem
              // einfgen.
              CGleichungszeile *GL = new CGleichungszeile;
              GL->kombinieren(SpecialRelations_from_file,relation.fpos);
              mpz_mul_ui(GL->Delta,GL->Delta,links);
	      mpz_mod(GL->Delta,GL->Delta,n);
              rechts=relation.factor/links;
	      //cout << relation.factor;

              // Kantenliste bis zu der Position, wo der Wert "links"
              // erstmalig enthalten ist, durchgehen -> dort endet der Zyklus.
	      // dabei noch die Wurzel des Produkts der Specialfaktoren
              // hinzumultiplizieren (beim Zyklus sind dies alle beteiligten
              // Large Primes in einfacher Potenz), also die Werte "rechts"
	      unsigned int Zyklenlaenge=1;
              for (Tkantenliste::iterator p=kantenliste.begin(); p!=kantenliste.end(); ++p)
	       {
                 Zyklenlaenge++;
		 //cout << " - " << p->factor;
                 mpz_mul_ui(GL->Delta,GL->Delta,rechts);
	         mpz_mod(GL->Delta,GL->Delta,n);
                 GL->kombinieren(SpecialRelations_from_file,(*p).fpos);
                 if ((*p).factor.DLP_enthaelt_Teiler(links)) break;
		 rechts=(*p).factor/rechts;
               }
              //cout << endl;
              cout << "DLP-Zyklenlnge betrgt " << Zyklenlaenge << endl;
	      GL->Kongruenzcheck(); // aus Sicherheitsgrnden...
     	      Relation_einfuegen(GL);
	    
  	      // Anschlieend eine verantwortliche (wir whlen die letzte Kante)
              // aus der Special-Relation-Menge entfernen:
              Special_Factor_Relations.erase(relation); // alten Specialfactor entfernen
              relation.factor.swap();
              Special_Factor_Relations.erase(relation); // alten Specialfactor entfernen (geswapte Kopie)
              relation.factor.swap();
              {
                streampos pos;
	        // nicht vergessen, die Relation auf File zu markieren!
                pos = SpecialRelations_to_file.tellp(); // Schreibpostion merken
                SpecialRelations_to_file.seekp(relation.fpos); SpecialRelations_to_file << "U "; // ungltig markieren
                SpecialRelations_to_file.seekp(pos); // wieder an die alte position
              }
 
              // Schlielich: Zyklensuche fortsetzen.
	      links=relation.factor/links;
              continue;
            }
           else
            {
              // kein Zyklus gefunden -> Kante und Knoten einfgen
              knotenmenge.insert(links);
              kantenliste.push_front(relation);
	      //cout << "Fge Kante " << relation.factor << " hinzu " << endl;
              continue;
            } 
         }
        // keine Kante zum linksseitigen Einfgen gefunden
        // -> Alternative 1: rechts einfgen (zum Zyklusfinden nicht notwendig)
        // -> Alternative 2: linksseitig eine Kante entfernen
	//cout << "Entferne Kante " << kantenliste.begin()->factor << endl;
#ifdef USE_OLD_GCC
        {
         TSpecialFactorRelation r = *kantenliste.begin();
         links=r.factor/links;
        }
#else
        links=kantenliste.begin()->factor/links;
#endif
        kantenliste.pop_front(); // also linksseitig eine Kante entfernen
      }
   }
  cout << "Size: " << Special_Factor_Relations.size() << ", " << arbeitsmenge.size() << endl;
  cout << "Zyklensuche beendet." << endl;
}

bool SpecialFactorRelation_vermerken(const Tfactor &DLP, CGleichungszeile *GL)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren,
  // so da man GL anschlieend ins Gleichungssystem eintragen kann.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat!" << endl;
     Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  TSpecial_Factor_Relations::iterator pos = Special_Factor_Relations.find(relation);
  if (pos==Special_Factor_Relations.end())
   {
     relation.fpos=GL->speichern(SpecialRelations_to_file,relation.factor);
     Special_Factor_Relations.insert(relation);
     relation.factor.swap();
     Special_Factor_Relations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
	SpecialFactors_Zyklensuche();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit durch bekannten Special Factor!" << endl;
     Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.in_mpz_ablegen(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->kombinieren(SpecialRelations_from_file,(*pos).fpos);
     //Relation_einfuegen(GL); // sollte an anderer Stelle gemacht werden!
     return false;
   }
}

bool SpecialFactorRelation_testen(const Tfactor &DLP)
{
  // Testen, ob Special-Factor erfolgreich gesplittet werden kann
  // true -> ja, false -> nein
  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     // cout << "Spezialfall: DLP ist Quadrat!" << endl;
     return true;
   }
  TSpecialFactorRelation relation;
  relation.factor=DLP;
  TSpecial_Factor_Relations::iterator pos = Special_Factor_Relations.find(relation);
  return (pos!=Special_Factor_Relations.end());
}


#ifndef IS_CLIENT
void tune_parameters(const unsigned int Dezimalstellen)
{
  ifstream in("qsieve.cfg");
  char eingabezeile[200];
  string zeile;
  unsigned int Stellen, gewaehlte_Stellenzahl=0;
  if (in)
    {
      cout << "Lese Konfigurationsdatei" << endl; 
      while (!in.eof())
	{
	  in.getline(eingabezeile,200,'\n');
	  if (eingabezeile[0]=='#') continue; // Kommentarzeilen berlesen
	  zeile=eingabezeile;
	  Stellen = atoi(zeile.substr(0, zeile.find(":")).c_str());
	  if (Stellen > Dezimalstellen || Stellen==0) break;
          gewaehlte_Stellenzahl=Stellen;
#ifdef USE_OLD_GCC
	  zeile.remove(0, zeile.find(":")+1);
	  rho_Phase=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  phi_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  phi_Phase2=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  elcu_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  elcu_Phase2=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  elcu_Kurven=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.remove(0, zeile.find(",")+1);
	  Primbasis_size=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.remove(0, zeile.find(":")+1);
	  sscanf(zeile.substr(0, zeile.find(":")).c_str(), "%lf", &Factor_Threshold);
	  zeile.remove(0, zeile.find(":")+1);
#else	  
	  zeile.erase(0, zeile.find(":")+1);
	  rho_Phase=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase2=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase2=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Kurven=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  Primbasis_size=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.erase(0, zeile.find(":")+1);
	  sscanf(zeile.substr(0, zeile.find(":")).c_str(), "%lf", &Factor_Threshold);
	  zeile.erase(0, zeile.find(":")+1);
#endif

	  M = atoi(zeile.c_str());
	  M = (M>Siebgroesse) ? M : Siebgroesse;
	}
    }
  else
    {
       static char ch = 'N';
       cout << "Konfigurationsdatei nicht gefunden!" << endl
	    << "verwende Defaultwerte..." << endl;
       cout << "qsieve.cfg not found! Using default values!" << endl;

       if (ch!='Y')
         {
           cout << "Bitte mit [Y]es besttigen!" << endl;
           cout << "Please confirm with [Y]es." << endl;
           cin >> ch;
           if (ch!='Y') exit(1);
         }
    }
  
  cout << "Die verwendeten Parameter fr die Faktorisierung lauten:" << endl;
  cout << "Angegegebene Stellenzahl: " << Dezimalstellen << endl;
  cout << "Ausgewhlte Parameter fr " << gewaehlte_Stellenzahl << " Stellen" << endl;
  cout << "Pollard-Rho: " << rho_Phase << endl;
  cout << "Pollard-Phi: Phase 1: " << phi_Phase1 << " Phase 2: " << phi_Phase2 << endl;
  cout << "Elliptische Kurven: Phase 1: " << elcu_Phase1
       << " Phase 2: " << elcu_Phase2
       << " #Kurven: " << elcu_Kurven << endl; 
  cout << "Gre der statischen Faktorbasis: " << Primbasis_size << endl;
  cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
  cout << "Siebintervall pro Polynom: [" << -M << "," << M << "]" << endl;

  if (Primbasis_size>MAX_Primbasis_size)
    {
      cerr << "MAX_Primbasis_size ist zu klein gewhlt (bzw. Primbasis_size zu hoch)!" << endl;
      exit(1);
    }
}
#endif /* von #ifndef IS_CLIENT */


const char* MAL(void)
{
  // liefert Leerstring bei erstem Aufruf, " * " sonst.
  // Aufgabe: Verhinderung aufwendiger Tests, um den Faktor 1 als
  // ersten bzw. letzten Faktors unterdrcken zu knnen.
  static bool erster_Aufruf = true;
  if (erster_Aufruf)
   {
     erster_Aufruf=false;
     return "";
   }
  else return " * ";
}

void pollard(const int runden) /* Pollard-rho-Methode */
{
  cout << "Trying Pollard-rho up to " << runden << endl;
  mpz_t a,a2;
  mpz_init(a); mpz_init(a2);
  mpz_set_ui(a,1); mpz_set(a2,a);
  
  for (int i=1; i<=runden; i++)
    {
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_sub(x,a2,a);
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  do
	    {
	      if (mpz_probab_prime_p(x,probab_prime_checks))
		{
		  cout << x << " ist Faktor. [in Runde " << i <<"]" << endl;
		  Factorization_to_file << MAL() << x << flush;
		}
	      else
		{
		  cout << x << " ist zusammengesetzter Faktor. [in Runde " << i <<"]" << endl;
		  Factorization_to_file << MAL() << x << " [composite]" <<  flush;
		}
	      mpz_div(n,n,x);
	      mpz_gcd(x,x,n);
	    } while (mpz_cmp_ui(x,1)!=0);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) break; /* Rest wahrscheinlich prim */
	}
    };
  mpz_clear(a); mpz_clear(a2);
}


#include "polynomial.cc" /* for fft-continuation */

void my_mpz_powm_ui(mpz_t &a, unsigned int n, const mpz_t m)
{
  // berechnet a=a^n mod m, n>1, n (mglichst) prim,
  // dies ist eine alternative Vorgehensweise zur binren Version, die
  // fr Primzahlen ein wenig schneller sein sollte...
  // Idee dahinter:
  // Wenn in n viele Bits gesetzt sind, mssen auch viele Zweierpotenzen
  // in der Exponentation kummuliert werden. Dann wird es in vielen
  // Fllen billiger sein, den Exponenten n anderweitig zusammenzusetzen.
  // Eine Mglichkeit besteht darin, n-1 in Faktoren zu zerlegen, die
  // ihrerseits wenig "Bitoverhead" haben.

  if (n<127) { mpz_powm_ui(a,a,n,m); return; } 

  register unsigned int h = n;
  register int b = 0;
  while (h) { b+=(h&1) ? 3 : -4; h>>=1; } // #gesetzte Bits bestimmen
  //cout << "b= " << b << endl;

  if (b<0)
   {
     //cout << "---> " << n << endl; 
     mpz_powm_ui(a,a,n,m);
     return;
   }

  //cout << "-> " << n << endl; 
  mpz_t r;
  mpz_init_set(r,a);
  n--;

  h=0; while ((n&1)==0) { n>>=1; h++; }
  if (h) mpz_powm_ui(r,r,1<<h,m);
  while (n%129==0) { n/=129; mpz_powm_ui(r,r,129,m); }
  while (n%65==0) { n/=65; mpz_powm_ui(r,r,65,m); }
  while (n%33==0) { n/=33; mpz_powm_ui(r,r,33,m); }
  while (n%17==0) { n/=17; mpz_powm_ui(r,r,17,m); }
  while (n%9==0) { n/=9; mpz_powm_ui(r,r,9,m); }
  while (n%5==0) { n/=5; mpz_powm_ui(r,r,5,m); }
  while (n%3==0) { n/=3; mpz_powm_ui(r,r,3,m); }
//  while (n%7==0) { n/=7; mpz_powm_ui(r,r,7,m); }
  while (n%11==0) { n/=11; mpz_powm_ui(r,r,11,m); }
  while (n%13==0) { n/=13; mpz_powm_ui(r,r,13,m); }
  while (n%19==0) { n/=19; mpz_powm_ui(r,r,19,m); }
  //while (n%23==0) { n/=19; mpz_powm_ui(r,r,23,m); }
  //while (n%29==0) { n/=29; mpz_powm_ui(r,r,29,m); }
  if (n>1) mpz_powm_ui(r,r,n,m);
  mpz_mul(a,a,r); mpz_mod(a,a,m);
  mpz_clear(r);
}

void pollard_phi(const int phase1, const int phase2)
{
  // Pollard-Phi Methode modifiziert in drei Phasen:
  // Phase 0: alle kleine Zahlen bis konstanter Obergrenze
  // Phase 1: (im wesentlichen) Primzahlen bis phase1 als Obergrenze
  //          (Primfaktor-1) kann/darf in alle diese Primzahlen zerfallen
  // Phase 2: (im wesentlichen) Primzahlen bis phase2 als Obergrenze
  //          bis auf einen(!) aus Phase 2 mssen alle anderen
  //          Faktoren von (Primfaktor-1) bereits vorher entdeckt worden
  //          sein, um Primfaktor entdecken zu knnen.
  //          Es stehen diverse Methoden fr Phase 2 zur Verfgung,
  //          die jedoch momentan nur ber eine nderung in diesem Sourcecode
  //          aktiviert werden knnen.
  //           - phi-phi continuation
  //           - standard continuation
  //           - improved standard continuation
  //           - improved standard continuation with pairing
  //           - improved standard continuation with pairing & fft

  mpz_t a,b;
  mpz_init(a); mpz_init(b);

  int i;
  short int d;
  
  // Phase 0: kleine Zahlen (mit Primzahlpotenzen)
  cout << "Phi-Phase 0" << endl;
  mpz_set_ui(a,2);
  for (i=5000; i<10000; i++)
    {
      mpz_powm_ui(a,a,i,n);
      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)  /* Teiler gefunden? */
	{
	  cout << "Faktor gefunden in Phase 0: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
	    }
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi0]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi0] [composite]" << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;
	}
    }
  
#if 0 /* normale phi-Phase 1 */
  {
   // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
   const int D=2*3*5*7*11*13;
   i=i-(i%D); // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   cout << "Phi-Phase 1" << endl;
   while (i<=phase1)
    {
      // Primzahlen aussieben
      bool sieve[D];
      for (int j=1; j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
      for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
       for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
       /* 
         Bemerkung:
          a) p*p<D+i ist als Primzahlbedingung fr das Intervall
             etwas zu streng, aber strt wegen Phase 0 nicht.
             (-> Primzahlen unter sqrt(D) werden nicht erkannt) 
          b) p-(i%p) ist zwar p, wenn p teilt i;
             aber: den Index 0 brauchen wir spter nicht!
       */
            
      for (int j=1, dj=4; j<D; j+=dj, dj=6-dj)
       {
         //if (sieve[j]!=is_prime(i+j))
         //  { cerr << "keine bereinstimmung: " << i+j << endl; }
         if (sieve[j]) my_mpz_powm_ui(a,i+j,n);
       }
      /* Bemerkung:
         Nicht durch die Literatur verwirren lassen:
         Das akkumulierte Produkt der (a-1)-Werte mu natrlich
         nicht gebildet werden. Das wre eine total berflssige
         Aktion, da ein Teiler auch bei nachfolgenden a's erhalten
         bleiben mu (aufgrund des kleinen Fermatschen Satzes!).
         Falls sich andererseits mehrere Teiler akkumulieren, so kann
         auch das akkumulierte Produkt dies nicht verhindern...
         Keine Ahnung, warum ich das akkumulierte Produkt frher hier
         gebildet habe, offenbar habe ich mich verwirren lassen...
         (Ist ja auch wie mit den Elefanten: "Warum klatscht du denn
          dauernd in die Hnde?" - "Um die Elefanten zu verscheuchen!" -
          "Welche Elefanten?" - "Siehtst du! Alle verscheucht!")
      */
      // Schleife vermeidet hufige gcd-Berechnung
      // in seltenen Fllen knnen aber mehrere Faktoren zusammenfallen
      // -> Phi-Methode versagt dann (ggf. Schleifendurchlufe vermindern)

      i+=D;
      cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;

      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
        {	  
	  cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
            }
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1] [composite]" << flush;
 	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig; 
	}
    }
   i--; d=4; while (!is_prime(i)) { i-=d; d=6-d; } // i auf vorige Primzahl adjustieren
   cout << "Phi-Phase 1 bis " << i << endl;
   if (i>=phase2) goto fertig;
  }
#else /* modifizierte phi-Phase 1 */
  {
   // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
   const unsigned int K=2;
   unsigned int k=0;
   mpz_t base[K];
   for (unsigned int j=0; j<K; ++j) mpz_init_set_ui(base[j],1);

   const int D=2*3*5*7*11*13;
   i=i-(i%D); // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   cout << "Phi-Phase 1 (modified)" << endl;
   while (i<=phase1)
    {
      // Primzahlen aussieben
      bool sieve[D];
      for (int j=1; j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
      for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
       for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
       /* 
         Bemerkung:
          a) p*p<D+i ist als Primzahlbedingung fr das Intervall
             etwas zu streng, aber strt wegen Phase 0 nicht.
             (-> Primzahlen unter sqrt(D) werden nicht erkannt) 
          b) p-(i%p) ist zwar p, wenn p teilt i;
             aber: den Index 0 brauchen wir spter nicht!
       */
      
      mpz_set_ui(b,1);
      for (int j=1, dj=4; j<D; j+=dj, dj=6-dj)
       {
         //if (sieve[j]!=is_prime(i+j))
         //  { cerr << "keine bereinstimmung: " << i+j << endl; }
         if (sieve[j]) 
          {
            my_mpz_powm_ui(a,i+j,n);
            mpz_sub(x,a,base[k]); mpz_mul(b,b,x); mpz_mod(b,b,n);
            mpz_set(base[k],a); k++; if (k>=K) k=0;
            //cout << "i+j=" << i+j << "b=" << b << endl;
          }
       }
      // Schleife vermeidet hufige gcd-Berechnung
      // in seltenen Fllen knnen aber mehrere Faktoren zusammenfallen
      // -> Phi-Methode versagt dann (ggf. Schleifendurchlufe vermindern)

      i+=D;
      cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;

      mpz_gcd(x,b,n);
      if (mpz_cmp_ui(x,1)!=0)
        {	  
	  cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
            }
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi1] [composite]" << flush;
 	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig; 
	}
    }
   i--; d=4; while (!is_prime(i)) { i-=d; d=6-d; } // i auf vorige Primzahl adjustieren
   cout << "Phi-Phase 1 bis " << i << endl;
   for (unsigned int j=0; j<K; ++j) mpz_clear(base[j]);
   if (i>=phase2) goto fertig;
  }
#endif

#if 0
  // eigene phi-phi-continuation (testweise)
  /* Idee dahinter:
      Es wird die Folge b(k+1)=b(k)^2 mod n betrachtet.
      Sie ist zyklisch mit einer Periode, die phi(phi(P)) [mit P teilt n]
      teilt.
      Die Folge wurde bezglich des Startwertes durch die Phase 1 bereits
      komprimiert, d.h. kleinere Faktoren von phi(P) wurden bereits entfernt.
      Mit etwas Glck knnen wir daher durch das Ablaufen der Folge
      den Zyklus ermitteln und haben damit einen Teiler gefunden.
      Noch effizienter wre diese Phase, wenn man analog zu Phase 1 eine
      schnelle b^2^k-Exponentiation zur Vefrgung htte; dann knnte man
      analog zu Phase 1 vorgehen...
   */

   cout << "Phi-Phase 2: phi-phi-continuation" << endl;
   mpz_t c; mpz_init(c);
   mpz_set(b,a); // Basis fr phi-Methode
    for (int j=0; j<phase2; )
    {
      mpz_set_ui(x,1); // gcd-Sammler
      for (int k=10000; k; --k)
       {
         mpz_powm_ui(b,b,2,n);
         mpz_sub(y,b,a); mpz_mul(x,x,y); mpz_mod(x,x,n);
       }
      j+=10000;
      cout << "Phase2: " << j << "/" << phase2 << "\r" << flush;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << j << endl;
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phiphi]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phiphi] [composite]" << flush;
	    }
	  mpz_div(n,n,x);
          mpz_mod(a,a,n); mpz_mod(b,b,n);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) break;
	}
    }
#endif

#if 0 /* standard continuation */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2" << endl;
   mpz_set(b,a); // neue Basis fr phi-Methode

   // Precomputing most common used exponents:
   const int pre_exp_size = 100;
   mpz_t bpow[pre_exp_size];
   for (int j=2; j<pre_exp_size; j+=2)
    {
      mpz_init(bpow[j]);
      mpz_powm_ui(bpow[j],b,j,n);
    }  

   while (!probab_prime(i)) { i+=d; d=6-d; }
   mpz_powm_ui(a,a,i,n); mpz_sub_ui(x,a,1);
  
   while (i<=phase2) 
    { 
      int runden=5000;
      while (runden>0)
	{
	  int i_prev=i;
          do { i+=d; d=6-d; } while (i%5==0 || i%7==0 || i%11==0 || i%13==0 ||
                                     i%17==0 || i%19==0 || i%23==0 || i%29==0 ||
	           		     i%31==0 || i%37==0 || i%41==0 || i%43==0 ||
	           		     i%47==0 || i%53==0 || i%59==0 || i%61==0);
				     // || powmod(2,i-1,i)!=1);
	 // obige do-Schleife soll mglichst schnell viele zusammengesetzte
         // Zahlen ausschlieen, (Fermat-Test hier zu teuer!) 
          if (i-i_prev<pre_exp_size) mpz_mul(a,a,bpow[i-i_prev]);
          else
           { // not a precomputed power of b:
	     cerr << "not a precomputed exponent " << i-i_prev << endl;
             mpz_powm_ui(y,b,i-i_prev,n);
	     mpz_mul(a,a,y);
           }
          mpz_mod(a,a,n);
	  mpz_sub_ui(y,a,1); mpz_mul(x,x,y); mpz_mod(x,x,n);
	  --runden;
	}
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2; 
	}
    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  for (int j=2; j<pre_exp_size; j+=2)
    mpz_clear(bpow[j]); // freeing memory for precomputed exponents
  }
#endif

#if 0 /* improved standard continuation */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved)" << endl;
   mpz_set(b,a); // neue Basis fr phi-Methode

   // Precomputing most common used exponents:
   const int pre_exp_size = 2*3*5*7*11*13;
   mpz_t bpow[pre_exp_size];
   for (int j=1; j<pre_exp_size; j+=2)
    {
      if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13))
       {
         mpz_init(bpow[j]);
         mpz_powm_ui(bpow[j],b,j-1,n);
       }
    }

   mpz_powm_ui(b,b,pre_exp_size,n); // Intervall-Multiplikator  b := b^pre_exp_size mod n;

   i=((i/pre_exp_size)*pre_exp_size)-1; d=2;
   mpz_powm_ui(a,a,i,n); mpz_sub_ui(x,a,1);

   int next_i = i+1+pre_exp_size;
   mpz_mul(a,a,b); mpz_mod(a,a,n); // ein Intervall weiterspringen

   bool sieve[pre_exp_size];
   for (int i1=pre_exp_size-1; i1>=0; i1-=2) sieve[i1]=false;
   for (int p=5, dp=2; p<pre_exp_size && p*p<next_i; p+=dp, dp=6-dp )
    for (int i1=p-((next_i-pre_exp_size)%p); i1<pre_exp_size; i1+=p) sieve[i1]=true;
   
   while (i<=phase2) 
    { 
      int runden=10000;
      while (runden>0)
	{
          // nchste Primzahl ermitteln...
          do { i+=d; d=6-d; } while (i<next_i && sieve[i%pre_exp_size]);
	  // obige do-Schleife soll mglichst schnell viele zusammengesetzte
	  // Zahlen ausschlieen (mittels Siebverfahren)
          // (ein Fermat-Test wre hier zu teuer!)

	  while (i>next_i) // "if" tut's bei gengend groen pre_exp_size auch...
	    {
              // ein Intervall weiterspringen
              next_i+=pre_exp_size;
              mpz_mul(a,a,b); mpz_mod(a,a,n);
              for (int i1=pre_exp_size-1; i1>=0; i1-=2) sieve[i1]=false;
              for (int p=5, dp=2; p<pre_exp_size && p*p<next_i; p+=dp, dp=6-dp )
               for (int i1=p-((next_i-pre_exp_size)%p); i1<pre_exp_size; i1+=p) sieve[i1]=true;
	    }
          mpz_sub(y,a,bpow[next_i-i]);
          mpz_mul(x,x,y); mpz_mod(x,x,n);
	  --runden;
	}
      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;

	  // nun noch die Exponenten anpassen (damit's schneller geht)
	  for (int j=1; j<pre_exp_size; j+=2)
           if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13)) mpz_mod(bpow[j],bpow[j],n);
	}
    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  for (int j=1; j<pre_exp_size; j+=2)
   if ((j%3)&&(j%5)&&(j%7)&&(j%11)&&(j%13))
    mpz_clear(bpow[j]); // freeing memory for precomputed exponents
  }
#endif

#if 0 /* improved standard continuation with pairing */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved with pairing)" << endl;

   // in a steht die Start-Basis fr Phase 2

  /* Grundidee:
     Ausnutzung der folgenden Gleichheit:
     (a^(i+d) -1) * (a^(i-d) -1) = a^i * ( (a^i +a^(-i)) - (a^d +a^(-d)) )
     
     1. Der Multiplikatior a^i kann weggelassen werden, da er das
        Teilerverhalten beim ggT nicht beeinflut.
     2. In Intervallen der Gre D wird nun i erhht.
         -> Vorberechnung von (a^d +a^(-d)) fr 0<d<D,
         -> auerdem: pro Intervall eine Bestimmung von (a^i + a^(-i))
     3. a^(i+D) lt sich sukzessiv ber a^(i+D)=a^i * a^D bestimmen;
        a^(-(i+D)) lt sich wegen a^(-i)=(a^(-1))^i analog bestimmen.
        -> pro Intervall zwei Multiplikationen, um a^(i+D) + a^(-(i+D))
           zu bestimmen.

    Bemerkung I:
      Die hierauf aufbauende Paarbildungsmethode konstruiert nicht
      die durch andere Methoden erreichbaren Primzahlpaare, sie ist
      aber dafr vergleichsweise einfach zu implementieren und sie
      kommt mit weniger Speicherplatz als die improved-standard-continuation
      aus. Sobald mehrere Primzahlpaare innerhalb eines Intervalls
      auftauchen, ist sie schneller als die improved standard continuation.
   
    Bemerkung II: (zu 3.)
      a^i+a^(-i) lt sich -- wegen x^2-x(a^1+a^(-1))+1=(x-a)*(x-a^(-1)) --
      auch ber Lucasfolgen effizient berechnen:
      a^i+a^(-i)=Vi(a^1+a^(-1))=:V[i];
      a^(i+D)+a^(-(i+D)))=V[i+D]=V[i]*V[D]-V[i-D].
      Man kann also die Anzahl der erforderlichen Multiplikationen pro
      Intervallsprung von 2 auf 1 reduzieren.
      Damit wren wir dann (mit anderer Herleitung und mit leichten
      Modifikationen) wieder auf bereits bekannte Fortsetzungen gestoen
      (vgl. Peter L. Montgomery, "Speeding the Pollard and Elliptic Curve
       Methods of Factorization", MathComp Vol 48 (177), Jan 1987, p.243-264;
       insbesondere Abschnitte 4, 5 und 9).
  */

   mpz_t b_inv, d, d_inv, b_plus_b_inv, d_plus_d_inv, prev_b_plus_b_inv;
   mpz_init(b_inv); mpz_init(d); mpz_init(d_inv); mpz_init(b_plus_b_inv);
   mpz_init(d_plus_d_inv); mpz_init(prev_b_plus_b_inv);

   // im folgenden werden b,b_inv, d,d_inv nicht mehr aktualisiert,
   // da die zugehrigen Summen direkt durch Lucasfolgen berechnet werden.

   // Precomputing most common used values:
   const int D = 2*3*5*7*11*13; // 2*3 unbedingt notwendig!
   mpz_t valtab[D/2];

   mpz_set(b,a); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_powm_ui(d,b,2,n); mpz_powm_ui(d_inv,b_inv,2,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
   for (int j=1; j<D/2; j+=2)
    {
      if (gcd(j,D)==1) mpz_init_set(valtab[j],b_plus_b_inv);
#if 0
       {
         mpz_init(valtab[j]);
         //mpz_powm_ui(x,a,j,n);
         //mpz_invert(y,x,n);
         //mpz_add(x,x,y);
         mpz_add(x,b,b_inv);
         mpz_mod(valtab[j],x,n); // valtab[j] = a^j+a^(-j) mod n

         cout << valtab[j] << " = " << b_plus_b_inv << endl;
         cout << endl;
         char ch;
         cin >> ch;
       }

      mpz_mul(b,b,d); mpz_mod(b,b,n);
      mpz_mul(b_inv,b_inv,d_inv); mpz_mod(b_inv,b_inv,n);
#endif
      mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
      mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
      mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
      mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
    }

   i=i-(i%D); // Startpunkt ermitteln
   mpz_powm_ui(d,a,D,n); mpz_invert(d_inv,d,n); mpz_mod(d_inv,d_inv,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_powm_ui(b,a,i,n); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_mul(x,b,d_inv); mpz_mod(x,x,n);
   mpz_mul(y,b_inv,d); mpz_mod(y,y,n);
   mpz_add(prev_b_plus_b_inv,x,y); mpz_mod(prev_b_plus_b_inv,prev_b_plus_b_inv,n);

   i-=D/2; // Intervall von [i-D/2,i+D/2] auf [i,i+D] normieren
   while (i<=phase2)
    {
      int runden = 10000; // Anzahl Runden bis zu gcd und Statusausgabe
      mpz_set_ui(y,1); // Produktsammler fr gcd

      while (i<=phase2 && runden>0)
       {
         // jetzt das Intervall [i,i+D] abarbeiten
#if 0
         mpz_add(x,b,b_inv); mpz_mod(x,x,n);
         cout << x << endl;
         cout << b_plus_b_inv << endl;
         cout << endl;
         char ch; cin >> ch;
#endif
         // Primzahlen aussieben
         bool sieve[D];
         for (int j=1-(i%2); j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
         for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
           for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
           /* 
             Bemerkung:
              p-(i%p) ist zwar p, wenn p teilt i;
              aber: den Index 0 brauchen wir spter nicht!
           */

         // Zahlenpaare i+D/2-j,i+D/2+j mit mindestens einer Primzahl auslesen
         for (int j=1, dj=4; j<D/2; j+=dj, dj=6-dj)
         if (sieve[D/2-j] || sieve[D/2+j])
          {
#if 0
            if (!( is_prime(i+D/2-j) || is_prime(i+D/2+j) ))
             {
               cout << "FEHLER!!"
                    << " i= " << i 
	  	    << " j= " << j 
		    << " D/2= " << D/2 
                    << endl;
               exit(1);
             }
#endif
            mpz_sub(x,b_plus_b_inv,valtab[j]); // valtab ohne Displacement!
            mpz_mul(y,y,x); mpz_mod(y,y,n);
            runden--;
          }

         // Werte fr nchstes Intervall bestimmen...
#if 0
         mpz_mul(b,b,d); mpz_mod(b,b,n);
         mpz_mul(b_inv,b_inv,d_inv); mpz_mod(b_inv,b_inv,n);
#endif
         i+=D; // neues Intervall (bis i haben wir nun abgearbeitet)
         mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
         mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
         mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
         mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
       }

      cout << "Phase2: " << i << "/" << phase2 << "\r" << flush;

      // nun (erst) den ggT ermitteln
      mpz_gcd(x,y,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;

	  // nun noch die Exponenten anpassen (damit's schneller geht)
	  for (int j=1; j<D/2; j+=2)
           if (gcd(j,D)==1) mpz_mod(valtab[j],valtab[j],n);
	}

    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  mpz_clear(b_inv); mpz_clear(d); mpz_clear(d_inv); mpz_clear(b_plus_b_inv);
  mpz_clear(d_plus_d_inv); mpz_clear(prev_b_plus_b_inv);
  for (int j=1; j<D/2; j+=2)
   if (gcd(j,D)==1) 
    mpz_clear(valtab[j]); // freeing memory for precomputed values
  }
#endif

#if 1 /* improved standard continuation with pairing & fft */
  {
   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Phi-Phase 2 (improved with pairing & fft)" << endl;

   /* zur Bedeutung nachfolgender Werte siehe elliptische Kurven,
      fft-continuation */
   int pg=256, pms=2*1050;
   if (phase2 >    591360/2*3) { pg=  512; pms= 2* 2310; }
   if (phase2 >   2365440/2*3) { pg= 1024; pms= 2* 4620; }
   if (phase2 >   9461760/2*3) { pg= 2048; pms= 2* 9240; }
   if (phase2 >  39137280/2*3) { pg= 4096; pms= 2*19110; }
   if (phase2 > 160849920/2*3) { pg= 8192; pms= 2*39270; }
   if (phase2 > 648560640/2*3) { pg=16384; pms= 2*79170; }
   //if (phase2 > 2611445760/2*3) { pg=32768; pms=2*159390; }

   const int D = pms;
   const int Polynom_Index = pg;

   cout << "Polynomgre=" << Polynom_Index << endl;
   cout << "Einzel-Intervallsize=" << D << endl;

   mpz_t sammel[Polynom_Index];
   for (int j=0; j<Polynom_Index; ++j) mpz_init(sammel[j]);

   // in a steht die Start-Basis fr Phase 2

  /* Grundidee:
     Ausnutzung der folgenden Gleichheit:
     (a^(i+d) -1) * (a^(i-d) -1) = a^i * ( (a^i +a^(-i)) - (a^d +a^(-d)) )
     
     1. Der Multiplikatior a^i kann weggelassen werden, da er das
        Teilerverhalten beim ggT nicht beeinflut.
     2. In Intervallen der Gre D wird nun i erhht.
         -> Vorberechnung von (a^d +a^(-d)) fr 0<d<D,
         -> auerdem: pro Intervall eine Bestimmung von (a^i + a^(-i))
     3. a^(i+D) lt sich sukzessiv ber a^(i+D)=a^i * a^D bestimmen;
        a^(-(i+D)) lt sich wegen a^(-i)=(a^(-1))^i analog bestimmen.
        -> pro Intervall zwei Multiplikationen, um a^(i+D) + a^(-(i+D))
           zu bestimmen.

    Bemerkung I:
      Die hierauf aufbauende Paarbildungsmethode konstruiert nicht
      die durch andere Methoden erreichbaren Primzahlpaare, sie ist
      aber dafr vergleichsweise einfach zu implementieren und sie
      kommt mit weniger Speicherplatz als die improved-standard-continuation
      aus. Sobald mehrere Primzahlpaare innerhalb eines Intervalls
      auftauchen, ist sie schneller als die improved standard continuation.
   
    Bemerkung II: (zu 3.)
      a^i+a^(-i) lt sich -- wegen x^2-x(a^1+a^(-1))+1=(x-a)*(x-a^(-1)) --
      auch ber Lucasfolgen effizient berechnen:
      a^i+a^(-i)=Vi(a^1+a^(-1))=:V[i];
      a^(i+D)+a^(-(i+D)))=V[i+D]=V[i]*V[D]-V[i-D].
      Man kann also die Anzahl der erforderlichen Multiplikationen pro
      Intervallsprung von 2 auf 1 reduzieren.
      Damit wren wir dann (mit anderer Herleitung und mit leichten
      Modifikationen) wieder auf bereits bekannte Fortsetzungen gestoen
      (vgl. Peter L. Montgomery, "Speeding the Pollard and Elliptic Curve
       Methods of Factorization", MathComp Vol 48 (177), Jan 1987, p.243-264;
       insbesondere Abschnitte 4, 5 und 9).
  */

   mpz_t b_inv, d, d_inv, b_plus_b_inv, d_plus_d_inv, prev_b_plus_b_inv;
   mpz_init(b_inv); mpz_init(d); mpz_init(d_inv); mpz_init(b_plus_b_inv);
   mpz_init(d_plus_d_inv); mpz_init(prev_b_plus_b_inv);

   // im folgenden werden b,b_inv, d,d_inv nicht mehr aktualisiert,
   // da die zugehrigen Summen direkt durch Lucasfolgen berechnet werden.

   // Precomputing most common used values:
   int Polynom_Index_i = 0;
   mpz_set(b,a); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_powm_ui(d,b,2,n); mpz_powm_ui(d_inv,b_inv,2,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
   for (int j=1; j<D/2; j+=2)
    {
      if (gcd(j,D)==1) mpz_set(sammel[Polynom_Index_i++],b_plus_b_inv);

      mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
      mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
      mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
      mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
    }

   cout << "unvergrerte Polynomgre: " << Polynom_Index_i << endl;
   polynomial::TPolynom Polynom; // hier steht das anschlieend konstruierte Polynom drin
   // Rest bis zur nchsten Zweierpotenz mit Nullen auffllen
   /* Bemerkung: Besser wre es, Punkte auerhalb des Suchbereichs
                 zu whlen (Motto: doppeltes, aber auerhalb des
                 Suchbereichs lchriges Netz)
    */
   { 
     int anz_Punkte=Polynom_Index_i;    
     while (anz_Punkte<Polynom_Index) mpz_set_ui(sammel[anz_Punkte++],0);
     Polynom_Index_i=polynomial::Polynom_aus_Nullstellen_konstruieren(Polynom,sammel,anz_Punkte,n);
     cout << "Polynom erzeugt. Polynomindex = " << Polynom_Index_i << endl;
   }

   i=i-(i%D); // Startpunkt ermitteln
   mpz_powm_ui(d,a,D,n); mpz_invert(d_inv,d,n); mpz_mod(d_inv,d_inv,n);
   mpz_add(d_plus_d_inv,d,d_inv); mpz_mod(d_plus_d_inv,d_plus_d_inv,n);
   mpz_powm_ui(b,a,i,n); mpz_invert(b_inv,b,n); mpz_mod(b_inv,b_inv,n);
   mpz_add(b_plus_b_inv,b,b_inv); mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
   mpz_mul(x,b,d_inv); mpz_mod(x,x,n);
   mpz_mul(y,b_inv,d); mpz_mod(y,y,n);
   mpz_add(prev_b_plus_b_inv,x,y); mpz_mod(prev_b_plus_b_inv,prev_b_plus_b_inv,n);

   i-=D/2; // Intervall von [i-D/2,i+D/2] auf [i,i+D] normieren
   while (i<=phase2)
    {
      cout << "Phase2: " << i << "/" << phase2 << endl << flush;
      cout << "Sammle Werte..." << endl;

      //Werte sammeln
      int sammel_index=0;
      while (sammel_index<(Polynom_Index_i-1)/2)
       {
         mpz_set(sammel[sammel_index++],b_plus_b_inv);
         // Werte fr nchstes Intervall bestimmen...
         i+=D; // neues Intervall (bis i haben wir nun abgearbeitet)
         mpz_set(x,prev_b_plus_b_inv); mpz_set(prev_b_plus_b_inv,b_plus_b_inv);
         mpz_mul(b_plus_b_inv,b_plus_b_inv,d_plus_d_inv);
         mpz_sub(b_plus_b_inv,b_plus_b_inv,x);
         mpz_mod(b_plus_b_inv,b_plus_b_inv,n);
       }

      // Polynomauswertung starten
      cout << "Phase2: " << i << "/" << phase2 << endl << flush;
      cout << "starte Polynomauswertung" << endl;
      cout << "Parameter: Polynomindex: " << Polynom_Index_i
           << ", Anzahl der Punkte: " << sammel_index << endl;
      polynomial::multipoint_eval(sammel,Polynom,Polynom_Index_i,sammel,sammel_index,n);
      cout << "Polynomauswertung beendet. Teste auf ggT." << endl;
      mpz_set(y,sammel[0]);
      for (int j=1; j<sammel_index; ++j)
        { mpz_mul(y,y,sammel[j]); mpz_mod(y,y,n); }
      // nun (erst) den ggT ermitteln
      mpz_gcd(x,y,n);
      if (mpz_cmp_ui(x,1)==0) continue;
      
      // nun ist/sind ein/mehrere Teiler(in/nen) gefunden,
      // mgliches Problem: vielleicht mehrere Teiler akkumuliert.
      // Da es aber nur log-viele Teiler geben kann, ist dieser
      // Abschnitt nicht zeitkritisch; also strt es wenig, alle
      // Werte nochmal durchzugehen (dies senkt die Wahrscheinlichkeit,
      // akkumulierte Teiler zu erhalten).
    
      for (int j=1; j<sammel_index; ++j)
       {
	 mpz_gcd(x,sammel[j],n);
         if (mpz_cmp_ui(x,1)!=0)
	  {
	   cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	   if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2]" << flush;
	    }
	   else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [phi2] [composite]" << flush;
	    }
	   mpz_div(n,n,x);
	   if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;
           // anschlieend noch Polynom vereinfachen:
	   for (int j=0; j<Polynom_Index_i; ++j) mpz_mod(Polynom[j],Polynom[j],n);
	  }
       }

    }

 fertig_phase2:
  cout << "Phi-Phase 2 bis " << i << endl;
  mpz_clear(b_inv); mpz_clear(d); mpz_clear(d_inv); mpz_clear(b_plus_b_inv);
  mpz_clear(d_plus_d_inv); mpz_clear(prev_b_plus_b_inv);
  for (int j=0; j<Polynom_Index; ++j) mpz_clear(sammel[j]);
  for (int j=0; j<Polynom_Index_i; ++j) mpz_clear(Polynom[j]);
  delete [] Polynom;
  }
#endif


 fertig:
  mpz_clear(a); mpz_clear(b);
}


void fermat_like_methode(unsigned int bis)
{
  /* Fermat-hnliche Methode: wenn t die Zahl n teilt
     und abs(t - n/t) klein ist, so kann hiermit ein Teiler gefunden werden.
     Alternativ ist auch ein beliebiger Vorfaktor fr abs(k*t - n/t)
     realisiert.
     Diese Methode ist zwar nicht besonders effizient, da sie eher einer
     Probedivision entspricht; sie hilft aber, solche Zahlen besonders
     schnell zu faktorisieren, die auf "allzu naive" Weise als
     "besonders schwer" zu faktorisierende Zahlen konstruiert sind.
  */
  cout << "starte Fermat-hnliche Methode" << endl;
  
  mpz_t c,r,h,h2;
  mpz_init(c); mpz_init(r); mpz_init(h); mpz_init(h2);
  
  unsigned int delta;
  
  for (unsigned int k=1; k<=bis; k++)
    {
      cout << "Vorfaktor ist " << k << "   " << endl;
      mpz_mul_ui(c,n,k);
      mpz_mul_ui(c,c,4); // 4c ist notwendig fr "Quadrat-Formel"
      
      mpz_sqrt(r,c);
      for(delta=0; delta<1000000; delta++)
	{
	  if (delta%100000==0)
	    { printf("delta=%i\r",delta); fflush(stdout); }
	  mpz_add_ui(h,r,delta); mpz_mul(h,h,h); mpz_sub(h,h,c);
	  if (mpz_perfect_square_p(h))
	    {
	      { printf("delta=%i\n",delta); fflush(stdout); }
	      cout << "Square detected..." << endl;
	      mpz_sqrt(h2,h); mpz_add_ui(h,r,delta); mpz_sub(h,h,h2);
	      mpz_gcd(h,h,n);
	      if (mpz_cmp_ui(h,1)!=0) 
		{
		  if (mpz_probab_prime_p(h,probab_prime_checks))
		    {
		      cout << h << " ist Faktor." << endl;
		      Factorization_to_file << MAL() << h << " [fer]" << flush;
		    }
		  else
		    {
		      cout << h << " ist zusammengesetzter Faktor." << endl;
		      Factorization_to_file << MAL() << h << " [fer] [composite]" << flush;
		    }
		  mpz_div(n,n,h);
		  fermat_like_methode(bis); // und noch einmal versuchen...
		  k=bis+1; break;
		}
	    }
	}
    }
  mpz_clear(h); mpz_clear(h2); mpz_clear(r);
  mpz_clear(c);
}



void phimat(mpz_t &n)
{
  mpz_t x,y;
  mpz_init(x); mpz_init(y);
  
  mpz_mul_ui(x,n,4); mpz_sqrt(x,x); // x=trunc(2*sqrt(n));
  mpz_add_ui(y,n,1); mpz_sub(y,y,x); // y=n+1-x
  mpz_set_ui(x,2); mpz_powm(x,x,y,n); // x=2^y mod n;

  cout << "trying phimat..." << x << endl;

  //mpz_add_ui(y,n,1); mpz_div_ui(y,y,2);
  unsigned long int z=0;
  while (z<1000000000)
    {
      int p=mpz_scan1(x,0); z+=p; 
      mpz_tdiv_q_2exp(x,x,p);
      //cout << z << ": " << x << endl;
      if (mpz_cmp_ui(x,1)==0) break;
      mpz_add(x,x,n);
    }

  cout << z << ":" << x << endl;
  
  if (mpz_cmp_ui(x,1)==0)
    {
      cout << "x=1 ... " << endl;
      mpz_mul_ui(x,n,4);mpz_sqrt(x,x); mpz_add_ui(x,x,z); // x=p+q
      mpz_mul(y,x,x);
      mpz_sub(y,y,n); mpz_sub(y,y,n); mpz_sub(y,y,n); mpz_sub(y,y,n);
      mpz_sqrt(y,y); // y=p+q
      mpz_add(y,x,y); mpz_div_ui(y,y,2); // y=p
      mpz_div(x,n,y); // x=q
      mpz_mul(x,x,y);
      cout << x << endl; 
      if (mpz_cmp(x,n)==0) // wirklich echte Teiler ermittelt?
	{
	  mpz_div(x,n,y); // x=q
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [composite]" << flush;
	    }
	  if (mpz_probab_prime_p(y,probab_prime_checks))
	    {
	      cout << y << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << y << flush;
	    }
	  else
	    {
	      cout << y << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << y << " [composite]" << flush;
	    }      
	  mpz_set_ui(n,1);
	}
    }
  
  mpz_clear(x); mpz_clear(y);
}



bool k_te_Wurzel (const mpz_t &n, mpz_t &wurzel, const unsigned int k)
{
  /* Berechnen der abgerundeten, ganzzahligen k. Wurzel der bergebenen Zahl.
     Die Wurzel wird in "wurzel" zurckgegeben.
     Der Funktionswert ist true, falls die Zahl die k-te Potenz der zurckgegebenen
     Wurzel ist, andernfalls false. */
  //cout << "k-wurzel rein..." << k << endl;
  mpz_t d,y;
  mpz_init(d); mpz_init(y);

  
  mpz_set_ui(y,1); int exp = 0;
  do { mpz_mul_2exp(y,y,k); exp++; }  while (mpz_cmp(y,n)<=0);
  mpz_set_ui(d,1); mpz_mul_2exp(d,d,exp);
  
  mpz_set_ui(wurzel,0);
  do
    {
      mpz_tdiv_q_2exp(d,d,1);
      mpz_add(wurzel,wurzel,d);
      // mpz_pow_ui(y,wurzel,k); // klappt aus unerfindlichen Grnden nicht, daher:
      mpz_set_ui(y,1); for (int i=k; i>0; i--) mpz_mul(y,y,wurzel); // langsamer Ersatz.

      if (mpz_cmp(y,n)>0) mpz_sub(wurzel,wurzel,d);
    } while (mpz_cmp_ui(d,0)>0);
  
  
  //mpz_pow_ui(y,wurzel,k);
  bool fl=(mpz_cmp(y,n)==0);
  
  mpz_clear(d); mpz_clear(y);
  //cout << "k-wurzel raus..." << endl;
  return fl;
}


bool Potenztest(const mpz_t &n)
{
  // (reine) Potenzen vorhanden?
  mpz_t y;
  mpz_init(y);
  for (unsigned int k=mpz_sizeinbase(n,256); k>1; k--)
    {
      //cout << "k=" << k << "?" << endl;
      if (k_te_Wurzel(n,y,k))
	{	
	  cout << y << "^" << k <<  " ist ";
	  Factorization_to_file << MAL() << y << "^" << k;
	  if (!mpz_probab_prime_p(y,probab_prime_checks))
	    {
	      cout << "zusammengesetzter ";
	      Factorization_to_file << " [composite]";
	    }
	  Factorization_to_file << flush;
	  cout << "Faktor." << endl;
	  mpz_clear(y);
	  return true; 
	}
    }
  mpz_clear(y);
  return false; // ein Fall fr Viagra...
}

#include "elliptic_curve.cc"

#ifndef IS_CLIENT
void easy_factor()
{
  /* einfache Faktorisierung versuchen:
     vielleicht kann die Zahl dadurch komplett faktorisiert werden
     oder zumindest um einige Faktoren reduziert werden... */
  
  if (mpz_sizeinbase(n, 10)>50)
    fermat_like_methode(1);
  /* Fermat-hnliche Faktorisierung versuchen fr den Fall, da die Zahl
     z.B. die Form (a-d)*(a+d) mit kleinem d hat... */
  
  cout << "Trialdivision up to 10000" << endl;
  {
    /* Probedivision mit 2 und 3 */
    int exp=0;
    while (mpz_div_ui(y,n,2)==0) { mpz_set(n,y); exp++; }
    if (exp>0)
      {
	cout << "2^" << exp << " ist Faktor." << endl;
	Factorization_to_file << MAL() << "2";
        if (exp>1) Factorization_to_file << "^" << exp;
      }
    exp=0;
    while (mpz_div_ui(y,n,3)==0) { mpz_set(n,y); exp++; }
    if (exp>0)
      {
	cout << "3^" << exp << " ist Faktor." << endl;
	Factorization_to_file << MAL() << "3";
        if (exp>1) Factorization_to_file << "^" << exp;
      }
    
    /* Probedivision bis 10000 */
    unsigned long int p=5;
    while (p<10000)
	{
	  exp=0;
	  while (mpz_div_ui(y,n,p)==0) { mpz_set(n,y); exp++; }
	  if (exp>0)
	    {
	      cout << p << "^" << exp << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << p;
              if (exp>1) Factorization_to_file << "^" << exp;
	    }
	  p+=2;
	  
	  exp=0;
	  while (mpz_div_ui(y,n,p)==0) { mpz_set(n,y); exp++; }
	  if (exp>0)
	    {
	      cout << p << "^" << exp << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << p;
              if (exp>1) Factorization_to_file << "^" << exp;
	    }
	  p+=4;
	} 
    
    Factorization_to_file.flush();
  }

  if (Potenztest(n)) mpz_set_ui(n,1);
  if (mpz_cmp_ui(n,1)==0) goto fertig; 
  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;

  {
    unsigned int dezimalstellen = mpz_sizeinbase(n,10);
    tune_parameters(dezimalstellen);
    pollard(rho_Phase); /* Faktorisierung nach Pollard-Rho-Methode versuchen */

    if (Potenztest(n)) mpz_set_ui(n,1);
    if (mpz_cmp_ui(n,1)==0) goto fertig; 
    if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;

    if (dezimalstellen != mpz_sizeinbase(n,10))
     {
       dezimalstellen=mpz_sizeinbase(n,10);
       tune_parameters(dezimalstellen);
     }

    if (phi_Phase1<=0 && phi_Phase2<=0)
     cout << "Phi-Phase ist deaktiviert!" << endl;
    else
     pollard_phi(phi_Phase1,phi_Phase2); /* Faktorisierung nach Pollard-Phi-Methode versuchen */
  }

fertig:  
  cout << "Verbleibende Zahl:" << endl << n << endl;
  
  if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) )
    {
      Factorization_to_file << endl;
      exit(0); /* 1 als Faktor: komplett faktorisiert */
    }  

  if (mpz_probab_prime_p(n,probab_prime_checks))
    {
      Factorization_to_file << MAL() << n << endl;
      cout << "Verbleibende Zahl ist wahrscheinlich prim!" << endl;
      exit(0);
    }
 //Factorization_to_file << " * REST" << endl; exit(0);
}
#endif /* von #ifndef IS_CLIENT */


#include "sqrt_modulo.cc"
#include "mpz_sqrtmod.cc"

inline int SQRT_kN_mod_Primzahl(const unsigned int Primzahl)
{
  return sqrtmod(mpz_mod_ui(x,kN,Primzahl),Primzahl);
}


int check_SQRT_kN_mod_Primzahl(const int Primzahl)
{
  // kann zu Debugging-Zwecken aufgerufen werden, um zu testen,
  // inwieweit die "SQRT_kN_mod_Primzahl"-Funktion korrekt arbeitet
  int w1 = SQRT_kN_mod_Primzahl(Primzahl);
  if (w1>Primzahl/2) w1=Primzahl-w1;
  
  mpz_set_ui(x,w1); mpz_mul(x,x,x); mpz_sub(x,x,kN);
  if (mpz_mod_ui(x,x,Primzahl)!=0)
    {
      cerr << "WURZELALARM! " << w1 << " mod " << Primzahl << endl;
      exit(1);
    }
  return w1;
}


void Statische_Faktorbasis_ermitteln()
{ 
  cout << "Ermittle die statische Faktorbasis." << endl;
  Primzahlen[0]=-1; /* nullte "Primzahl" ist -1 (Bercksichtigung des
		       Vorzeichens bei den Relationen! */
  log_Primzahlen[0]=0; // Vorzeichen hat keinen Logarithmuswert

  int Primzahl=2;
  // die Primzahl 2 ist ein Sonderfall. Ich gehe davon aus, da sie
  // auf jeden Fall in der Faktorbasis vorhanden sein soll.
  Primzahlen[1]=2; log_Primzahlen[1]=static_cast<TSiebElement>(ceil(log_Multiplikator*log(Primzahl)));

  Primzahl=1; // Initialwert 1 nur, wenn in der Schleife +=2 gezhlt wird!
  mpz_t wuqu; mpz_init(wuqu); // Fr Berechnung der Quadratwurzeln von Primzahlpotenzen
  for (int Primzahl_nr=2; Primzahl_nr<Primbasis_size; Primzahl_nr++)
    {
      do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; while(!is_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(x,Primzahl);
	}
      while (mpz_legendre(kN,x)<0); /* Primzahl gltig fr Primzahlbasis? */
      
      //printf("%i. Primzahl der Faktorbasis ist %i\r",Primzahl_nr,Primzahl);
      Primzahlen[Primzahl_nr]=Primzahl;
      
      log_Primzahlen[Primzahl_nr]=static_cast<TSiebElement>(ceil(log_Multiplikator*log(Primzahl)));
      // fr effizientes Sieben erforderlich (erspart dauernde Neuberechnung)
      
      SQRT_kN_der_Primzahlen[Primzahl_nr]=SQRT_kN_mod_Primzahl(Primzahl);
      // fr sptere Ermittlung der Deltawerte zum Sieben erforderlich
      
      // ggf. Prfung, ob Wurzel korrekt ist:
      mpz_set_ui(x,SQRT_kN_der_Primzahlen[Primzahl_nr]);
      mpz_mul(x,x,x); mpz_sub(x,kN,x); 
      if (mpz_mod_ui(x,x,Primzahl)!=0)
       {
         cerr << "Wurzel nicht korrekt!" << endl;
         cerr << SQRT_kN_der_Primzahlen[Primzahl_nr] << "^2 <> "
              << mpz_mod_ui(x,kN,Primzahl) << " (mod " << Primzahl << ")" << endl;
         exit(1);
       }

      if (Primzahl<=2*M/Primzahl && Primzahl>2 && mpz_mod_ui(x,kN,Primzahl)!=0)
	// Primzahl darf nicht zu gro sein, ungleich 2 und kein Teiler von kN, um mit Potenzen gesiebt zu werden
	{
	  FB_maxQuadrate=Primzahl_nr; /* nach Abschlu der Schleife ist dies der Wert, bis zu dem
					 Quadrate auf normale Weise gesiebt werden. */
	  
	  // Wurzeln von kN modulo Primzahl^pot ermitteln	  
	  // Initialwerte:
	  int P_Potenz = Primzahl*Primzahl;
	  int Wu       = SQRT_kN_der_Primzahlen[Primzahl_nr];
	  
	  for (short int pot=2; ;pot++)
	    {
	      //cout << "berechne Wurzel fr " << Primzahl << "^" << pot << endl;

	      mpz_set_ui(y,2*Wu); mpz_set_ui(x,P_Potenz);
	      if (mpz_invert(y,y,x)==0)
		{ 
		  cerr << "PPot-Wurzel fr " << Primzahl  << ": Inverses existiert nicht!" << endl;
		  break;
		}
	      
              if (anz_weitere_Primzahlpotenzen>=FB_max_weitere_Potenzen)
               { 
                 static char weitermachen = 'n';
                 if (weitermachen!='y')
                  {
	           cerr << "berlauf ab " << Primzahl << "^" << pot << endl;
                   cerr << "Es kann nicht optimal mit Primzahlpotenzen gesiebt werden," << endl
                        << "da der hierfr reservierte Speicherplatz nicht ausreicht." << endl;
                   cerr << "Sie sollten \"FB_max_weitere_Potenzen\" vergroessern!" << endl;
                   cerr << "Reserved memory for storing prime powers is to small..." << endl;
                   cerr << "sieving will be less efficient..." << endl;
                   cerr << "continue? (y/n) " << flush;
                   cin >> weitermachen;
                   if (weitermachen=='n') exit(1);
                  }
                 break; // berlauf durch Schleifenabbruch verhindern
               }

	      mpz_mod_ui(x,kN,P_Potenz); mpz_set_ui(wuqu,Wu); mpz_mul_ui(wuqu,wuqu,Wu); mpz_sub(x,x,wuqu);
	      mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod_ui(x,x,P_Potenz);
	      register int h = mpz_get_ui(x); if (h>P_Potenz-h) h=P_Potenz-h; // auf die "kleinste" Wurzel normieren
	      Primzahlpotenzen[anz_weitere_Primzahlpotenzen]=P_Potenz;
	      SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]=h;
	      if ( pot==2 && Primzahl_nr<sieben_ab_Primzahlindex )
		// Exponent 2, weil erstes Auftreten des Faktors
		log_Primzahl_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]
                 =static_cast<TSiebElement>(ceil(log_Multiplikator*2*log(Primzahl)));
	      else
		// Exponent 1, weil nochmaliges Auftreten des Faktors
		// beachte: wenn man schon einmal durch p geteilt hat, ergibt ein weiteres Teilen den Teiler p^2
		log_Primzahl_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]
                 =static_cast<TSiebElement>(ceil(log_Multiplikator*log(Primzahl)));
	      
	      // ggf. Prfung, ob Wurzel korrekt ist:
	      mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]);
	      mpz_mul(x,x,x); mpz_sub(x,kN,x); 
	      if (mpz_mod_ui(x,x,P_Potenz)!=0)
		{
		  cerr << "PPot-Wurzel nicht korrekt!" << endl;
		  cerr << "Primzahl=" << Primzahlen[Primzahl_nr] << " Nr. " << Primzahl_nr << endl;
		  cerr << "kN= " << kN << endl;
		  cerr << "SQRT(kN) mod p=" << SQRT_kN_der_Primzahlen[Primzahl_nr] << endl;
		  cerr << "SQRT(kN) mod p^i=" << SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen] << endl;
		  cerr << "aber falsch: " << x << endl;
		  exit(1);
		}	  
	      
	      anz_weitere_Primzahlpotenzen++; // dies war eine weitere aufgenommene Primzahlpotenz

	      if (P_Potenz>2*M/Primzahl) break; // nchste Primzahlpotenz wird zu gro...
	      P_Potenz*=Primzahl; Wu=SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen-1];
	    }
	  
	}
    }
  mpz_clear(wuqu);

  biggest_Prime_in_Factorbase=Primzahl; 
  cout << "Primzahlen der statischen Faktorbasis ermittelt." << endl;
  
  cout << "Hinweis: ";
  if (Primzahl<Siebgroesse)
    cout << "Alle Primzahlen der statischen Faktorbasis passen in das Sieb!" << endl;
  else
    cout << "Einige Primzahlen der statischen Faktorbasis sind grer als das Sieb!" << endl;
}



void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset)
{
  TSieb_Delta sieb_delta;
  int d1,d2;
  unsigned int inv_A2, PolyB;
  
  sieb_delta.factor=DynRel.factor;

  inv_A2=invmod(mpz_mod_ui(x,Polynom.A2,DynRel.factor),DynRel.factor);
  PolyB=mpz_mod_ui(x,Polynom.B,DynRel.factor);
  
  //mpz_set_ui(x,DynRel.SQRT_kN);
  //mpz_sub(x,x,Polynom.B); mpz_mul_ui(x,x,inv_A2);
  //d1=mpz_mod_ui(x,x,DynRel.factor);

  d1=DynRel.SQRT_kN-PolyB; if (d1<0) d1+=DynRel.factor;
  d1=mulmod(d1,inv_A2,DynRel.factor);
  sieb_delta.delta=(d1-offset)%DynRel.factor;
  if (sieb_delta.delta<0) sieb_delta.delta+=DynRel.factor;
  sieb_delta.delta+=offset;
  if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
    Sieb_Delta_Heap.push(sieb_delta);
  
  //mpz_set_ui(x,DynRel.SQRT_kN); mpz_neg(x,x);
  //mpz_sub(x,x,Polynom.B); mpz_mul_ui(x,x,inv_A2); 
  //d2=mpz_mod_ui(x,x,DynRel.factor);
  d2=DynRel.factor-DynRel.SQRT_kN-PolyB; if (d2<0) d2+=DynRel.factor;
  d2=mulmod(d2,inv_A2,DynRel.factor);

  sieb_delta.delta=(d2-offset)%DynRel.factor;
  if (sieb_delta.delta<0) sieb_delta.delta+=DynRel.factor;
  sieb_delta.delta+=offset;
  if (d1 != d2)
    {
      if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
	Sieb_Delta_Heap.push(sieb_delta);
    }
}

void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor)
{
  // Wenn ein neuer dynamischer Primfaktor entdeckt wird, so empfiehlt es sich,
  // in der Menge der Specialfaktoren (welches zusammengesetzte Faktoren sind) einen
  // Teilbarkeitstest auf diesen Primfaktor durchzufhren. Auf diese Weise knnen
  // weitere Primfaktoren isoliert werden, die als dynamische Faktoren zur Verfgung stehen.
  
  static int Trashcounter = 0; /* Die (nach Elimination eines Faktors) nutzlos auf dem File
				  befindlichen Relationen */
  
  TSpecial_Factor_Relations::iterator pos;
  TSpecialFactorRelation SearchRelation;
  SearchRelation.factor.set_for_search(Primfaktor);

  while (true)
   {
    //cout << "splitten " << Primfaktor << endl;
    /* Wegen Rekusion etc. immer wieder korrekt neu aufsetzen, wenn
       ein Specialfaktor gekegelt wurde -> nchsthherer Special-Factor */
    pos=Special_Factor_Relations.lower_bound(SearchRelation);
    if (pos==Special_Factor_Relations.end()) break;
    if (!(*pos).factor.DLP_enthaelt_Teiler(Primfaktor)) break;

    TSpecialFactorRelation gemerkte_Relation = *pos;
    Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
    gemerkte_Relation.factor.swap();
    Special_Factor_Relations.erase(gemerkte_Relation); // alten Specialfactor entfernen (geswapte Kopie)
    gemerkte_Relation.factor.swap();
     
    mpz_t x; mpz_init(x); // als lokale Variable erforderlich
    mpz_set_ui(x,(*pos).factor/Primfaktor);

    // zusammenpuzzeln...
    CGleichungszeile *GL = new CGleichungszeile;
    GL->kombinieren(SpecialRelations_from_file,gemerkte_Relation.fpos);
    {
      streampos pos;
      pos = SpecialRelations_to_file.tellp(); // position merken
      SpecialRelations_to_file.seekp(gemerkte_Relation.fpos); SpecialRelations_to_file << "U "; // ungltig markieren
      SpecialRelations_to_file.seekp(pos); // position restaurieren
    }

    TDynamicFactorRelation relation;
    relation.factor=Primfaktor;
    TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(relation);
    if (p==Dynamic_Factor_Relations.end()) 
      { 
        cerr << "Fehler in dynamic factorbase! (Specialfaktoren splitten)" << endl;
        exit(1);
      }
    mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
    if ((*p).GLcached()) GL->kombinieren( *((*p).GLcached()) );
    else GL->kombinieren(DynamicRelations_from_file,(*p).fpos);
	
    if (mpz_probab_prime_p(x,probab_prime_checks))
     if (mpz_cmp_ui(x, Specialfactor_Sieb_Threshold)<=0)
       { 
         // einen dynamischen Faktor entdeckt!

	 // es liegt nun der ermittelte dynamic Factor in x korrekt vor
	 GL->relevanter_Faktor=dynamic_factor;
	 relation.factor=mpz_get_ui(x);
	      
	 // aber ist er neu oder existierte er schon in der dyn. FB?
	 p = Dynamic_Factor_Relations.find(relation);
	 if (p==Dynamic_Factor_Relations.end())
	   { 
	     // er ist neu!
	     //cout << "Special Factor: Dynamischer Faktor erzeugt!" << endl;
	     Special_to_dynamic_Factor_hit++;
	     relation.fpos = GL->speichern(DynamicRelations_to_file,relation.factor);
	     relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	     Deltas_berechnen_fuer(relation, Sieboffset+Siebgroesse);
             // falls geeignet (klein, kurz, Speicher frei,...), Relation cachen
             if (!relation.insert_cache(GL)) delete GL; // nicht gecached
	     Dynamic_Factor_Relations.insert(relation); // Dynamicfactor einfgen
	     Special_Factoren_durch_Primfaktor_splitten(relation.factor);
       	   }
         else 
	   { 
	     // er existierte schon in der dyn. FB -> er fhrt zu einem Treffer im GLS!
	     cout << "Special Factor: Special Hit!" << endl;
	     // gefundenen dynamischen Faktor kombinieren:
      	     mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
	     if ((*p).GLcached()) GL->kombinieren( *((*p).GLcached()) );
	     else GL->kombinieren(DynamicRelations_from_file,(*p).fpos);
	     Special_hit++; Relation_einfuegen(GL);
	   } 
       }
     else // x prim und x>SpecialFactorThreshold
       {
         // unser x ist zwar nun prim, liegt aber auerhalb des Bereichs fr
         // die dynamische Faktorbasis -> er ist nutzlos
         // ---> entfernen
         cout << "Specialfaktor aus Specialfaktormenge ist nutzlos geworden!" << endl;
	 delete GL;
       }
    else // x immer noch zusammengesetzt
     {
       cerr << "Specialfaktor trotz Elimination eines Primfaktors noch zusammengesetzt!" << endl;
       // es liegt nun ein ermittelter neue special Factor in x korrekt vor
       cerr << "splitten: in der aktuellen Version nur Double Large Prime implementiert!" << endl;
       cerr << "gesplitteter DLP sollte daher *nicht* zusammengesetzt sein." << endl;
       exit(1);	    
     }
	
    mpz_clear(x); // lokale Variable wieder freigeben
    Trashcounter++; // Die auf dem File verbleibenden sinnlosen Relationen zhlen 
   }

 // Wenn sich sehr viel Mll angesammelt hat: Meldung erstatten
 if (Trashcounter>200000)
   {
     // Hier knnte man nun eine Kompaktifizierung des Special-Relation-Files
     // durchfhren. Das wrde aber fr den laufenden Serverbetrieb zu lange
     // dauern. Stattdessen wird ab jetzt lediglich eine Meldung ausgegeben,
     // in welcher ein Runterfahren des Programms und ein anschlieender
     // Neustart desselben empfohlen wird. (Beim Recovery wird eine
     // Kompaktifizierung automatisch durchgefhrt.)
     clog << "Das Special-Relations-File hat einen hohen Anteil an redundanten" << endl;
     clog << "Relationen, die eliminiert werden knnten." << endl;
     clog << "Ein Runterfahren (Abbruch) und erneutes Starten (Recovery) des" << endl;
     clog << "Programms wrde eine Kompaktifizierung des Files initiieren." << endl;
     Trashcounter-=100; // in Krze die Meldung wiederholen...
   }
}

//Vorwrtsdeklarationen:
int hole_Wert (istream &in);
void schreibe_Wert (ostream &out, const int w);

void Special_Factoren_durch_Primfaktor_splitten_LAZY(const int Primfaktor)
{
  // Wenn ein neuer dynamischer Primfaktor entdeckt wird, so empfiehlt es sich,
  // in der Menge der Specialfaktoren (welches zusammengesetzte Faktoren sind) einen
  // Teilbarkeitstest auf diesen Primfaktor durchzufhren. Auf diese Weise knnen
  // weitere Primfaktoren isoliert werden, die als dynamische Faktoren zur Verfgung stehen.
  // Im Gegensatz zur ansonsten gleichnamigen BUSY-Methode verwandelt die LAZY-Methode
  // den abgesplitteten Primfaktor in einen Korrekturfaktor. Die eigentliche
  // Auswertung der Relation wird damit auf einen spteren Zeitpunkt verschoben
  // und mglicherweise ganz verhindert, falls die betroffene Relation nicht
  // bentigt wird.

  static int Trashcounter = 0; /* Die (nach Elimination eines Faktors) nutzlos auf dem File
				  befindlichen Relationen */
  
  TSpecial_Factor_Relations::iterator pos;
  TSpecialFactorRelation SearchRelation;
  SearchRelation.factor.set_for_search(Primfaktor);

  while (true)
   {
    //cout << "splitten " << Primfaktor << endl;
    /* Wegen Rekusion etc. immer wieder korrekt neu aufsetzen, wenn
       ein Specialfaktor gekegelt wurde -> nchsthherer Special-Factor */
    pos=Special_Factor_Relations.lower_bound(SearchRelation);
    if (pos==Special_Factor_Relations.end()) break;
    if (!(*pos).factor.DLP_enthaelt_Teiler(Primfaktor)) break;

    TSpecialFactorRelation gemerkte_Relation = *pos;
    Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
    gemerkte_Relation.factor.swap();
    Special_Factor_Relations.erase(gemerkte_Relation); // alten Specialfactor entfernen (geswapte Kopie)
    gemerkte_Relation.factor.swap();
    
    TDynamicFactorRelation relation;
    relation.factor=Primfaktor;
    TDynamic_Factor_Relations::iterator p_el,p;
    p_el = Dynamic_Factor_Relations.find(relation);
    if (p_el==Dynamic_Factor_Relations.end()) 
      { 
        cerr << "Fehler in dynamic factorbase! (Specialfaktoren splitten)" << endl;
        exit(1);
      }

    relation.factor=(*pos).factor/Primfaktor; // Dies ist der neue Restfaktor
    // es liegt nun der ermittelte dynamic Factor korrekt vor
	      
    // aber ist er neu oder existierte er schon in der dyn. FB?
    p = Dynamic_Factor_Relations.find(relation);
    if (p==Dynamic_Factor_Relations.end())
     { 
       // er ist neu!
       //cout << "Special Factor: Dynamischer Faktor erzeugt!" << endl;
       Special_to_dynamic_Factor_hit++;
       relation.fpos = DynamicRelations_to_file.tellp();
       // nun die Relation modifiziert umkopieren:
       {
         streampos rpos=SpecialRelations_from_file.tellg(); // Position merken
         SpecialRelations_from_file.seekg(gemerkte_Relation.fpos); // Position merken
         Tfactor ret; string s;
         // Dateiformatflags retten
         ios::fmtflags oldFlagsR = SpecialRelations_from_file.flags(); // alte Formatflags retten
         SpecialRelations_from_file.unsetf(ios::hex); SpecialRelations_from_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         ios::fmtflags oldFlagsW = DynamicRelations_to_file.flags(); // alte Formatflags retten
         DynamicRelations_to_file.unsetf(ios::hex); DynamicRelations_to_file.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
         SpecialRelations_from_file >> s; // Startzeichen
         if (s != "G")
          {
           cerr << "Error: Falsche Position beim Einlesen der Gleichungszeile" << endl;
           exit(1);
          }
         SpecialRelations_from_file >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
         SpecialRelations_from_file >> s; // delta
         DynamicRelations_to_file << "G " << setprecision(20) << relation.factor << " " << s << " ";

         // ab jetzt wird hexadezimal gearbeitet:
         SpecialRelations_from_file.unsetf(ios::dec); SpecialRelations_from_file.setf(ios::hex); // Hexadezimalausgabe
         SpecialRelations_from_file.unsetf(ios::showbase); // ohne Basiskennzeichnung
         DynamicRelations_to_file.unsetf(ios::dec); DynamicRelations_to_file.setf(ios::hex); // Hexadezimalausgabe
         DynamicRelations_to_file.unsetf(ios::showbase); // ohne Basiskennzeichnung

         int distance=hole_Wert(SpecialRelations_from_file);
         while (distance>=0)
          {
            schreibe_Wert(DynamicRelations_to_file,distance);
            distance=hole_Wert(SpecialRelations_from_file);
          }
         schreibe_Wert(DynamicRelations_to_file,-2);
         schreibe_Wert(DynamicRelations_to_file,Primfaktor);
         if (distance==-2)
          {
            distance=hole_Wert(SpecialRelations_from_file);
            while (distance>=0)
             {
               schreibe_Wert(DynamicRelations_to_file,distance);
               distance=hole_Wert(SpecialRelations_from_file);
             }            
          }
         schreibe_Wert(DynamicRelations_to_file,-1); DynamicRelations_to_file << endl;
         SpecialRelations_from_file.flags(oldFlagsR); // alte Formatflags wiederherstellen
         DynamicRelations_to_file.flags(oldFlagsW); // alte Formatflags wiederherstellen
         SpecialRelations_from_file.seekg(rpos); // position restaurieren
       }
       relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
       Deltas_berechnen_fuer(relation, Sieboffset+Siebgroesse);
       Dynamic_Factor_Relations.insert(relation); // Dynamicfactor einfgen
       Special_Factoren_durch_Primfaktor_splitten_LAZY(relation.factor);
     }
    else 
     { 
       // er existierte schon in der dyn. FB -> er fhrt zu einem Treffer im GLS!
       cout << "Special Factor: Special Hit!" << endl;
       CGleichungszeile *GL = new CGleichungszeile;
       GL->kombinieren(SpecialRelations_from_file,gemerkte_Relation.fpos);
       // gefundene dynamische Faktoren kombinieren:
       mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
       if ((*p).GLcached()) GL->kombinieren( *((*p).GLcached()) );
       else GL->kombinieren(DynamicRelations_from_file,(*p).fpos);
       // nicht vergessen: auch Primfaktor eliminieren!
       mpz_mul_ui(GL->Delta,GL->Delta,(*p_el).factor); mpz_mod(GL->Delta,GL->Delta,n);
       if ((*p_el).GLcached()) GL->kombinieren( *((*p_el).GLcached()) );
       else GL->kombinieren(DynamicRelations_from_file,(*p_el).fpos);
       Special_hit++; Relation_einfuegen(GL);
     } 

    { // alte Special-Relation streichen
      streampos pos;
      pos = SpecialRelations_to_file.tellp(); // position merken
      SpecialRelations_to_file.seekp(gemerkte_Relation.fpos); SpecialRelations_to_file << "U "; // ungltig markieren
      SpecialRelations_to_file.seekp(pos); // position restaurieren
    }

    Trashcounter++; // Die auf dem File verbleibenden sinnlosen Relationen zhlen 
   }

 // Wenn sich sehr viel Mll angesammelt hat: Meldung erstatten
 if (Trashcounter>200000)
   {
     // Hier knnte man nun eine Kompaktifizierung des Special-Relation-Files
     // durchfhren. Das wrde aber fr den laufenden Serverbetrieb zu lange
     // dauern. Stattdessen wird ab jetzt lediglich eine Meldung ausgegeben,
     // in welcher ein Runterfahren des Programms und ein anschlieender
     // Neustart desselben empfohlen wird. (Beim Recovery wird eine
     // Kompaktifizierung automatisch durchgefhrt.)
     clog << "Das Special-Relations-File hat einen hohen Anteil an redundanten" << endl;
     clog << "Relationen, die eliminiert werden knnten." << endl;
     clog << "Ein Runterfahren (Abbruch) und erneutes Starten (Recovery) des" << endl;
     clog << "Programms wrde eine Kompaktifizierung des Files initiieren." << endl;
     Trashcounter-=100; // in Krze die Meldung wiederholen...
   }
}


// fr die Speicherung der gefundenen dynamischen Faktoren.
// Bemerkung: Um die dynamischen Faktoren bereits in der Siebphase
//            aufzusammeln, ist das Array global.
//             (Dafr wird die Anzahl der gefundenen dynamischen Faktoren
//              fr die aktuelle Relation dem Konstruktor bergeben.)
const short int MaxHits=50; // bei Bedarf erhhen...
static struct {int Faktor, Exponent;} Hits[MaxHits]; // Dynamische Special-Factoren der Relation aufsammeln
// wichtiger Hinweis: Der Exponent zum Faktor wird nicht beim Sieben, sondern erst bei der Refaktorisierung
// ermittelt. Falls also mit Primzahlpotenzen gesiebt werden sollte, braucht der Exponent nicht eingetragen
// zu werden. Als Faktor sollte in jedem Falle die Primzahl und nicht die Primzahlpotenz eingetragen werden!

CGleichungszeile::CGleichungszeile(const signed int Siebpos, const short int HitCount=0)
  /* Konstruktor: neue Relation einfgen.
     "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */
{
  // cout << "initialisiere Gleichungszeile" << endl;
  Relation_sparse = new CTinyFBsizetypeVector;
  Relation_dense=NULL;
  mpz_init(Delta);
  Polynom.Werte_holen(Siebpos,Delta,y);
  // in y: reduziertes Quadrat (ist nun zu faktorisieren...) 
  //cout << "y= "; mpz_out_str(stdout,10,y); cout << endl;
  mpz_invert(Delta,Delta,n); /* wir verwenden den Kehrwert von Delta,
				das erspart eine zustzliche Variable;
				gleichzeitig ist damit die zu findende
				quadratische Kongruenz auf Delta^2=1 (mod n)
				normiert. */
  
  int z;
  if (mpz_sgn(y)<0) // Vorzeichenbehandlung 
    {
      Relation_sparse->append(0); 
      /* "nullte Primzahl" ist -1 wegen Bercksichtigung
	 des Vorzeichens. Also die 0 fr Position von  -1 einfgen */
      mpz_neg(y,y); /* -1 war Teiler -> positiv machen */
    }
  
  // ersten Faktor der statischen Faktorbasis ebenfalls gesondert behandeln,
  // denn die generierten Polynome haben Probleme, die Primzahl 2 zu sieben...
  z=0;
  while (mpz_div_ui(x,y,Primzahlen[1])==0) { mpz_set(y,x); z++; }
  if (z%2==1) Relation_sparse->append(1); /* ungerade Protenz der i. Primzahl markieren */
  if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
  {
    if (z<4) mpz_mul_ui(Delta,Delta,Primzahlen[1]);
    else
      {
	mpz_set_ui(x,Primzahlen[1]); mpz_powm_ui(x,x,z/2,n);
	mpz_mul(Delta,Delta,x);
      }
  }
  
  
  // find factors in dynamic factorbase
  short int HitCount2 = HitCount;
  while (HitCount2--)
    {
      int Exponent=0; // Initialisierung des Exponenten	
      while (mpz_div_ui(x,y,Hits[HitCount2].Faktor)==0)
	{
	  mpz_set(y,x); // getestete Division verwirklichen 
	  Exponent++; // Exponent fr Faktor inkrementieren
	}
      //if (Exponent==0)
      //{ // nur testweise fr debuggingzwecke
      //  cerr << "Falscher Teiler in Hits[]: " << Hits[HitCount2].Faktor << endl;
      //  exit(1);
      //}
      if (Exponent>1)
	{
	  mpz_set_ui(x, Hits[HitCount2].Faktor);
	  mpz_powm_ui(x,x,Exponent/2, n);
	  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
	  Exponent%=2;
	}
      Hits[HitCount2].Exponent = Exponent;
    }
  

  // find factors in stactic factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )
  int i;
  for (i=2; i<Primbasis_size; i++)
    {

      { // shortcut:
        /* keine aufwendige Division durchfhren, wenn diese
           Primzahl gar nicht ausgesiebt wurde */ 
        register int d;
        d=Siebpos%Primzahlen[i]; if (d<0) d+=Primzahlen[i];
        // Bemerkung: %-operation does *not* eliminate the sign!
        //            ugly: -1=1 (mod 2), but (-1)%2=-1, 1%2=1
        // -> so we have to eliminate the sign!!!
        // and don't try to use unsigned int d, this won't work!

        // for debugging reasons:
        /*
        bool b1=(mpz_div_ui(x,y,Primzahlen[i])==0);
        bool b2=(d==Delta_der_Primzahlen[i][0] || d==Delta_der_Primzahlen[i][1]);
        if (b1!=b2) 
         {
           cerr << "error in shortcut: " << Siebpos << "; "
                << Primzahlen[i] << ": " << d << ", "
                << Delta_der_Primzahlen[i][0] << ", "
                << Delta_der_Primzahlen[i][1] << endl;
           exit(1);
         }
         */

        if (d!=Delta_der_Primzahlen[i][0] && d!=Delta_der_Primzahlen[i][1]) continue;
      }

      // nun den Faktor rausdividieren
      mpz_div_ui(y,y,Primzahlen[i]); z=1;
      while (mpz_div_ui(x,y,Primzahlen[i])==0) { mpz_set(y,x); z++; }

      // Relation updaten:
      if (z%2==1)
	Relation_sparse->append(i); /* ungerade Protenz der i. Primzahl markieren */
      if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
	{
	  if (z<4) mpz_mul_ui(Delta,Delta,Primzahlen[i]);
	  else
	    {
	      mpz_set_ui(x,Primzahlen[i]); mpz_powm_ui(x,x,z/2,n);
	      mpz_mul(Delta,Delta,x);
	    }
	  mpz_mod(Delta,Delta,n); // Wert klein halten...
	} 


      // -------------------------------------------------------------------
      // hier wre die Implementierung eines shortcuts mglich:
      // a) Falls es wahrscheinlich erscheint, da der verbleibende Faktor
      // prim ist, kann man dies testen und ggf. abbrechen.
      // (Wenn darberhinaus die Wahrscheinlichkeit fr einen dynamischen
      // Faktor gering ist, sollte man dies mit der Quadratwurzel anstatt
      // einem Primzahltest machen.)
      // Es ist sichergestellt, da ein isolierter statischer Restfaktor
      // in y korrekt bearbeitet wird, da dieser shortcut in hnlicher
      // Form frher implementiert war. Leider kehrt die Double-Large-Prime-
      // Variante die Wirkung dieser Optimierungsmglichkeit um, da dann
      // der verbleibende Rest eben meist nicht prim ist.
      // Aus diesem Grund wurde der shortcut deaktiviert.
      // b) weniger zeitintensiv wre ein Test auf 1, denn dann kann
      // definitiv abgebrochen werden. Leider drfte sich auch das nur
      // lohnen, wenn die Wahrscheinlichkeit, da die Relation vollstndig
      // in der statischen Faktorbasis zerfllt, sehr hoch ist (dann drfte
      // aber der Primzahltest effektiver sein.)
      // c) Aus diesen Grnden wurde ein frher implementierter,
      // hnlich wirkender shortcut wieder entfernt.
      // (Allerdings wurde die korrekte Behandlung eines statischen
      // Restfaktors im Programm belassen, da sie wenig zeitintensiv ist.)
      // -------------------------------------------------------------------
    }
  //clog << i << "/" << Primbasis_size << endl;
  
  // Restfactor in static factorbase...
  bool existiert_grosser_Restfaktor = (mpz_cmp_ui(y,biggest_Prime_in_Factorbase)>0);

  if (existiert_grosser_Restfaktor && collecting_phase_beendet)
    { relevanter_Faktor=special_factor; return; } // vorzeitiges Abbrechen

  if (!existiert_grosser_Restfaktor)
    {
      int p=mpz_get_si(y);
      if (p>1) 
	{
          //cout << "Restfaktor: ab=" << i;
          // binre Suche:
          int b = Primbasis_size-1-i;
          while (b>>=1) if (p>=Primzahlen[i+b]) i+=b;
          //cout << " bin=" << i;
	  while (i<Primbasis_size && p!=Primzahlen[i]) i++;
          //cout << " bis=" << i << endl;
	  if (i<Primbasis_size)
	    Relation_sparse->append(i); /* Index der Primzahl eintragen, falls
					   Primzahl in Faktorbasis vorhanden ist */
	  else 
	    {
	      cerr << "Fehler: unexpected Factorwert: ";
	      mpz_out_str(stderr,10,y); cerr << " -- " << p << endl;
	      // da der Fehler bisher nur bei kleinen Zahlen auftritt,
              // scheint es ein berlauf beim Sieben zu sein.
              // eventuell ist die Faktorbasis zu gro...
              // Oder ist es doch ein BUG?
	      exit(1);
	    }
	  mpz_set_ui(y,1);
	}
      
      relevanter_Faktor=groesster_Faktor_in_Relation();
      // mit der aufgenommenen Relation lt sich dieser Faktor eliminieren
    }
#ifndef IS_CLIENT  
  // kombiniere Faktoren der dynamic Factorbase
  HitCount2 = HitCount;
  while (HitCount2--) if(Hits[HitCount2].Exponent!=0)
    {
      //cout << "kombiniere dynamischen Factor " << Hits[HitCount2].Faktor << endl;
      TDynamicFactorRelation FaRelSearch;
      FaRelSearch.factor=Hits[HitCount2].Faktor;
      TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(FaRelSearch);
      if (p==Dynamic_Factor_Relations.end())
	{
	  cerr << "Fehler in dynamic factorbase! " << FaRelSearch.factor << " not found!" << endl;
          // entweder tatschlich ein Fehler ODER
          // denkbarer (unwahrscheinlicher) Sonderfall:
          // ein gepushtes Primzahl-Quadrat auerhalb der dynamischen
          // Faktorbasis geht (sogar) mit ungerader Potenz ein...
          // ...dann liegt zwar kein Fehler vor, aber die Relation ist
          // trotzdem ungltig. Der Fall erscheint mir aber derart
          // unwahrscheinlich, da ich dann trotzdem abbreche.
          // Hinzu kommt: Wenn ein Client einen solchen Faktor sieben sollte,
          // dann wird diese Bearbeitung hier gar nicht aufgerufen und
          // der Server wird im Fehlerfall die Relation nicht akzeptieren.
          // Dann gibt es zwar eine mgliche dynamische Relation weniger,
          // aber das sollte bei der Seltenheit dieses Falles nicht stren.
	  exit(1);
	}
      mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
      if ((*p).GLcached()) kombinieren( *((*p).GLcached()) );
      else kombinieren(DynamicRelations_from_file,(*p).fpos);
      //cout << "fertig kombiniert von " << *pos << endl;
    }
#endif
  if (existiert_grosser_Restfaktor)
    {
      if (mpz_probab_prime_p(y,probab_prime_checks))
	if (mpz_cmp_ui(y, Specialfactor_Sieb_Threshold)<=0)
	  {
	    // es liegt ein neuer dynamic Factor vor
	    //cout << "neuer dynamic-factor..." << endl;
	    relevanter_Faktor=dynamic_factor;
	    TDynamicFactorRelation relation;
	    relation.factor=mpz_get_ui(y);
	    relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	    Deltas_berechnen_fuer(relation, Sieboffset+Siebgroesse);

#ifndef IS_CLIENT
#if 0
            if (Fuellung_GLS+2*SizeOfRelation()>Primbasis_size)
             {
               Relation_optimieren(0);
	       relevanter_Faktor=dynamic_factor;
             }
#endif
	    relation.fpos = speichern(DynamicRelations_to_file,relation.factor);
            if (relation.insert_cache(this)) relevanter_Faktor=cached_dynamic_factor;
#else
	    // an das Ausgabefile
	    //	    communication_stream << "Relation! ";
	    relation.fpos=speichern(communication_stream,relation.factor, HitCount);
#endif
	    Dynamic_Factor_Relations.insert(relation);
#ifndef IS_CLIENT
	    Special_Factoren_durch_Primfaktor_splitten(relation.factor);
#endif
	    return;
	  }
	else 
	  {
	    // es liegt ein Special-Factor vor, den wir aber ignorieren,
	    // weil er prim ist
	    relevanter_Faktor=special_factor;
	    return;
	  }
      else // es liegt ein Specialfactor vor, der nicht prim ist
	{
	  relevanter_Faktor=special_factor;
	  TSpecialFactorRelation relation;
	  if (!relation.factor.DLP_get(y))
	    {
	      //cerr << "Hinweis: Special Factor " << y << " ungltig." << endl;
	      return; /* Treffer nicht bernehmen, da er ungltig ist. */
	    }
#ifndef IS_CLIENT
	  if (!SpecialFactorRelation_vermerken(relation.factor,this))
           relevanter_Faktor=groesster_Faktor_in_Relation();
	   // wenn Factor nicht vermerkt wurde, heit das, da er
           // neutralisiert werden konnte. -> normale Relation liegt vor.
#else
	  // an das Ausgabefile
	  //	  communication_stream << "Relation! ";
	  relation.fpos=speichern(communication_stream, relation.factor, HitCount);
#endif
	  return;
	}
    }

#ifdef IS_CLIENT
      // Ausgabe an File
      //      communication_stream << "Relation! ";
      speichern(communication_stream,1, HitCount);
#endif
      
  if (HitCount<4) Erfolgsaufteilung[HitCount]++; else Erfolgsaufteilung[4]++;
  // bei Erfolgsaufteilung[0]: Relation vollstndig durch statische FB entdeckt
  // bei Erfolgsaufteilung[1]: Relation durch Eliminierung eines Faktors aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[2]: Relation durch Eliminierung zweier Faktoren aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[3]: Relation durch Eliminierung dreier Faktoren aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[4]: Relation durch Eliminierung >3 Faktoren aus dyn. FB entdeckt
  
  //cout << "GL: " << *this << endl;
  //cout << "relevanter Faktor: " << relevanter_Faktor << endl;
  //cout << "Initialisierung der Gleichungszeile abgeschlossen!" << endl;
}

void CGleichungszeile::Ausgabe() const
{
  if (Relation_dense) 
    cout << Relation_dense->count(1) << " Primes (dense) ";
  else 
    for (int i=0; i<Relation_sparse->size(); i++) cout << Relation_sparse->v[i] << " ";
  
  cout << "<=> ";
  mpz_out_str(stdout,10,Delta);
  cout << endl;
}


ostream& operator<< (ostream& ostr, const CGleichungszeile& GL)
{
  if (GL.Relation_dense) ostr << "dense"; else ostr << "sparse";
  ostr << "1=";
  
  char *str;
  str = new char [mpz_sizeinbase(GL.Delta,10)+2]; mpz_get_str(str, 10, GL.Delta);
  ostr << str << "^2";
  delete [] str; 
  
  if (GL.Relation_dense)
    { 
      for (int i=GL.Relation_dense->first(); i>=0; i=GL.Relation_dense->next(i))
	ostr << "*p" << i;
    }
  else
    {
      for (int pos=0; pos<GL.Relation_sparse->size(); ++pos)
	ostr << "*p" << GL.Relation_sparse->v[pos];
    }
  
  return ostr;
}


void CGleichungszeile::convert_Relation_to_dense()
{
  //cout << "Konvertiere von Sparsedarstellung nach Densedarstellung." << endl;
  if (Relation_dense)
    { cerr << "Relation ist bereits dense!" << endl; return; }
  Relation_dense = new myBitString;
  for (int pos=Relation_sparse->size()-1; pos>=0; --pos)
    Relation_dense->set(Relation_sparse->v[pos]);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // Sparsedarstellung lschen
}

void CGleichungszeile::convert_Relation_to_sparse()
{
  //cout << "Konvertiere von Densedarstellung nach Sparsedarstellung." << endl;
  if (Relation_sparse)
    { cerr << "Relation ist bereits sparse!" << endl; return; }
  Relation_sparse = new CTinyFBsizetypeVector(Relation_dense->count(1));
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->fast_append(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // Densedarstellung lschen
}


bool CGleichungszeile::check_Relation(const Tfactor Restfaktor) const
{
  mpz_t h;
  mpz_init(h);
  mpz_mul(h,Delta,Delta); mpz_mod(h,h,n);
  
  if (Relation_dense)
    { 
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(h,h,Primzahlen[i]); else mpz_neg(h,h);
	  mpz_mod(h,h,n);
	}
    }
  else
    {
      for (int pos=0; pos<Relation_sparse->size(); ++pos)
	{
	  if (Relation_sparse->v[pos]>0) mpz_mul_ui(h,h,Primzahlen[Relation_sparse->v[pos]]);
	  else mpz_neg(h,h);
	  mpz_mod(h,h,n);
	}
    }
  

  mpz_t h2;
  mpz_init(h2); Restfaktor.in_mpz_ablegen(h2);
  //mpz_set_d(h2,Restfaktor);
  mpz_mul(h,h,h2); mpz_mod(h,h,n);
  mpz_clear(h2);
  
  if (mpz_cmp_ui(h,1)!=0)
    {
      cerr << "Relation mit Faktor " << setprecision(20) << Restfaktor
	   << " ist nicht korrekt: 1<>" << h << endl;
      mpz_clear(h);
      return false;  
    }
//  else   
//    {
//      cerr << "Relation mit Faktor " << setprecision(20) << Restfaktor
//           << " ist korrekt: 1=" << h << endl;
//    }

  mpz_clear(h);
  return true;  
}


void CGleichungszeile::kombinieren(const CGleichungszeile& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Kombiniere Gleichungszeilen..." << endl;
  //cout << "Gl.1: "; Kongruenzcheck();
  //cout << "Gl.2: "; GL2.Kongruenzcheck();
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen
      
      for (int i=GL2.Relation_dense->last(); i>=0; i=GL2.Relation_dense->prev(i))
	if (Relation_dense->test(i))
	  if (i>0)
	    {
	      mpz_mul_ui(x,x,Primzahlen[i]);
	      if (mpz_cmp(x,n)>=0)
		{
		  mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		  mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		}
	    } //else mpz_neg(Delta,Delta);
      
      (*Relation_dense).xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test(GL2.Relation_sparse->v[pos]))
	      if (GL2.Relation_sparse->v[pos]>0)
		{
		  mpz_mul_ui(x,x,Primzahlen[GL2.Relation_sparse->v[pos]]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	    Relation_dense->invert(GL2.Relation_sparse->v[pos]);
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(Primbasis_size);
	neueMenge.reset();
	while (p2<GL2.Relation_sparse->size() && p1<Relation_sparse->size())
	  {
	    while (GL2.Relation_sparse->v[p2]<Relation_sparse->v[p1])
	      {
		neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
	      }
	    while (Relation_sparse->v[p1]<GL2.Relation_sparse->v[p2])
	      {
		neueMenge.fast_append(Relation_sparse->v[p1]); ++p1;
		if (p1==Relation_sparse->size()) goto fertig;
	      }
	    while (GL2.Relation_sparse->v[p2]==Relation_sparse->v[p1])
	      {
		if (Relation_sparse->v[p1]>0)
		  {
		    mpz_mul_ui(x,x,Primzahlen[Relation_sparse->v[p1]]);
		    if (mpz_cmp(x,n)>=0)
		      {
			mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
			mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		      }
		  } //else mpz_neg(Delta,Delta);
		++p1; ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
		if (p1==Relation_sparse->size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=GL2.Relation_sparse->size()) { neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2; }
	while (p1!=Relation_sparse->size()) { neueMenge.fast_append(Relation_sparse->v[p1]); ++p1; }
	Relation_sparse->optisized_copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
        //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
      }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Gleichungszeile; diese also dazumultiplizieren!
  // Relationswurzeln beider Gleichungszeilen, also dazumultiplizieren:
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}

// Routinen, die ein mehrfaches Kombinieren beschleunigen:
// NEU: Multikombinationsroutinen sollten das Kombinieren einer
//      Relation mit vielen anderen Relationen beschleunigen, da
//      das Update der Faktorbasis-Exponenten der Relation
//      an das Ende aller Verknfungen geschoben werden kann und
//      nicht mehr simultan durchgefhrt werden mu.

static unsigned int Exponentenfeld[MAX_Primbasis_size] = {0};
static unsigned int multi_kombinieren_Zaehler = 0;
 
void CGleichungszeile::multi_kombinieren_init()
{
  // Initialisierung des Exponentenvektors
  // Hinweis: Der Aufruf dieser Routine darf auch entfallen,
  //          sofern sichergestellt ist, da der letzte
  //          multi_kombinieren_...-Aufruf ein  multi_kombinieren_exit() war.
  if (!multi_kombinieren_Zaehler) return;
  multi_kombinieren_Zaehler=0;
  for (int i=0; i<Primbasis_size; i++) Exponentenfeld[i]=0;
}

void CGleichungszeile::multi_kombinieren_exit()
{
  if (!multi_kombinieren_Zaehler) return; // gibt nichts zu tun...
  //cout << "Multikombinieren_exit nach " << multi_kombinieren_Zaehler
  //     << " Aufrufen." << endl;

  optisize(); // optimize Size of vector (sparse<->dense, etc.)

  // nun die Faktor-Exponenten in das Delta ziehen:
  mpz_set_ui(y,1);
  Exponentenfeld[0]=0; // Primzahl "-1" immer unbercksichtigt lassen
  for (int i=1; i<Primbasis_size; i++)
   if (Exponentenfeld[i])
     {
       if (Exponentenfeld[i]==1) mpz_mul_ui(y,y,Primzahlen[i]);
       else
        {
          mpz_set_ui(x,Primzahlen[i]);
          mpz_powm_ui(x,x,Exponentenfeld[i],n);
          mpz_mul(y,y,x);
        };
       if (mpz_cmp(y,n)>=0)
        {
         mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
         mpz_set_ui(y,1);
        }
       Exponentenfeld[i]=0; // Exponenten zurcksetzen...
     }
  mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
  multi_kombinieren_Zaehler=0;
}

void CGleichungszeile::multi_kombinieren_main(const CGleichungszeile& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Multi-Kombiniere Gleichungszeilen..." << endl;
  ++multi_kombinieren_Zaehler; // Anzahl Aufrufe zhlen

  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Exponentenfeld vermerken
  */

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen

      if (SizeOfRelation()<GL2.SizeOfRelation()) // fr minimalen Schleifenaufwand
       {
         for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
  	  if (GL2.Relation_dense->test(i)) ++Exponentenfeld[i];
       }
      else
       {
         for (int i=GL2.Relation_dense->first(); i>=0; i=GL2.Relation_dense->next(i))
	  if (Relation_dense->test(i)) ++Exponentenfeld[i];
       }
      (*Relation_dense).xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test(GL2.Relation_sparse->v[pos]))
              ++Exponentenfeld[GL2.Relation_sparse->v[pos]];
	    Relation_dense->invert(GL2.Relation_sparse->v[pos]);
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(Primbasis_size);
	neueMenge.reset();
	while (p2<GL2.Relation_sparse->size() && p1<Relation_sparse->size())
	  {
	    while (GL2.Relation_sparse->v[p2]<Relation_sparse->v[p1])
	      {
		neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
	      }
	    while (Relation_sparse->v[p1]<GL2.Relation_sparse->v[p2])
	      {
		neueMenge.fast_append(Relation_sparse->v[p1]); ++p1;
		if (p1==Relation_sparse->size()) goto fertig;
	      }
	    while (GL2.Relation_sparse->v[p2]==Relation_sparse->v[p1])
	      {
		++Exponentenfeld[Relation_sparse->v[p1]];
		++p1; ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
		if (p1==Relation_sparse->size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=GL2.Relation_sparse->size()) { neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2; }
	while (p1!=Relation_sparse->size()) { neueMenge.fast_append(Relation_sparse->v[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
        //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
      }
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Gleichungszeile; diese also dazumultiplizieren!
  // Relationswurzeln beider Gleichungszeilen, also dazumultiplizieren:
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  //-------------------------------------------------------------------
  // WICHTIG: Nach dem letzten Aufruf von multi_kombinieren_main
  // mu die Relation mit dem Exponentenfeld abgeglichen werden!!!
  // Hierzu ist ein abschlieendes multi_kombinieren_exit aufzurufen!
  //-------------------------------------------------------------------

  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}



Tfactor CGleichungszeile::kombinieren(istream &in, const streampos pos)
{
  streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  Tfactor f=kombinieren(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}

Tfactor CGleichungszeile::multi_kombinieren_main(istream &in, const streampos pos)
{
  streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  Tfactor f=multi_kombinieren_main(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}


int hole_Wert (istream &in)
{
  static int ahead_value = 0;
  if (ahead_value>0) { int w=ahead_value; ahead_value=0; return w; }

  char c;
  in >> c;
  if ( c>='A' && c<='Z' ) return c-('A'-1); // A=1,...,Z=26
  if ( c>='g' && c<='z' )
   {
     int h = (c-'g');
     ahead_value=(h%5)+1; // nchster "vorausgelesener" Wert
     return (h/5)+1; 
   }
  if (c=='0') return 0; // da keine fhrenden Nullen
  if (c=='$') return -1;
  if (c=='%') return -2;

  // ansonsten normal gespeicherte Zahl:
  int w;
  in.unget();
  in >> w;
  return w;
}

void schreibe_Wert (ostream &out, const int w)
{
  static int prev_value = 0;
  if (w<=5 && w>=1 && prev_value>0) // dann kann kodiert werden
    {
      out << char( (prev_value-1)*5 + w-1 + 'g' );
      prev_value=0;
      return;
    }

  if (prev_value>0)
   {
     out << char(prev_value+('A'-1)); // Wert ausgeben
     prev_value=0;
   }

  if (w>26) 
   {
     out << w << " "; // Wert unkodiert ausgeben
     return;
   }
  if (w>0)
   {
     if (w<=4) { prev_value=w; return; } // nchstes Mal abwarten
     out << char(w+('A'-1)); // Wert als char kodiert ausgeben
     return;
   }
  switch (w)
   {
     case  0: out << "0"; break;
     case -1: out << "$"; break;
     case -2: out << "%"; break;
     default:
      cerr << "unbekannter Wert gelesen!" << endl;
      exit(1);
   }
}

Tfactor CGleichungszeile::kombinieren(istream &in)
{
  //cout << "kombiniere Gleichungszeile vom File" << endl;
  Tfactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      cerr << "Error: Falsche Position beim Einlesen der Gleichungszeile" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // Thorsten: "x ist frei"
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=hole_Wert(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index))
	    if (index>0)
	      {
		mpz_mul_ui(x,x,Primzahlen[index]);
		if (mpz_cmp(x,n)>=0)
		  {
		    mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		    mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		  }
	      } //else mpz_neg(Delta,Delta);
	  Relation_dense->invert(index);
	  distance=hole_Wert(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(Primbasis_size);
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      
      distance=hole_Wert(in); index=distance; 
      while (distance>=0 && p1<Relation_sparse->size())
	{
	  while (index<Relation_sparse->v[p1])
	    {
	      neueMenge.append(index); distance=hole_Wert(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (Relation_sparse->v[p1]<index)
	    {
	      neueMenge.append(Relation_sparse->v[p1]); ++p1;
	      if (p1==Relation_sparse->size()) goto fertig;
	    }
	  while (index==Relation_sparse->v[p1])
	    {
	      if (index>0)
		{
		  mpz_mul_ui(x,x,Primzahlen[index]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	      distance=hole_Wert(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==Relation_sparse->size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=hole_Wert(in); index+=distance; }
      while (p1!=Relation_sparse->size()) { neueMenge.append(Relation_sparse->v[p1]); ++p1; }
      Relation_sparse->optisized_copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
      //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
    }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
 
  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      TDynamicFactorRelation FaRelSearch;
      // Korrekturfaktoren
      FaRelSearch.factor=hole_Wert(in);
      while (FaRelSearch.factor>0)
	{
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(FaRelSearch);
	  if (p==Dynamic_Factor_Relations.end())
	    {
	      cerr << "Unbekannter Korrekturfaktor " << FaRelSearch.factor << endl;
	      ret=0; return ret; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if ((*p).GLcached()) kombinieren( *((*p).GLcached()) ); // dynamische Relation im Hauptspeicher gecached
	  else kombinieren(DynamicRelations_from_file,(*p).fpos); // auf Hintergrundspeicher
	  //cout << "fertig kombiniert von " << FaRelSearch.factor << endl;
	  FaRelSearch.factor=hole_Wert(in);
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

Tfactor CGleichungszeile::multi_kombinieren_main(istream &in)
{
  //cout << "Multi-Kombiniere Gleichungszeilen vom File..." << endl;
  ++multi_kombinieren_Zaehler; // Anzahl Aufrufe zhlen

  Tfactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      cerr << "Error: Falsche Position beim Einlesen der Gleichungszeile" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // Thorsten: "x ist frei"
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=hole_Wert(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index)) ++Exponentenfeld[index];
	  Relation_dense->invert(index);
	  distance=hole_Wert(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(Primbasis_size);
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      
      distance=hole_Wert(in); index=distance; 
      while (distance>=0 && p1<Relation_sparse->size())
	{
	  while (index<Relation_sparse->v[p1])
	    {
	      neueMenge.append(index); distance=hole_Wert(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (Relation_sparse->v[p1]<index)
	    {
	      neueMenge.append(Relation_sparse->v[p1]); ++p1;
	      if (p1==Relation_sparse->size()) goto fertig;
	    }
	  while (index==Relation_sparse->v[p1])
	    {
	      ++Exponentenfeld[index];
	      distance=hole_Wert(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==Relation_sparse->size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=hole_Wert(in); index+=distance; }
      while (p1!=Relation_sparse->size()) { neueMenge.append(Relation_sparse->v[p1]); ++p1; }
      Relation_sparse->optisized_copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
      //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
    }

  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      TDynamicFactorRelation FaRelSearch;
      // Korrekturfaktoren
      FaRelSearch.factor=hole_Wert(in);
      while (FaRelSearch.factor>0)
	{
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(FaRelSearch);
	  if (p==Dynamic_Factor_Relations.end())
	    {
	      cerr << "Unbekannter Korrekturfaktor " << FaRelSearch.factor << endl;
	      ret=0; return ret; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if ((*p).GLcached()) multi_kombinieren_main( *((*p).GLcached()) ); // dynamische Relation im Hauptspeicher gecached
	  else multi_kombinieren_main(DynamicRelations_from_file,(*p).fpos); // auf Hintergrundspeicher
	  //cout << "fertig kombiniert von " << FaRelSearch.factor << endl;
	  FaRelSearch.factor=hole_Wert(in);
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

streampos CGleichungszeile::speichern(ostream &out, const Tfactor factor, const short int HitCount = 0) const
{
  //cout << "Speichern einer Gleichungszeile" << endl;
#ifndef IS_CLIENT
  //check_Relation(factor); // Test, ob die Kongruenz auch aufgeht (sollte sie nmlich!)
  /* Der Check ist nur erforderlich, wenn sichergestellt werden soll, da
     der Server nicht durch falsche Relationen defektiert werden kann.
     (Oder zu Debuggingzwecken ;)
     WICHTIG: Der Check klappt (wegen der Korrekturfaktoren) nicht im Clientmodus!
     Hinweis: Der Server kontrolliert Relationen aber bereits
              beim Empfang. -> hier nur noch zu Debuggingzwecken aktivieren.
  */
  
  streampos pos;
  // out.seekp(0,ios::end);  // sollte jetzt schon am Ende sein
  pos = out.tellp(); // fileposition ans Ende setzen und auslesen
#endif
  char *str;

  str = new char [mpz_sizeinbase(Delta,mpzbase_f)+2]; mpz_get_str(str,mpzbase_f, Delta);
  out << "G " << setprecision(20) << factor << " " << str << " ";
  delete [] str;

  // ab hier wird hexadezimal gearbeitet:
  ios::fmtflags oldFlags = out.flags(); // alte Formatflags retten
  out << hex; // Hexadezimalausgabe
  out.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /* Bemerkung:
     Statt der Werte werden die Abstnde gespeichert, das spart Platz */
  if (Relation_dense)
    { 
      int prev = 0;
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
          schreibe_Wert(out,i-prev);
	  prev=i;
	} 
    }
  else
    {
      int prev = 0;
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
 	  int h = Relation_sparse->v[i];
          schreibe_Wert(out,h-prev);
	  prev=h;
	}
    }

#ifndef IS_CLIENT
  schreibe_Wert(out,-1); out << endl;
  out.flags(oldFlags); // alte Formatflags wiederherstellen
  return pos;
#else
  int Anzahl_echte_Faktoren = 0;
  for (int i=0; i<HitCount; i++) if (Hits[i].Exponent>0) Anzahl_echte_Faktoren++;
  if (Anzahl_echte_Faktoren>0)
    {
      schreibe_Wert(out,-2);
      //cout << "Korrekturfaktoren fr Relation mit Faktor " << setprecision(20) << factor << endl;
      for (int i=0; i<HitCount; i++)
	{ 
	  //cout << Hits[i].Faktor << "^" << Hits[i].Exponent << endl; 
	  if (Hits[i].Exponent>0) schreibe_Wert(out,Hits[i].Faktor);
	}
    }
  schreibe_Wert(out,-1); out << endl;
  out.flags(oldFlags); // alte Formatflags wiederherstellen
  return 0;
#endif
}

void CGleichungszeile::Kongruenzcheck() const
{
  mpz_set_ui(y,1); 
  if (Relation_sparse)
    {
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  if (Relation_sparse->v[i]>0) 
	    mpz_mul_ui(y,y,Primzahlen[Relation_sparse->v[i]]);
	  else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  else
    {
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(y,y,Primzahlen[i]); else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  
  mpz_invert(x,Delta,n); mpz_powm_ui(x,x,2,n);
  
  cout << "Kongruenzcheck: ";
  mpz_out_str(stdout,10,x); printf(" = ");
  mpz_out_str(stdout,10,y); printf(" (mod N)\n");
  if (mpz_cmp(x,y)!=0) 
    {
      cout << "Fehler! Werte sind NICHT kongruent!" << endl;
      exit(1);
    }
  
}


void CGleichungszeile::Quadratische_Kongruenz_ausrechnen() const
{
  /* Die ermittelte Gleichungszeile erfllt die Kongruenz (A=1) mod n.
     Wegen A-1=(A+1)*(A-1)=0 mod n knnen wir daher nun durch ggT(A+1,n)
     einen Teiler von n ermitteln und haben gute Chancen, da es sich dabei
     auch um einen echten (also nichttrivialen) Teiler handelt... */

  cout << endl << "Geeignete Relation ermittelt..." << endl;

  if (!Relation_ist_leer()) cout << "Fehler! Relation ist noch nicht leer!" << endl;

  cout << Delta << " = 1 (mod N)" << endl;
   
  mpz_powm_ui(x,Delta,2,n);
  printf("d.h.: ");
  cout << x << " = 1 (mod N)" << endl;
  
  if (mpz_cmp_ui(x,1)!=0) 
    {
      cout << "Fehler! Quadrate sind NICHT kongruent!" << endl;
      exit(1);
    }
  
  mpz_add_ui(x,Delta,1);
  mpz_gcd(x,x,n); mpz_div(y,n,x); /* beachte: n und nicht kN
				     ist die eigentlich zu
				     faktorisierende Zahl! */
  
  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
      cout << "Faktorisierung war leider trivial..." << endl;
    }
  else
    {
      cout << endl << "Faktorisierung gefunden!" << endl;
      cout << x << " * ";
      cout << y << endl;
      cout << "= ";
      cout << n << endl;
      
      if (mpz_probab_prime_p(x,probab_prime_checks))
	{
	  mpz_div(n,n,x);
	  cout << x << " ist Faktor." << endl;
	  Factorization_to_file << MAL() << x << flush;
	}
      else
      	if (Potenztest(x)) mpz_div(n,n,x); // ggf. Potenzen streichen...

      if (mpz_probab_prime_p(y,probab_prime_checks))
	{
	  mpz_div(n,n,y);
	  cout << y << " ist Faktor." << endl;
	  Factorization_to_file << MAL() << y << flush;
	}
      else
	if (Potenztest(y)) mpz_div(n,n,y); // ggf. Potenzen streichen...
      

      if (mpz_cmp_ui(n,1)==0) // Faktorisierung vollstndig?
	{ 
	  cout << "Faktorisierung erfolgreich beendet." << endl;
          Factorization_to_file << endl;
	  
	  // Streams schlieen
          SpecialRelations_buffer.close();
	  DynamicRelations_buffer.close();
	  StaticRelations_buffer.close();
	  Recovery_buffer.close();

	  // Files lschen
	  remove(SpecialRelationsFile);
	  remove(DynamicRelationsFile);
	  remove(StaticRelationsFile);	
	  remove(RecoveryFile);
	  
	  exit(0); // Programm erfolgreich beenden...
	}
    }
  cout << "Setze Faktorisierung fort..." << endl;
}
  
unsigned int CGleichungszeile::SizeOfRelation_verknuepft_mit(const CGleichungszeile& GL2) const
{
  int Zaehler = 0;
  
  if (Relation_dense)
    if (GL2.Relation_dense)
      {
	for (int i=GL2.Relation_dense->first(); i>=0; i=GL2.Relation_dense->next(i))
	  if (Relation_dense->test(i)) Zaehler++;
      }
    else 
      {
        for (int i=0; i<GL2.Relation_sparse->size(); ++i)
	  if (Relation_dense->test(GL2.Relation_sparse->v[i])) Zaehler++;
      }
  else // Relation is sparse
    if (GL2.Relation_dense)
      {
	for (int i=0; i<Relation_sparse->size(); ++i)
	  if (GL2.Relation_dense->test(Relation_sparse->v[i])) Zaehler++;
      }
    else
      {
	int j=0;
	for (int i=0; i<GL2.Relation_sparse->size(); ++i)
	  {
	    while (j<Relation_sparse->size()
		   && Relation_sparse->v[j]<GL2.Relation_sparse->v[i]) ++j;
	    if (j>=Relation_sparse->size()) break;
	    if (Relation_sparse->v[j]==GL2.Relation_sparse->v[i]) {j++; Zaehler++;}
	  }
      }
  
  return SizeOfRelation()+GL2.SizeOfRelation()-2*Zaehler;
}


void CGleichungszeile::Relation_optimieren(int bitpos=1)
{ 
  /* optimiere ab bitpos:
     bitpos = 0 -> letzter Faktor der Relation (Vorsicht: wenn dieselbe
                   Relation gerade eingefgt wurde, ergibt eine erneute
                   Kombination eine leere Relation!
     bitpos = 1 -> vorletzter Faktor
     bitpos = 2 -> vorvorletzter Faktor, usw.
  */
  cout << "Optimiere " << groesster_Faktor_in_Relation() << endl;

  multi_kombinieren_init();
  while (true)
    {
      signed int vorletzter_Faktor;
      if (Relation_sparse) 
	if (Relation_sparse->size()>bitpos)
	  vorletzter_Faktor=Relation_sparse->v[Relation_sparse->size()-1-bitpos];
	else break;
      else
	{
	  vorletzter_Faktor=Relation_dense->last(1);
	  for (int j=0; j<bitpos && vorletzter_Faktor>=0; j++)
	    vorletzter_Faktor=Relation_dense->prev(vorletzter_Faktor,1);
	  if (vorletzter_Faktor<0) break;
	}
      if (GLS[vorletzter_Faktor]==NULL) { bitpos++; continue; }
      cout << "Optimiere " << groesster_Faktor_in_Relation()
      	   << " mit " << vorletzter_Faktor << endl;
      multi_kombinieren_main(*GLS[vorletzter_Faktor]); 
    }
  multi_kombinieren_exit();
}


void Gleichungssystem_optimieren()
{
  /* Mglichst viele redundante Faktoren aus den Relationen entfernen.
     Dadurch dafr sorgen, da die Relationen mglichst dnn mit Faktoren
     besetzt sind.
     -> Nachteil: bei dicht besetztem Gleichungssystem hohe Laufzeit! */
  cout << "Optimiere Gleichungssystem" << endl;
  for (int i=1; i<Primbasis_size; i++) if (GLS[i]) 
    {
      cout << " GL " << i << ", " << flush;
      GLS[i]->Relation_optimieren();
    }
  cout << endl;
}


#include <time.h>
time_t StartZeitpunkt; // Zeitpunkt, zu dem das Sieben gestartet wurde
double StartFuellung; // Fllung des GLS bei Start des Aussiebens

void Restzeit_kalkulieren(void)
{
  // eigentlich wird hier weniger die Restzeit abgeschtzt als
  // vielmehr die durchschnittlich bentigte Zeit zum Sieben einer
  // Relation ermittelt...
  // Fr eine Restzeitabschtzung mte man das Zeitverhalten eigentlich
  // extrapolieren (und zustzlich die Schwankungen zwischen den
  // "Zwischenankunftszeiten" der Relationen ausmitteln).
  // Dies wre aber ein hoffnungsloses Unterfangen, da die Systemlast
  // schwankt und (im Serverbetrieb) die Anzahl der siebenden Clients
  // unbekannt ist (und zudem auch die Fileclients die Ergebnisse
  // verflschen wrden).
  if (Fuellung_GLS==StartFuellung) return;
  double t=difftime(time(NULL),StartZeitpunkt);
  t/=(Fuellung_GLS-StartFuellung);
  double Restzeit=(Primbasis_size-Fuellung_GLS)*t;
  double days = floor(Restzeit/(60*60*24));
  int hours,mins,secs;
  Restzeit=fmod(Restzeit,60*60*24);
  secs=static_cast<int>(Restzeit); mins=secs/60; hours=mins/60;
  mins%=60; secs%=60;
  if (days)
   cout << "geschtzte Restzeit: " << days << "d " << hours << "h" << endl;
  else
   cout << "geschtzte Restzeit: " << hours << "h " << mins << "m "
        << secs << "s" << endl;
}

void Statuslegende_ausgeben()
{
  cout << "Bedeutung der nachfolgenden Statusinformationen:" << endl;
#ifndef USE_NETWORK 
  cout << "#Relationen{Verteilung}/Faktorbasis,#dynamic-FB,#Special-Factors(Verteilung)"
       << " [Siebpos] -> Fortschritt" << endl;
#else
  cout << "#Relationen{Verteilung}/Faktorbasis,#dynamic-FB,#Special-Factors(Verteilung)"
       << " -> Fortschritt" << endl;
#ifndef IS_CLIENT
 cout << "Hinweis: Verteilungen = vom Server vorgenommene Verteilungen." << endl;
 cout << "         Extern vorgenomme Verteilungen bleiben unbercksichtigt." << endl;
#endif
#endif
 StartZeitpunkt=time(NULL); StartFuellung=Fuellung_GLS; // Initialisierung fr Restzeitberechnung
}

void Statusmeldung()
{
  // aktuellen Stand ausgeben  
  cout << Fuellung_GLS
       << "{" << Erfolgsaufteilung[0]
       << "," << Erfolgsaufteilung[1]
       << "," << Erfolgsaufteilung[2]
       << "," << Erfolgsaufteilung[3]
       << "," << Erfolgsaufteilung[4]
       << "}/" << Primbasis_size << ","
       << Dynamic_Factor_Relations.size() << ","
       << Special_Factor_Relations.size()
       << "(" << Special_to_dynamic_Factor_hit << "," << Special_hit
       << "," << gefundene_DLP_Zyklen << ")";
#ifndef USE_NETWORK
  // Sieboffset macht nur bei standalone-Version Sinn
  cout << " [" << Sieboffset << "]";
#endif
  cout.setf(ios::left, ios::adjustfield); 	
  cout << " -> ca. " << setw(6) << setprecision(5)
       << 100.0*Fuellung_GLS/Primbasis_size
       << "%  " << endl;
}


void Relation_einfuegen(CGleichungszeile *GL, const bool do_multi_kombinieren_init=true)
{ /* hier wird eine neu gefundene Relation in das bisherige Gleichungssystem
     eingefgt. Falls sie redundant ist bzw. zu einer Faktorisierung fhrt,
     so wird dies automatisch erkannt. */
#ifdef IS_CLIENT
  if (GL->relevanter_Faktor >= 0) Fuellung_GLS++; // beim Client mangels Gleichungssystem: jede gefundene Relation
  delete GL;
  return;
#else  
  if (GL->relevanter_Faktor==dynamic_factor || GL->relevanter_Faktor==special_factor)
    { delete GL; return; } /* diese Relationen sind auf Hintergrundspeicher ausgelagert
			      und brauchen deshalb nicht im Hauptspeicher zu verbleiben */
  
  if (GL->relevanter_Faktor==cached_dynamic_factor)
    return; /* Diese Faktoren sind zwar auch auf dem Hintergrundspeicher, verbleiben aber
	       zwecks schnelleren Zugriffs auch im Hauptspeicher */
  
  if (do_multi_kombinieren_init) GL->multi_kombinieren_init();
  while (!GL->Relation_ist_leer())
    {
      //clog << "kombiniere " << GL.relevanter_Faktor;
      if (GL->relevanter_Faktor>=Primbasis_size) cerr << "zu grosser Wert!!" << endl;
      CGleichungszeile* pos = GLS[GL->relevanter_Faktor];
      //clog << " ... ";
      if ( pos!=NULL )
	{ /* Relation mit diesem Eintrag existiert schon -> neue Kombination mglich! */
	  //clog << "kombiniere..." << endl;
#if 1 /* Minimierung der Relationen ein-/ausschalten */
          if (
	       (GL->Relation_sparse) && (pos->Relation_sparse) // beide in Sparse-Darstellung
               && GL->SizeOfRelation()<=pos->SizeOfRelation()
	       &&GL->zweitgroesster_Faktor_in_Relation()<=pos->zweitgroesster_Faktor_in_Relation()
             )
	    { /* weniger Primzahlen in der Relation bedeuten schnelleres Abarbeiten.
	  	 Da beide Gleichungszeilen quivalent sind, knnen sie problemlos so
		 getauscht werden, da die platzsparende im Gleichungssystem verbleibt.
		 Andererseits bestimmt das Maximum der zweitgrten Faktoren zweier
		 Relationen die nchste Verknpfung. Also ist auch dies zu minimieren.
                 Beim Zielkonflikt wren beide Kriterien also abzuwgen. Dies geschieht
		 hier nicht. Nur wenn beide Kriterien zugleich die Lage nicht
                 verschlechtern, werden die Relationen getauscht.
                 Man beachte, da der Tausch das Ergebnis der aktuellen Verknpfung
                 nicht verndert, sondern nur zuknftige Verknpfungen
	         beschleunigen kann bzw. den Speicherbedarf vermindert.
              */
	      //clog << "Optimiere Relationen" << endl;
	      GL->multi_kombinieren_exit(); 
	      GLS[GL->relevanter_Faktor]=GL;
	      register CGleichungszeile* h = pos; pos=GL; GL=h; // Gleichungszeilen tauschen
              // GL->multi_kombinieren_init(); <- darf entfallen
	    }
#endif
          // grter Faktor mu eliminiert werden:
 	  //GL->Ausgabe(); (*pos).Ausgabe();
 	  GL->multi_kombinieren_main(*pos); 
	}
      else
	{
	  GL->multi_kombinieren_exit();
	  GLS[GL->relevanter_Faktor]=GL; // Gleichungszeile ins Gleichungssystem eintragen
	  Fuellung_GLS++; // eine Relation mehr im Gleichungssystem
          //if (Fuellung_GLS+2*GL->SizeOfRelation()>Primbasis_size) GL->Relation_optimieren();
          //if (GL->Relation_dense) GL->Relation_optimieren();
	  // speichern
          GL->speichern(StaticRelations_to_file);
          Statusmeldung();
          Restzeit_kalkulieren();
	  return;
	}
      //clog << "ok." << endl;
    }

  GL->multi_kombinieren_exit();
  //clog << "Kombinierung abgeschlossen." << endl;

  /* Relation redundant	-> Faktorisierung probieren! */
  //cout << endl << "leere Relation entdeckt." << endl;
  //GL->Ausgabe();
  Statusmeldung();
  GL->Quadratische_Kongruenz_ausrechnen();
  //     GL->clear(); delete GL; // Gleichungszeile leeren
  delete GL;
#endif
}



TSiebElement log_Threshold = 0; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert

void SiebSchwellwert_kalkulieren()
{
  // Hier wird der Threshold (berschlagsmig) ermittelt
  
  // Voraussetungen zum Aufruf der Initialisierung:
  // 1. Logische Siebintervallgre M (fr Intervall [-M,M])
  // 2. kN (die mit Vorfaktor versehene zu faktorisierende Zahl)
  // 3. die statische Faktorbasis
  // mssen bereits bekannt und berechnet sein.
 
  // logarithmischen Schwellwert kalkulieren:
  mpz_div_ui(x,kN,2); mpz_sqrt(x,kN); mpz_mul_ui(x,x,M);
  double d = log(fabs(mpz_get_d(x))) - (Factor_Threshold*log(biggest_Prime_in_Factorbase));
  if (Primzahlen[1]==2) d-=log(2); // Falls 2 in der Faktorbasis, dann teilt 2 alle Siebeintrge
  d*=log_Multiplikator;
  cout << "Der Threshold zum Erkennen einer Relation betrgt "<< d << endl;
  log_Threshold=static_cast<TSiebElement>(floor(d));
  if (log_Threshold!=floor(d)) 
   {
     cerr << "Mit den momentan einkompilierten Typen/Konstanten kann diese" << endl;
     cerr << "Faktorisierung leider nicht durchgefhrt werden." << endl; 
     cerr << "TSiebElement ist zu klein fr Threshold!" << endl;
     cerr << "Bitte modifizieren Sie log_Multiplikator und/oder TSiebElement" << endl;
     cerr << "und kompilieren Sie dann das Programm erneut." << endl;
     exit(1);
   }
}

void Sieb_initialisieren()
{
  //cout << "Starte Initialisierung des Siebs von " << Sieboffset
  //   << " bis " << Sieboffset+Siebgroesse << endl;

  register TSiebElement w = log_Threshold;
  /* log_Threshold bezeichnet den Schwellwert, der zum Erkennen der Relation
     unterschritten werden mu. Es ist besser, den Threshold bereits bei der
     Initialisierung zu bercksichtigen (d.h. abzuziehen), da beim Auslesen
     des Siebs dann nur auf negative Werte getestet werden braucht. */
  
  // Initialisierung des Siebs mit dem berschlagsmigen Schwellwert
  register int i=Siebgroesse; do Sieb[--i]=w; while (i);

  // nachfolgender Block dient (falls aktiviert) nur noch Validierungszwecken:
  /*
  Polynom.Werte_holen(Sieboffset,x,y);
  TSiebElement w_test = static_cast<TSiebElement>(log_Multiplikator*log(fabs(mpz_get_d(y))));
  TSiebElement FB_Threshold
    = static_cast<TSiebElement>(log_Multiplikator*Factor_Threshold*log(biggest_Prime_in_Factorbase));
  w_test-=FB_Threshold;
  if (Primzahlen[1]==2) w_test-=static_cast<TSiebElement>(log_Multiplikator*log(2));
  cout << "value " << y << " -> logwert: " << w_test << "<->" << log_Threshold << endl; 
  */

  //cout << "Initialisierung abgeschlossen." << endl;
}

#ifdef ASM386
#define SIEBASM386
#if TSIEBELEMENTSIZE == 1
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subb %4,(%2,%1)\n\t" \
       "subb %4,(%2,%0)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subb %4,(%2,%1)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (Sieb), "g" (P), "q" (lp), "0" (d0), "1" (d1));
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (Sieb), "g" (P), "q" (lp));
  // Bemerkung: "memory" ist eigentlich zuviel, da nur Sieb[...] verndert wird;
  // nur "asm volatile" ist besser (wenn der Compiler das mitmacht).
  // Am besten wre es, wenn man wte, wie man als contraint angeben
  // knnte, da Sieb[...] verndert wurde.
#elif TSIEBELEMENTSIZE == 2
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subw %4,(%2,%1,2)\n\t" \
       "subw %4,(%2,%0,2)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subw %4,(%2,%1,2)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (Sieb), "g" (P), "q" (lp), "0" (d0), "1" (d1));
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (Sieb), "g" (P), "q" (lp));
#else
#undef SIEBASM386
#warning "Kein handoptimierter Asssemblerkode fr Aussieben mit TSiebElement!"
#endif
#endif

void schnelles_aussieben()
{
  //cout << "Aussieben mit Primzahlen der statischen FB: " << Sieboffset<< endl;
  
  /* wir sieben erst ab einer greren Primzahl aus der Faktorbasis,
     da wir sowieso unsauber und schnell sieben...
     ( je kleiner die Primzahl, um langsamer ist das Sieben mit dieser
     Zahl; und bei der gewhlten Implementierung ohne Bercksichtigung
     von Primzahlpotenzen ist auch der zustzlich gemachte Fehler
     vernachlssigbar und wird zudem ber Factor_Threshold kompensiert.
     Aber Vorsicht: Shortcut-Konstante nicht zu hoch ansetzen, sonst fallen
     zu viele Relationen raus!
     Achtung: als Startwert mindestens 2 whlen, damit -1 und 2 ignoriert
     werden knnen. Dann gibt es auch 2 Deltawerte. )
     -> relevant fr Schleife mit den kleinen Primzahlen
  */

  int nr=Primbasis_size-1;

  // groe Primzahlen (maximal 1 Siebtreffer pro Delta) 
  for (;; --nr) 
    {
      register int Primzahl=Primzahlen[nr];

      if (Primzahl<Siebgroesse) break;

      // neue d0/d1-Ermittlung (sollte schneller sein)
      register int d0,d1;
      {
	register int h=Sieboffset%Primzahl;
	if (h<0) h+=Primzahl; // damned signed %-Operation...
	d0=Delta_der_Primzahlen[nr][0]-h;
	d1=Delta_der_Primzahlen[nr][1]-h;
	if (d1<0) d1+=Primzahl;
	if (d0<0) d0+=Primzahl;
      }
      
      register TSiebElement log_Primzahl = log_Primzahlen[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSiebElement) (log_Multiplikator*log(Primzahl)) */
      
      if (d0<Siebgroesse) Sieb[d0]-=log_Primzahl;
      if (d1<Siebgroesse) Sieb[d1]-=log_Primzahl;
    }

  // kleine Primzahlen (mehrere Siebtreffer pro Delta)
  for (; nr>=sieben_ab_Primzahlindex; --nr) 
    {
      register int Primzahl = Primzahlen[nr];
      //cout << "Quadratisches Sieb mit " << Primzahl << endl;
      
//#define ALTE_D_ERMITTLUNG
#ifdef ALTE_D_ERMITTLUNG
      register int d0=(Delta_der_Primzahlen[nr][0]-(Sieboffset+Siebgroesse))%Primzahl;
      register int d1=(Delta_der_Primzahlen[nr][1]-(Sieboffset+Siebgroesse))%Primzahl;
      if (d1>=0) d1-=Primzahl;
      if (d0>=0) d0-=Primzahl;
#else
      // neue d0/d1-Ermittlung (sollte schneller sein)
      register int d0,d1;
      {
	register int h=(Sieboffset+Siebgroesse)%Primzahl;
	if (h<0) h+=Primzahl; // damned signed %-Operation...
	d0=Delta_der_Primzahlen[nr][0]-h;
	d1=Delta_der_Primzahlen[nr][1]-h;
	if (d1>=0) d1-=Primzahl;
	if (d0>=0) d0-=Primzahl;
      }
#endif
      
      register TSiebElement log_Primzahl = log_Primzahlen[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSiebElement) (log_Multiplikator*log(Primzahl)) */
      
      d0+=Siebgroesse;
      d1+=Siebgroesse;

      //if (d0==d1) cout << Primzahl << " teilt Vorfaktor!" << endl;
      
#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, Primzahl);
#else 
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so da d0<d1
      while (d0>=0)
	{
	  Sieb[d0]-=log_Primzahl; d0-=Primzahl;
	  Sieb[d1]-=log_Primzahl; d1-=Primzahl;
	}
      if (d1>=0) Sieb[d1]-=log_Primzahl;
#endif
    }

  // hhere Primzahlpotenzen
  for (nr=0; nr<anz_weitere_Primzahlpotenzen; nr++)
    {
      register int PPotenz = Primzahlpotenzen[nr];
      
#ifdef ALTE_D_ERMITTLUNG
      register int d0=(Delta_der_Primzahlpotenzen[nr][0]-(Sieboffset+Siebgroesse))%PPotenz;
      register int d1=(Delta_der_Primzahlpotenzen[nr][1]-(Sieboffset+Siebgroesse))%PPotenz;
      if (d1>=0) d1-=PPotenz;
      if (d0>=0) d0-=PPotenz;
#else
      // neue d0/d1-Ermittlung (sollte schneller sein)
      register int d0,d1;
      {
	register int h=(Sieboffset+Siebgroesse)%PPotenz;
	if (h<0) h+=PPotenz; // damned signed %-Operation...
	d0=Delta_der_Primzahlpotenzen[nr][0]-h;
	d1=Delta_der_Primzahlpotenzen[nr][1]-h;
	if (d1>=0) d1-=PPotenz;
	if (d0>=0) d0-=PPotenz;
      }
#endif
      
      register TSiebElement log_Primzahl = log_Primzahl_der_Primzahlpotenzen[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSiebElement) (log_Multiplikator*log(Primzahl)) */
      
      // beachte: da mit den Primzahlen selbst auch gesiebt wird, darf nicht log(Primzahlpotenz), sondern
      // nur log(Primzahl) abgezogen werden!
      // Analogie: Wenn man schon durch p geteilt hat, mu man nocheinmal durch p teilen,
      //           um insgesamt durch p^2 zu teilen!

      d0+=Siebgroesse;
      d1+=Siebgroesse;

#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, PPotenz);
#else
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so da d0<d1
      while (d0>=0)
	{
	  Sieb[d0]-=log_Primzahl; d0-=PPotenz;
	  Sieb[d1]-=log_Primzahl; d1-=PPotenz;
	}
      if (d1>=0) Sieb[d1]-=log_Primzahl;      
#endif
    }

}


void aussieben_mit_Quadraten()
{
  // Quadrate von statischen (und ggf. dynamischen Faktoren) sieben.
  // Wichtig: Daher nur den einfachen Logarithmus des Faktors subtrahieren!
  // Bemerkung: Zur Effizienzsteigerung wird davon ausgegangen, da jedes Quadrat nur einmal
  //            innerhalb des gesamten Siebintervalls trifft!

  //cout << "Quadrate sieben Start" << endl;
  if (Sieb_Delta_QuadrateHeap.empty()) return; // es gibt nichts zu sieben...
  TSieb_Delta sieb_delta = Sieb_Delta_QuadrateHeap.top();
  register int delta = sieb_delta.delta-Sieboffset;
  
  while (delta<Siebgroesse)
    {
      Sieb_Delta_QuadrateHeap.pop(); // Element aus der Priorittswarteschlange nehmen
      Sieb[delta]-=static_cast<TSiebElement>(ceil(log_Multiplikator*log(sieb_delta.factor))); /* Teiler im Sieb vermerken */
      if (Sieb_Delta_QuadrateHeap.empty()) break;
      sieb_delta = Sieb_Delta_QuadrateHeap.top();
      delta = sieb_delta.delta-Sieboffset;
    }
  //cout << "Quadrate sieben Ende" << endl;
}


void aussieben_mit_Dynamic_Factoren()
{
  if (Sieb_Delta_Heap.empty()) return; // es gibt nichts zu sieben...
  short int HitCount=0;
  TSieb_Delta sieb_delta = Sieb_Delta_Heap.top();
  register int delta = sieb_delta.delta-Sieboffset;
  goto Schleifentest;
  
 Schleifenanfang_weiterer_Faktor:
  ++HitCount;
  //if (HitCount>=MaxHits) { cerr << "MaxHits erhhen!"; exit(1); } 
 Schleifenanfang_normal:
  Sieb_Delta_Heap.pop(); // Element aus der Priorittswarteschlange nehmen
  Sieb[delta]-=static_cast<TSiebElement>(ceil(log_Multiplikator*log(sieb_delta.factor))); /* Teiler im Sieb vermerken */
  
  // Da dynamische Faktoren selten auftreten, diese Faktoren schon whrend
  // des Siebens vermerken...
  // ...das erhht die Effizienz bei der Refaktorisierung...    
  Hits[HitCount].Faktor=sieb_delta.factor;
  
  sieb_delta.delta += sieb_delta.factor; // nchste Hit-Position         	      
  if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
    Sieb_Delta_Heap.push(sieb_delta); // wieder in Priorittswarteschlange eintragen
  else
    if (Sieb_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben? 
  
  sieb_delta = Sieb_Delta_Heap.top(); // nchsten Eintrag aus Priorittswarteschlange holen
  register int delta2; delta2 = sieb_delta.delta-Sieboffset;
  if (delta==delta2) goto Schleifenanfang_weiterer_Faktor;  
  if (Sieb[delta]<0) // Relation gefunden?
    {
      Sieb[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      Relation_einfuegen(new CGleichungszeile(delta+Sieboffset, HitCount+1)); /* dann einfgen */
    }
  delta=delta2;
  
  HitCount=0; // und auf ein neues
 Schleifentest:
  if (delta<Siebgroesse) goto Schleifenanfang_normal; // noch im aktuellen (physikalischen) Intervall?
  return; // fertig mit diesem Intervall
  
 Schlange_leer:
  if (Sieb[delta]<0) // Relation gefunden?
    {
      Sieb[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      Relation_einfuegen(new CGleichungszeile(delta+Sieboffset,HitCount+1)); // dann einfgen
    }
  return; 
}


void schnelles_auslesen_des_Siebs()
{
  //cout << "Auslesen des (gesiebten) Siebintervalls" << endl;
  register int d=Siebgroesse;
  do
    if (Sieb[--d]<0) // Relation gefunden?
      Relation_einfuegen(new CGleichungszeile(Sieboffset+d)); /* dann einfgen */
  while (d);
}



/* ---------- Vorfaktorbestimmung ------------ */

// Um verschiedene Vorfaktorbestimmungsmethoden (mit jeweiliger Neu-Compilierung)
// testen zu knnen, ist die Prozedur " void bestes_kN_ermitteln() "
// auf das File "Vorfaktorbestimmung.cc" ausgelagert.

#include "Vorfaktorbestimmung.cc"

/* ------------------------------------------- */



void CPolynom::erstes_Polynom_ermitteln(const mpz_t &fuer_kN, const int M)
{
  
  mpz_set(kN,fuer_kN);
  mpz_div_ui(kN_div2,kN,2); // kN/2 fr Berechnung negativer Modulowerte
  
  //mpz_t rkN; mpz_init(rkN); mpz_sqrt(rkN,kN);
  //const double W1=sqrt(2)/2;
  //const double W2=-1/(2*sqrt(2));
  //double A0=W1*mpz_get_d(rkN)/M;
  //double B0=0;
  //double C0=W2*M*mpz_get_d(rkN);
  //mpz_clear(rkN);
  //mpz_set_d(D,sqrt(A0)); // Initialwert fr Polynome

  // exakter:
  mpz_mul_ui(D,kN,2); mpz_sqrt(D,D); mpz_div_ui(D,D,2); mpz_div_ui(D,D,M); mpz_sqrt(D,D);
  // D ist der Initalwert fr das "optimale" Polynom
  
  // Sonderfall:
  //  Falls D (bei kleinen Zahlen) in der statischen Faktorbasis
  //  liegt, gibt es Probleme. Ebenso, falls D dynamischer Faktor ist.
  if (mpz_cmp_ui(D,Specialfactor_Sieb_Threshold)<=0)
   {
     if (Factor_Threshold<=1.0)
      {
        collecting_phase_beendet = true; // keine dynamischen Faktoren sammeln!
        cout << "Dynamische Faktoren deaktiviert." << endl;
      }
     else
      cout << "Polynom kollidiert eventuell mit dynamischer Faktorbasis..." << endl;
   }
  if (mpz_cmp_ui(D,biggest_Prime_in_Factorbase)<=0)
   {
     cout << "Polynom kollidiert mit statischer Faktorbasis... korrigiere dies..." << endl;
     mpz_set_ui(D,biggest_Prime_in_Factorbase+1);
     collecting_phase_beendet = true; // keine dynamischen Faktoren sammeln!
   }

  // nun zur eigentlichen Polynomberechnung:
  // Let A ~ D^2, D (probab. prime), (D/kN)=1, D=3 (mod 4)
  // fr D=3 (mod 4) geht die Wurzelermittlung einfacher,
  // es reicht aber D=1 (mod 2)
  // meine Implementation lt alle Flle zu, da fr D=5 (mod 8)
  // die Wurzelberechnung ebenfalls recht einfach zu implementieren ist
  // und fr D=1 (mod 4 bzw. 8) die Ermittlung mit Lucassequenzen mglich ist...

  while (mpz_mod_ui(x,D,2)!=1) mpz_add_ui(D,D,1);

  naechstes_Polynom_ermitteln(); // hier folgt die weitere Berechnung
}

void CPolynom::naechstes_Polynom_ermitteln(const int step=0)
{
  mpz_add_ui(D,D,step*4); // step dient nur dazu, die Polynomintervalle
                          // fr die clients verteilen zu knnen

  // Let A ~ D^2, D (probab. prime), (D/kN)=1, D=3 (mod 4)
  // fr D=3 (mod 4) geht die Wurzelermittlung einfacher,
  // es reicht aber D=1 (mod 2)
  // meine Implementation lt alle Flle zu, da fr D=5 (mod 8)
  // die Wurzelberechnung ebenfalls recht einfach zu implementieren ist
  // und fr D=1 (mod 4 bzw. 8) die Ermittlung mit Lucassequenzen mglich ist...

  do mpz_add_ui(D,D,2);
  while ( (!mpz_probab_prime_p(D,probab_prime_checks)) || (mpz_jacobi(D,kN)!=1) );
  mpz_mul(A,D,D); mpz_mul_ui(A2,A,2);

  if (mpz_mod_ui(x,D,4)==1) // D=1 (mod 4)
   {
     // h1 = sqrt(kN) (mod D)
     mpz_sqrtmod(h1,kN,D);

     // h0 = h1/kN (mod D)
     mpz_invert(h0,kN,D); mpz_mul(h0,h0,h1); mpz_mod(h0,h0,D);
   }
  else // D=3 (mod 4)
   {
     // Bemerkung: dieser Fall knnte durch obigen Fall ebenfalls
     // bernommen werden, aber die Berechnung gestaltet sich
     // fr D=3 (mod 4) noch einfacher (Invertieren nicht erforderlich) 

     // h0 = (kN)^((D-3)/4) mod D
     mpz_sub_ui(h0,D,3); mpz_div_ui(h0,h0,4); mpz_powm(h0,kN,h0,D);
  
     // h1 = kN*h0 mod D  (=sqrt(kN) (mod D)) 
     mpz_mul(h1,kN,h0); mpz_mod(h1,h1,D);
   }

  // h2 = (2*h1)^-1 * ((kN-h1^2) /D) mod D
  mpz_mul(h2,h1,h1); mpz_sub(h2,kN,h2);
  {
    mpz_mod(x,h2,D);
    if (mpz_cmp_ui(x,0)!=0)
      { 
	cerr << "Fehler: Division mit Rest ";
	mpz_out_str(stderr,10,x); cerr << " !" << endl;
	exit(1);
      }
  }
  mpz_div(h2,h2,D);
  mpz_mul_ui(x,h1,2); mpz_invert(x,x,D); mpz_mul(h2,x,h2);
  mpz_mod(h2,h2,D);
  
  // B = h1 + h2*D mod A
  mpz_mul(B,h2,D); mpz_add(B,B,h1); mpz_mod(B,B,A);
  if (mpz_mod_ui(x,B,2)==0) mpz_sub(B,A,B); 
  
  // C = (B^2-kN) / (4*A)
  mpz_mul(C,B,B); mpz_sub(C,C,kN); mpz_div(C,C,A); mpz_div_ui(C,C,4);

#if 0
  // vollstndige (aber wohl unntige Ausgabe)
  cout << "Die Parameter des Polynoms A*x^2+B*x+C lauten: " << endl;
  cout << "A=" << A << endl;
  cout << "B=" << B << endl;
  cout << "C=" << C << endl;
  cout << "D=" << D << " mit A=D^2" << endl;
#else
  // kurze Ausgabe der wesentlichen Infos:
  cout << "Neues Polynom mit D=" << D << " ermittelt." << endl;
#endif
  
  mpz_mul_ui(D2_inv_mod_kN,D,2); mpz_invert(D2_inv_mod_kN,D2_inv_mod_kN,kN);
  //testen();
}

void CPolynom::testen(const signed int Siebpos=0)
{
  mpz_set_si(h2,Siebpos);
  
  // Variante 1: ( (2*A*wert+B) / (2*D) )^2 mod kN
  mpz_mul_ui(x,A,2); mpz_mul(x,x,h2); mpz_add(x,x,B);
  //mpz_div(x,x,D); mpz_div_ui(x,x,2);
  mpz_mul(x,x,D2_inv_mod_kN); mpz_mod(x,x,kN);
  mpz_mul(x,x,x); mpz_mod(x,x,kN);
  if (mpz_cmp(x,kN_div2)>0) { mpz_sub(x,kN,x); mpz_neg(x,x); } // ggf. negativen Wert berechnen!
  
  // Variante 2: A*wert^2 + B*wert + C mod kN
  
  mpz_mul(h0,A,h2); mpz_mul(h0,h0,h2);
  mpz_mul(h1,B,h2);
  mpz_add(y,h0,h1); mpz_add(y,y,C); mpz_mod(y,y,kN);
  if (mpz_cmp(y,kN_div2)>0) { mpz_sub(y,kN,y); mpz_neg(y,y); } // ggf. negativen Wert berechnen!
  
  cout << "Test: Q(" << Siebpos << ")" << endl;
  cout << " =" << x << endl;
  cout << " =" << y << endl;

  if (mpz_cmp(x,y)!=0)
    {
      cerr << "Fehler: Werte sind verschieden!" << endl;
      exit(1);
    }
}

void CPolynom::Werte_holen(const signed int Siebpos, mpz_t &Wurzelwert, mpz_t &Q) const
{  // Variante 1: ( (2*A*wert+B) / (2*D) )^2 mod kN
  mpz_set_si(Wurzelwert,Siebpos);
  mpz_mul(Wurzelwert,Wurzelwert,A2);
  mpz_add(Wurzelwert,Wurzelwert,B);
  mpz_mul(Wurzelwert,Wurzelwert,D2_inv_mod_kN); mpz_mod(Wurzelwert,Wurzelwert,kN);
  
  mpz_mul(Q,Wurzelwert,Wurzelwert); mpz_mod(Q,Q,kN);
  if (mpz_cmp(Q,kN_div2)>0) mpz_sub(Q,Q,kN); // ggf. negativen Wert berechnen!
  
  //cout << "Werte fr " << Siebpos << " ";
  //mpz_out_str(stdout,10,Wurzelwert); cout << " -> ";
  //mpz_out_str(stdout,10,Q); cout << endl;
}


void CPolynom::Daten_sichern(ostream &ostr) // fr Recovery etc.
{
  // Daten sichern, die fr die Ermittlung des aktuellen Polynoms wichtig sind
  // D ist der Wert, der das aktuelle Polynom aus der Funktionsschar bestimmt.
  ostr << D << endl;
}

void CPolynom::Daten_laden(istream &in) // fr Recovery etc.
{
  // Daten laden, die fr die Ermittlung des aktuellen Polynoms wichtig sind 
  in >> D; in.ignore(1,'\n');
  mpz_sub_ui(D,D,4); // um vier vermindern
  naechstes_Polynom_ermitteln(); // und neu aufsetzen
}

void alle_Deltas_neu_berechnen()
{
  /* Beim Wechsel auf ein neues Polynom werden alle bisherigen Deltawerte der Primzahlen
     ungltig. Daher ist es erforderlich, diese neu zu berechnen */

  cout << "Berechne die Deltas neu" << endl;

  // zunchst alle (mglicherweise) noch bestehenden dynamischen Faktoren aus der Warteschlange nehmen:
  //Sieb_Delta_Heap.clear();
  while (!Sieb_Delta_Heap.empty()) Sieb_Delta_Heap.pop(); // fuer priority_queue, die kein clear hat

  // ebenso die Quadrate:
  //Sieb_Delta_QuadrateHeap.clear();
  while (!Sieb_Delta_QuadrateHeap.empty()) Sieb_Delta_QuadrateHeap.pop(); // fuer priority_queue, die kein clear hat
 
  
  // als erstes die Deltas aus der statische Faktorbasis ermitteln
  unsigned int inv_A2_modp, PolyB;
  mpz_t inv_A2, P, wuq;
  mpz_init(inv_A2); mpz_init(P); mpz_init(wuq);
  int i=1;
  if (Primzahlen[i]==2) i++; // Primzahl 2 berspringen, da Polynom damit Probleme hat...
  for (; i<Primbasis_size; i++)
    {
      //mpz_set_ui(P,Primzahlen[i]);
      //mpz_invert(inv_A2,Polynom.A2,P);
      inv_A2_modp=invmod(mpz_mod_ui(x,Polynom.A2,Primzahlen[i]),Primzahlen[i]);
      PolyB=mpz_mod_ui(x,Polynom.B,Primzahlen[i]);
      
      //mpz_set_ui(x,SQRT_kN_der_Primzahlen[i]);
      //mpz_sub(x,x,Polynom.B); mpz_mul_ui(x,x,inv_A2_modp); 
      //Delta_der_Primzahlen[i][0]=mpz_mod_ui(x,x,Primzahlen[i]);
      register int d;
      d=SQRT_kN_der_Primzahlen[i]-PolyB; if (d<0) d+=Primzahlen[i];
      Delta_der_Primzahlen[i][0]=mulmod(d,inv_A2_modp,Primzahlen[i]);
      
      //mpz_set_ui(x,SQRT_kN_der_Primzahlen[i]); mpz_neg(x,x);
      //mpz_sub(x,x,Polynom.B); mpz_mul_ui(x,x,inv_A2_modp); 
      //Delta_der_Primzahlen[i][1]=mpz_mod_ui(x,x,Primzahlen[i]);
      d=Primzahlen[i]-SQRT_kN_der_Primzahlen[i]-PolyB; if (d<0) d+=Primzahlen[i];
      Delta_der_Primzahlen[i][1]=mulmod(d,inv_A2_modp,Primzahlen[i]);

#if 0
      //ggf. testen, ob Werte in Ordnung sind...
      Polynom.Werte_holen(Delta_der_Primzahlen[i][0],x,y);
      if (mpz_mod_ui(y,y,Primzahlen[i])!=0) 
	{ 
	  cerr << "Deltawert0 " << Delta_der_Primzahlen[i][0] << " liefert Teiler fr "
	       << Primzahlen[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_der_Primzahlen[i][0]-=mpz_get_ui(y);
	  if (Delta_der_Primzahlen[i][0]<0) Delta_der_Primzahlen[i][0]+=Primzahlen[i];
	}
      Polynom.Werte_holen(Delta_der_Primzahlen[i][1],x,y);
      if (mpz_mod_ui(y,y,Primzahlen[i])!=0)
	{ 
	  cerr << "Deltawert1 " << Delta_der_Primzahlen[i][1] << " liefert Teiler fr "
	       << Primzahlen[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_der_Primzahlen[i][1]-=mpz_get_ui(y);
	  if (Delta_der_Primzahlen[i][1]<0) Delta_der_Primzahlen[i][1]+=Primzahlen[i];
	}
#endif

#ifdef SIEVING_LARGE_SQUARES
      if (i>FB_maxQuadrate && Vorfaktor_von_kN%Primzahlen[i]!=0)
	{
	  // Fr alle "groen" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) zusammen mit den dynamischen Faktoren sieben.
	  mpz_set_ui(P,Primzahlen[i]); mpz_mul(P,P,P); // Quadrat der Primzahl holen
	  mpz_invert(inv_A2,Polynom.A2,P);

	  // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
	  // Wurzeln von kN modulo Primzahl^2 ermitteln
	  int Wu = SQRT_kN_der_Primzahlen[i];
	  mpz_set_ui(y,2*Wu);
	  if (mpz_invert(y,y,P)==0)
	    {
	      cerr << "PQ-Wurzel fr " << Primzahlen[i] << ": Inverses existiert nicht!" << endl;
	      exit(1);
	    }
	  
	  mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
	  mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);
	  
	  // ggf. Prfung, ob Wurzel korrekt ist:
	  /* 
	     mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
	     if (mpz_cmp_ui(x,0)!=0)
	      {
	        cerr << "PQ-Wurzel nicht korrekt!" << endl;
	        cerr << "Primzahl=" << Primzahlen[i] << endl;
	        cerr << "kN= " << kN << endl;
	        cerr << "SQRT(kN) mod p=" << SQRT_kN_der_Primzahlen[i] << endl;
	        cerr << "SQRT(kN) mod p^2=" << wuq << endl;
	        cerr << "aber falsch: " << x << endl;
	        exit(1); 
	      }
	  */
	  
	  // nun die Deltas:

	  // Delta1
	  mpz_sub(x,wuq,Polynom.B); mpz_mul(x,x,inv_A2);mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,M);
	  if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
	  if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
	    {
	      TSieb_Delta sieb_delta;
	      sieb_delta.factor=Primzahlen[i];
	      sieb_delta.delta=mpz_get_ui(x);
	      Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
	    }
	  
	  // Delta2
	  mpz_neg(x,wuq); mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,M);
	  if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
	  if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
	    {
	      TSieb_Delta sieb_delta;
	      sieb_delta.factor=Primzahlen[i];
	      sieb_delta.delta=mpz_get_ui(x);
	      Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
	    }
	  
	}
#endif
      
    }

#ifdef SIEVING_MORE_LARGE_SQUARES
  // und weitere Quadrate pushen:
  {
    unsigned int Primzahl = biggest_Prime_in_Factorbase;
    unsigned int Obergrenze = 3*biggest_Prime_in_Factorbase;
    cout << "pushe weitere Prim-Quadrate bis " << Obergrenze << endl; 
    while (Primzahl<Obergrenze)
     {
       do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; 
           while( Primzahl%3==0 || Primzahl%5==0 || Primzahl%7==0
                  || Primzahl%11==0 || Primzahl%13==0 || Primzahl%17==0
                  || !probab_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(P,Primzahl);
	}
       while (mpz_legendre(kN,P)<0); /* Primzahl gltig fr Primzahlbasis? */

       mpz_mul(P,P,P); // Quadrat der Primzahl holen
       mpz_invert(inv_A2,Polynom.A2,P);

       // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
       // Wurzeln von kN modulo Primzahl^2 ermitteln
       int Wu = SQRT_kN_mod_Primzahl(Primzahl);
       mpz_set_ui(y,2*Wu);
       if (mpz_invert(y,y,P)==0)
        {
          cerr << "PQ-Wurzel fr " << Primzahl << ": Inverses existiert nicht!" << endl;
          exit(1);
	}
	  
       mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
       mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#if 0	  
       // ggf. Prfung, ob Wurzel korrekt ist:
       mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
       if (mpz_cmp_ui(x,0)!=0)
        {
          cerr << "PQ-Wurzel nicht korrekt!" << endl;
          cerr << "Primzahl=" << Primzahl << endl;
          cerr << "kN= " << kN << endl;
          cerr << "SQRT(kN) mod p=" << Wu << endl;
          cerr << "SQRT(kN) mod p^2=" << wuq << endl;
          cerr << "aber falsch: " << x << endl;
          exit(1); 
        }
#endif	  
       // nun die Deltas:

       // Delta1
       mpz_sub(x,wuq,Polynom.B); mpz_mul(x,x,inv_A2);mpz_mod(x,x,P);
       mpz_sub_ui(x,x,M);
       if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
       if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
        {
          TSieb_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_ui(x);
          Sieb_Delta_Heap.push(sieb_delta); // und ablegen
          Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
          // Bemerkung: Die Primzahl wird im Sieb_Delta_Heap *und* 
          // gleichzeitig im Sieb_Quadrate_Heap abgelegt.
          // Damit ist sichergestellt, da
          // - der Wert in den integer-Bereich pat
          // - der Wert doppelt (d.h. als Quadrat) im Sieb vermerkt wird.
          // Der Effekt eines zu hufigen Siebens im Falle, da
          // Primzahl bereits dynamischer Faktor ist, wird vernachlssigt.
          cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
        }
	  
       // Delta2
       mpz_neg(x,wuq); mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,M);
       if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
       if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
        {
          TSieb_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
	  sieb_delta.delta=mpz_get_ui(x);
	  Sieb_Delta_Heap.push(sieb_delta); // und ablegen
          Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
          // Bemerkung: Die Primzahl wird im Sieb_Delta_Heap *und* 
          // gleichzeitig im Sieb_Quadrate_Heap abgelegt.
          // Damit ist sichergestellt, da
          // - der Wert in den integer-Bereich pat
          // - der Wert doppelt (d.h. als Quadrat) im Sieb vermerkt wird.
          // Der Effekt eines zu hufigen Siebens im Falle, da
          // Primzahl bereits dynamischer Faktor ist, wird vernachlssigt.
	  cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
        }
	  
     }
  }
#endif

if (!Sieb_Delta_QuadrateHeap.empty())
 cout << "Es wurden " << Sieb_Delta_QuadrateHeap.size() << " Quadrate gepusht." << endl;
  
  for (i=0; i<anz_weitere_Primzahlpotenzen; i++)
    {
      // Fr diese Zahlen auch die Deltas bzgl. der jeweiligen Potenz ermitteln
      // (fr das Sieben zusammen mit den statischen Faktoren)

      //mpz_set_ui(P,Primzahlpotenzen[i]);
      //mpz_invert(inv_A2,Polynom.A2,P);
      inv_A2_modp=invmod(mpz_mod_ui(x,Polynom.A2,Primzahlpotenzen[i]),Primzahlpotenzen[i]);
      PolyB=mpz_mod_ui(x,Polynom.B,Primzahlpotenzen[i]);

      //mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[i]);
      //mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      //Delta_der_Primzahlpotenzen[i][0]=mpz_mod_ui(x,x,Primzahlpotenzen[i]);
      register int d;
      d=SQRT_kN_der_Primzahlpotenzen[i]-PolyB; if (d<0) d+=Primzahlpotenzen[i];
      Delta_der_Primzahlpotenzen[i][0]=mulmod(d,inv_A2_modp,Primzahlpotenzen[i]);
      
      //mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[i]); mpz_neg(x,x);
      //mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      //Delta_der_Primzahlpotenzen[i][1]=mpz_mod_ui(x,x,Primzahlpotenzen[i]);
      d=Primzahlpotenzen[i]-SQRT_kN_der_Primzahlpotenzen[i]-PolyB; if (d<0) d+=Primzahlpotenzen[i];
      Delta_der_Primzahlpotenzen[i][1]=mulmod(d,inv_A2_modp,Primzahlpotenzen[i]);
    }
  
  mpz_clear(inv_A2); mpz_clear(P); mpz_clear(wuq);
  
  // dann die dynamische Faktorbasis behandeln:
  TDynamic_Factor_Relations::iterator pos;
  for (pos=Dynamic_Factor_Relations.begin();
       pos!=Dynamic_Factor_Relations.end(); ++pos) Deltas_berechnen_fuer(*pos, -M);
  
  cout << "Deltaberechnung beendet." << endl;
}


short signed int gZ_priority (const char ch)
{
  switch (ch)
   {
     case '+': case '-': return 1;
     case '*': case '/': case ':': return 2;
     case '^': return 3;
     default: return 0;
   } 
}

bool gZ_priority_le (const char a, const char b)
{
  return gZ_priority(a)<=gZ_priority(b);
}

bool gZ_priority_gr (const char a, const char b)
{
  return gZ_priority(a)>gZ_priority(b);
}

#include "numbpart.cc" /* for calculation of partition numbers */

bool get_Zahl (mpz_t &n, char* &s, const char op='R', const char upto_op='\0')
{ 
  /* Zeichenkette in multiple-precision-Zahl umwandeln.
     Die Zeichenkette kann ein beliebiger Term mit Klammern,
     und den Operationszeichen +,-,*,:,/,^ sein, solange er numerisch
     auswertbar ist.
  */
  
  while (s[0]==' ') s++; // ggf. Leerzeichen berlesen
  
  bool term_structure_ok = true;
  char* s_anf = s; // Start-Position des aktuellen Subterms
  char h;
  mpz_t tmp;
  mpz_init(tmp);

  if (s[0]=='(')
    {
      s++; // ffnende Klammer berspringen
      term_structure_ok=get_Zahl(tmp,s); // geklammerten Term holen
      if (s[0]==')') s++; // schlieende Klammer berspringen
      else 
	{
	  term_structure_ok=false;
	  cerr << "error: ')' expected at '" << s << "'" << endl; 
	}
    }
  else
   if (s[0]=='F' || s[0]=='L' || s[0]=='P' || s[0]=='p') // Fibonacci-/Lucas-/Partition-Index folgt...
    {
      /*
         F - Fibonnaccizahl-Index
         L - Lascaszahl-Index
         P - Partitionszahl-Index
         p - ermittelt die nachfolgende Primzahl (falls Argument nicht prim ist)
      */
      s++; // Funktionssymbol F, L, P bzw. p berspringen
      term_structure_ok=get_Zahl(tmp,s,s_anf[0]); // Term als Fibonacci-/Lucaszahl auswerten
    }
   else
    { // ungeklammerten Term holen
      char* anf=s;
      while (s[0]>='0' && s[0]<='9') s++;
      h=s[0]; s[0]='\0';
      if (mpz_set_str(tmp, anf, 10))
	{
	  term_structure_ok=false;
	  s[0]=h;
	  cerr << "error: number expected at '" << anf << "'" << endl;
	}
      s[0]=h;
    }
  
  while (s[0]==' ') s++; // Leerzeichen berlesen
  h=s[0]; // nchstes Operationszeichen
  
  if (term_structure_ok) switch (op)
    {
    case '\0':
      break;
    case 'R' :
      //cout << "Reset" << endl;
      mpz_set(n,tmp);
      break;
    case 'F' : // Fibonacci-Nummer
      {
       if (mpz_cmp_ui(tmp,10000)>0)
        {
         cerr << "Fibonacci-Index zu gro: " << tmp << endl;
         exit(1);
        }
       if (mpz_sgn(tmp)<0)
        {
         cerr << "Fibonacci-Index ist negativ: " << tmp << endl;
         exit(1);
        }
       cout << "berechne Fibonacci-Zahl" << endl;
       unsigned int i=mpz_get_ui(tmp);
       if (i<1) mpz_set_ui(n,0);
       if (i==1) mpz_set_ui(n,1);
       if (i>1)
        {
         mpz_t f0,f1,f2;
         mpz_init(f0); mpz_init(f1); mpz_init(f2);
         mpz_set_ui(f0,0); mpz_set_ui(f1,1);
         for (unsigned int j=2; j<=i; ++j)
          {
            mpz_add(f2,f1,f0); mpz_set(f0,f1); mpz_set(f1,f2);
          }
         mpz_set(n,f2);
         mpz_clear(f0); mpz_clear(f1); mpz_clear(f2);
        }
      }
      break;
    case 'L' : // Lucas-Nummer
      {
       if (mpz_cmp_ui(tmp,10000)>0)
        {
         cerr << "Lucas-Index zu gro: " << tmp << endl;
         exit(1);
        }
       if (mpz_sgn(tmp)<0)
        {
         cerr << "Lucas-Index ist negativ: " << tmp << endl;
         exit(1);
        }
       cout << "berechne Lucas-Zahl" << endl;
       unsigned int i=mpz_get_ui(tmp);
       if (i<1) mpz_set_ui(n,2);
       if (i==1) mpz_set_ui(n,1);
       if (i>1)
        {
         mpz_t f0,f1,f2;
         mpz_init(f0); mpz_init(f1); mpz_init(f2);
         mpz_set_ui(f0,2); mpz_set_ui(f1,1);
         for (unsigned int j=2; j<=i; ++j)
          {
            mpz_add(f2,f1,f0); mpz_set(f0,f1); mpz_set(f1,f2);
          }
         mpz_set(n,f2);
         mpz_clear(f0); mpz_clear(f1); mpz_clear(f2);
        }
      }
      break;
    case 'P' : // Partitionszahl
      {
       if (mpz_cmp_ui(tmp,1000000)>0)
        {
         cerr << "Partitionszahl-Index zu gro: " << tmp << endl;
         exit(1);
        }
       if (mpz_sgn(tmp)<0)
        {
         cerr << "Partitionszahl-Index ist negativ: " << tmp << endl;
         exit(1);
        }
       cout << "berechne Partitionszahl" << endl;
       numbpart::numbpart(n,mpz_get_ui(tmp));
      }
      break;
    case 'p' : // nachfolgende Primzahl von tmp ermitteln (sofern tmp nicht prim ist)
      {
        mpz_set(n,tmp);
        while (!mpz_probab_prime_p(n,50)) mpz_add_ui(n,n,1);
      }
      break;
    case '+' :
      if (gZ_priority_gr(h,op))
	{
	  //cout << "Punkt vor Strich" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h,'+'); h=s[0];
	}	
      //cout << "Addition" << endl;	
      mpz_add(n,n,tmp);
      break;
    case '-' :
      if (gZ_priority_gr(h,op))
	{
	  //cout << "Punkt vor Strich" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h,'-'); h=s[0];
	}	
      //cout << "Subtraktion" << endl;
      mpz_sub(n,n,tmp);
      break;
    case '*' : 
      if (gZ_priority_gr(h,op))
	{
	  //cout << "Exponentiation vor Multiplikation" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h,'*'); h=s[0];
	}	
      //cout << "Multiplikation" << endl;
      mpz_mul(n,n,tmp);
      break;
    case '/' :
    case ':' :
      {
        if (gZ_priority_gr(h,op))
  	  {
	    //cout << "Exponentiation vor Division" << endl; 
	    term_structure_ok=get_Zahl(tmp,++s,h,'/'); h=s[0];
	  }	
        //cout << "Division" << endl;
        mpz_t q;
        mpz_init(q);
        mpz_div(q,n,tmp);
        mpz_mul(tmp,q,tmp);
        if (mpz_cmp(tmp,n)) // bleibt Rest?
         {
           cout << "Warnung: ganzzahlige Division ohne Restbercksichtigung!" << endl;
         }
        mpz_set(n,q);
        mpz_clear(q);
      }
      break;
    case '^' :
      if (h=='^') cout << "Warnung: a^b^c ambiguent. Setze (a^b)^c." << endl; 
      // cout << "Exponentiation" << endl;
      mpz_pow_ui(n,n,mpz_get_ui(tmp));
      break;
    default :
      s=--s_anf;
      cerr << "error: operation +-*/:^ expected at '" << s << "'" << endl;	
      term_structure_ok=false;
      break;	 
    }
  mpz_clear(tmp);

  if (term_structure_ok && h!='\0' && h!=')' )
   {
     // Term ist bisher in Ordnung,
     // es liegt kein Klammerungsende vor und ein Operator folgt.
     if (upto_op=='\0') return get_Zahl(n,++s,h);
     if (gZ_priority_le(h,upto_op)) return term_structure_ok;
     else return get_Zahl(n,++s,h,upto_op);
   }
  else return term_structure_ok;
}

#ifdef LAZY_EVALUATION
const bool ohne_multi_kombinieren_init = false; // true ->multi-init, false ->kein multi_init bei Relation_einfuegen

bool SpecialFactorRelation_LAZY_vermerken(const Tfactor &DLP, const char* GL_String)
{
  // Special-Factor in die Special-Factor-Set eintragen (doppelt, damit
  // beide Large-Primes schnell zugreifbar sind)
  // Falls der Special-Factor schon vorhanden sein sollte, kombinieren
  // und ins Gleichungssystem eintragen.

  if (DLP.LP1()==DLP.LP2()) // Sonderfall: DoubleLargePrime=LargePrime^2
   {
     cout << "Spezialfall: DLP ist Quadrat! (in LAZY)" << endl;
     istrstream is(GL_String);
     CGleichungszeile* GL = new CGleichungszeile();
     GL->multi_kombinieren_init();
     Tfactor DLP_2 = GL->multi_kombinieren_main(is);
     if (DLP!=DLP_2) { cerr << "DLP-Fehler irgendwo in LAZY!" << endl; exit(1); }
     Special_hit++;
     // immer beachten: Delta ist nicht das Quadrat, sondern wird quadriert,
     // daher zu Delta immer nur Wurzeln multiplizieren!
     mpz_mul_ui(GL->Delta,GL->Delta,DLP.LP1()); mpz_mod(GL->Delta,GL->Delta,n);
     Relation_einfuegen(GL,ohne_multi_kombinieren_init); // LAZY -> einfgen geshieht hier...
     return false;
   }
  
  TSpecialFactorRelation relation;
  relation.factor=DLP;

  TSpecial_Factor_Relations::iterator pos = Special_Factor_Relations.find(relation);
  if (pos==Special_Factor_Relations.end())
   {
     //cout << "Speichere DLP-Relation (LAZY)" << endl;
     //relation.fpos=GL->speichern(SpecialRelations_to_file,relation.factor);
     relation.fpos=SpecialRelations_to_file.tellp();
     SpecialRelations_to_file << GL_String;

     Special_Factor_Relations.insert(relation);
     relation.factor.swap();
     Special_Factor_Relations.insert(relation);

     // regelmig Zyklensuche auf den Special-Relations durchfhren:
     static int naechste_Zyklensuche = 50000;
     if (!--naechste_Zyklensuche)
      {
        naechste_Zyklensuche=50000;
	SpecialFactors_Zyklensuche();
      }
     return true;
   }
  else
   { 
     // Special-Factor ist bereits bekannt
     cout << "Special Hit durch bekannten Special Factor! (LAZY)" << endl;
     istrstream is(GL_String);
     CGleichungszeile* GL = new CGleichungszeile();
     GL->multi_kombinieren_init();
     Tfactor DLP_2 = GL->multi_kombinieren_main(is);
     if (DLP!=DLP_2) { cerr << "DLP-Fehler irgendwo in LAZY!" << endl; exit(1); }
     Special_hit++;
     mpz_t x;
     mpz_init(x); DLP.in_mpz_ablegen(x);
     mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
     mpz_clear(x);
     GL->multi_kombinieren_main(SpecialRelations_from_file,(*pos).fpos);
     //GL->multi_kombinieren_exit() kann entfallen, wenn zugehriges init bei Relation_einfuegen entfllt
     Relation_einfuegen(GL,ohne_multi_kombinieren_init); // LAZY -> einfgen geshieht hier...
     return false;
   }
}
#endif /* LAZY */


#ifdef IS_STANDALONE
void bearbeite_ecm()
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // Random-Seed setzen (Zeitfunktion drfte hierzu zufllig genug sein)
  
  unsigned int dezimalstellen = mpz_sizeinbase(n,10);
  tune_parameters(dezimalstellen); // Parameter abhngig von der Gre der Zahl!

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // endl-berlesen
  while (Anzahl_Kurven<elcu_Kurven)
  {

    Anzahl_Kurven++;
    cout << "Elliptische Kurve Nr. " << Anzahl_Kurven << "/" << elcu_Kurven << endl;

    int ecm_sigma = rand(); // zufllige Kurve whlen
    elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
    elliptic_curve.go(ecm_sigma,elcu_Phase1,elcu_Phase2); /* Faktorisierung mit elliptischer Kurve versuchen */

    // Anzahl bearbeiteter Kurven ins Recovery-File schreiben...
    Recovery_to_file.seekp(fpos); 
    Recovery_to_file << Anzahl_Kurven << endl << flush;

    // Teiler gefunden?
    if (dezimalstellen==mpz_sizeinbase(n,10)) continue; // kein neuer Faktor
    
    // Stelligkeit von n hat sich gendert -> neuer Faktor

    if (mpz_probab_prime_p(n,probab_prime_checks))
      {
        Factorization_to_file << MAL() << n << endl;
        cout << "Restfaktor: " << n << endl;
        cout << "Verbleibende Zahl ist wahrscheinlich prim!" << endl;
        mpz_set_ui(n,1); // n ist faktorisiert
      }

    if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
      { 
        cout << "Faktorisierung erfolgreich beendet." << endl;
        Factorization_to_file << endl;
    
        // Streams schlieen
        SpecialRelations_buffer.close();
        DynamicRelations_buffer.close();
        StaticRelations_buffer.close();
        Recovery_buffer.close();

        // Files lschen
        remove(SpecialRelationsFile);
	remove(DynamicRelationsFile);
	remove(StaticRelationsFile);	
	remove(RecoveryFile);
	exit(0); // Programm erfolgreich beenden...
      }

    // fr weitere ecm-Runden Parameter neu tunen...
    dezimalstellen=mpz_sizeinbase(n,10);
    tune_parameters(dezimalstellen);

    // nun mu noch das recovery-File angepat werden, da sich
    // die zu faktorisierende Zahl verkleinert hat!
    Recovery_buffer.close(); // File schlieen
    Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::trunc); // wieder ffnen (leeres File erzeugen!)
    Recovery_to_file << n << endl; // zu faktorisierende Zahl ablegen
    fpos = Recovery_from_file.tellg(); // Position merken
    Recovery_to_file << Anzahl_Kurven << endl << flush; // Anzahl bisher bearbeiteter Kurven vermerken

  }
  cout << "ECM-Phase (standalone) beendet." << endl;
}
#endif


#ifdef IS_SERVER

void bearbeite_daten_stream_ecm(int server_socket)
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // Random-Seed setzen (Zeitfunktion drfte hierzu zufllig genug sein)

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // endl-berlesen

  while (Anzahl_Kurven<elcu_Kurven)
  {
    int communication_socket = accept(server_socket, NULL, NULL);
    if (communication_socket == 0) continue; // auf Verbindung warten

    unix_buffer buffer(communication_socket);
    iostream daten(&buffer);

    daten >> s;

    if (s == "ECM?")
     {
       int ecm_sigma = rand();
       daten << ecm_sigma << " "
             << elcu_Phase1 << " "
             << elcu_Phase2 << " "
             << n << endl; // sigma und n bertragen

       Anzahl_Kurven++;
       cout << "Elliptische Kurve Nr. " << Anzahl_Kurven << "/" << elcu_Kurven
            << " mit sigma=" << ecm_sigma
            << " verschickt." << endl;

       // Anzahl vergebener Kurven ins Recovery-File schreiben...
       Recovery_to_file.seekp(fpos); 
       Recovery_to_file << Anzahl_Kurven << endl << flush;

       daten << flush;
       close (communication_socket);
       continue;
     }

    if (s == "Faktor(ECM)!") // Faktoren annehmen...
      {
        int sigma;
        mpz_t x;
        mpz_init(x);
        daten >> sigma >> x; // Kurveninformation (sigma) und Faktor holen
        mpz_gcd(x,x,n); // ggt bilden (Faktor knnte bereits entdeckt sein!)
        while (mpz_cmp_ui(x,1)!=0)
         { // theoretisch knnte Faktor mehrfach teilen...
           if (mpz_probab_prime_p(x,probab_prime_checks))
             {
               cout << x << " ist Faktor." << endl;
               if (mpz_sizeinbase(x,10)<20)
                 Factorization_to_file << MAL() << x << " [ecm]" << flush;
               else
                 Factorization_to_file << MAL() << x << " [ecm,sigma=" << sigma << "]" << flush;
             }
           else
             {
               cout << x << " ist zusammengesetzter Faktor." << endl;
               if (mpz_sizeinbase(x,10)<20)
                 Factorization_to_file << MAL() << x << " [ecm] [composite]" << flush;
               else
                 Factorization_to_file << MAL() << x << " [ecm,sigma=" << sigma << "] [composite]" << flush;
             }
	   mpz_div(n,n,x); // Faktor rausdividieren
	   mpz_gcd(x,x,n); // und erneut testen...
         }
        mpz_clear(x);
        daten << flush;
        close (communication_socket);

        if (mpz_probab_prime_p(n,probab_prime_checks))
          {
            Factorization_to_file << MAL() << n << endl;
            cout << "Restfaktor: " << n << endl;
            cout << "Verbleibende Zahl ist wahrscheinlich prim!" << endl;
            mpz_set_ui(n,1); // n ist faktorisiert
          }

        if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
          { 
            cout << "Faktorisierung erfolgreich beendet." << endl;
            Factorization_to_file << endl;
    
            // Streams schlieen
            SpecialRelations_buffer.close();
            DynamicRelations_buffer.close();
	    StaticRelations_buffer.close();
  	    Recovery_buffer.close();

	    // Files lschen
	    remove(SpecialRelationsFile);
	    remove(DynamicRelationsFile);
	    remove(StaticRelationsFile);	
	    remove(RecoveryFile);
	    exit(0); // Programm erfolgreich beenden...
          }
        tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning

        // nun mu noch das recovery-File angepat werden, da sich
        // die zu faktorisierende Zahl verkleinert hat!
        Recovery_buffer.close(); // File schlieen
        Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::trunc); // wieder ffnen (leeres File erzeugen!)
        Recovery_to_file << n << endl; // zu faktorisierende Zahl ablegen
        fpos = Recovery_from_file.tellg(); // Position merken
        Recovery_to_file << Anzahl_Kurven << endl << flush; // Anzahl bisher bearbeiteter Kurven vermerken

	continue;
      }

    cerr << "Falsche Anfrage \"" << s << "\"" << endl;
    daten << flush;
    close (communication_socket);
  }
  cout << "Verteilte ECM-Phase (serverseitig) beendet." << endl;
}

void bearbeite_daten_stream(iostream &daten)
{
  string s;
  Tfactor factor;
  TDynamic_Factor_Relations::iterator pos;
  CGleichungszeile* GL;
  daten >> s;
  
  if (s == "Hunger!")
    {
      cout << "Sende zu faktorisierende Zahl." << endl;
      daten << n << " " << Primbasis_size << " " 
	    << Factor_Threshold << " " << M << endl;
      return;
    }

  if (s == "ECM?")
   {
     daten << "-1" << endl; // in dieser Phase keine elliptischen Kurven (-> sigma<0)
   }

  if (s == "Faktor(ECM)!") // Faktoren werden noch akzeptiert...
    {
      int sigma;
      mpz_t x;
      mpz_init(x);
      daten >> sigma >> x; // Kurveninformation (sigma) und Faktor holen
      mpz_gcd(x,x,n); // ggt bilden (Faktor knnte bereits entdeckt sein!)
      while (mpz_cmp_ui(x,1)!=0)
	{ // theoretisch knnte Faktor mehrfach teilen...
          if (mpz_probab_prime_p(x,probab_prime_checks))
            {
              cout << x << " ist Faktor." << endl;
              if (mpz_sizeinbase(x,10)<20)
                Factorization_to_file << MAL() << x << " [ecm]" << flush;
              else
                Factorization_to_file << MAL() << x << " [ecm,sigma=" << sigma << "]" << flush;
            }
          else
            {
              cout << x << " ist zusammengesetzter Faktor." << endl;
              if (mpz_sizeinbase(x,10)<20)
                Factorization_to_file << MAL() << x << " [ecm] [composite]" << flush;
              else
                Factorization_to_file << MAL() << x << " [ecm,sigma=" << sigma << "] [composite]" << flush;
            }
	  mpz_div(n,n,x); // Faktor rausdividieren
	  mpz_gcd(x,x,n); // und erneut testen...
        }
      mpz_clear(x);

      if (mpz_probab_prime_p(n,probab_prime_checks))
        {
          Factorization_to_file << MAL() << n << endl;
          cout << "Restfaktor: " << n << endl;
          cout << "Verbleibende Zahl ist wahrscheinlich prim!" << endl;
          mpz_set_ui(n,1); // n ist faktorisiert
        }

      if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
        { 
          cout << "Faktorisierung erfolgreich beendet." << endl;
          Factorization_to_file << endl;
	  
          // Streams schlieen
          SpecialRelations_buffer.close();
          DynamicRelations_buffer.close();
	  StaticRelations_buffer.close();
	  Recovery_buffer.close();

	  // Files lschen
	  remove(SpecialRelationsFile);
	  remove(DynamicRelationsFile);
	  remove(StaticRelationsFile);	
	  remove(RecoveryFile);
	  exit(0); // Programm erfolgreich beenden...
        }

      /* Falls der Programmflu diese Stelle erreicht, so knnen
         verschiedene Flle eingetreten sein:
         Fall 1: Ein defektierender Client hat einen falschen Faktor
                 als ECM-Faktor gemeldet -> schadet sonst nicht weiter,
                 da er nicht abgespalten konnte.
         Fall 2: Ein langsamer Client hat im ECM-Mode als "Nachzgler" einen
                 bereits entdeckten Faktor aufgrund seiner veralteten Daten
                 "wiederentdeckt" -> schadet sonst nicht weiter, da der
                 Faktor nicht abgespalten werden konnte.
         Fall 3: Ein Client hat in seiner ECM-Phase einen neuen Faktor
         !!!!!!  entdeckt, der nicht zu einer kompletten Faktorisierung
                 gefhrt hat. -> Da sich der Server aber bereits in der
                 Siebphase befindet (sehr wahrscheinlich sogar noch in der
                 Anfangsphase), sind extreme Laufzeitverluste zu befrchten,
                 da das Sieb nicht auf die nun verkleinerte Zahl umgestellt
                 werden kann.
                 -> Richtig schlimm wird es aber erst, wenn das Programm
                 abgebrochen wird, da die Recovery-Daten nun inkonsistent
                 sind. (Abgespaltener Faktor kann nur in den Recovery-Daten
                 gepeichert werden, wenn die Siebdaten gelscht werden!
                 Umgekehrt wre es unklug, den gefundenen Faktor einfach
                 zu ignorieren.)
                 Bemerkung: Fall 3 lt sich beim verteilten Faktorisieren
                 nicht ausschlieen, da der Server keine Annahmen ber den
                 Status der Clients treffen kann. Es bringt also nichts,
                 die Siebphase solange verzgern zu wollen, bis alle Clients
                 den ECM-Mode verlassen haben. (Abgesehen davon wre das
                 dazu quivalent, alle ECM-Faktoren in der Siebphase einfach
                 zu ignorieren.)

           <- diese Signatur dient nur dem schnellen Auffinden von "Programm-Baustellen"
	  Bercksichtigung insbesondere des dritten Falls ist noch nicht
          implementiert. Am besten wre es wohl, die Recovery-Daten anzupassen
          und das Programm mit Neustart-Aufforderung zu beenden.
          Andererseits wird durch den "RECOVERY"-Marker in "factorizations.txt"
          sowieso auf mgliche Inkonsistenzen aufmerksam gemacht...
      */            
                 
      return;
    }
  
  if (s=="DynamicFactors?_ab_fpos")
    {
      streampos startfpos; daten >> startfpos;
      cout << "Sende dynamische Faktoren ab Fileposition " << startfpos << endl;
      DynamicRelations_to_file.seekp(0,ios::end); daten << DynamicRelations_to_file.tellp() << " ";
      int prev=0; // wir bertragen die Abstnde
      for (pos= Dynamic_Factor_Relations.begin();
	   pos != Dynamic_Factor_Relations.end(); ++pos)
	if ((*pos).fpos>=startfpos)
	  {
	    daten << (*pos).factor-prev << " ";
            prev=(*pos).factor;
	  }
      daten << "-1" << endl;
      return;
    }
  
  if (s=="Polynom?")
    {
      int intervall;
      daten >> intervall;
      cout << "Sende neues Polynomintervall (" << intervall << ")" << endl;
      Polynom.Daten_sichern(daten);
      Polynom.naechstes_Polynom_ermitteln(intervall);
      Polynom.Daten_sichern(daten);
      
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.Daten_sichern(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
      return;
    }

#ifndef LAZY_EVALUATION 
  // BUSY-Variante (bisherige Implementierung):
  // alle Relationen sofort vollstndig auswerten
  if (s == "Relation!")
    { bool Relationsblock_okay = true;
      int anz_empfangene_Relationen = 0;
      do // Block von Relationen empfangen
     {
      static int counter = 100;
      if (--counter==0) { counter=100; Statusmeldung(); }
      //cout << "Empfange bertragene Relation." << endl;
      anz_empfangene_Relationen++;
      GL = new CGleichungszeile();
      factor = GL->kombinieren(daten);
      //daten.ignore(1,'\n');
      daten >> s;
      if (!GL->check_Relation(factor))
       { // Entweder Irrtum oder Angriff auf den Server mit
         // ungltiger Relation
         cerr << "Server mit ungltiger Relation attackiert!" << endl;
         cerr << "-> Mitteilung an defektierenden client." << endl;
         daten << "ignoriert!" << flush;
	 delete GL; // Speicher der Relation wieder freigeben
         Relationsblock_okay=false;
         continue; // und wieder bereit fr neue Daten 
       }
      if (factor.Typ()==static_prime)
	{
	  // cout << "Habe statische Relation empfangen!" << endl;
	  ++Erfolgsaufteilung[0];
	  Relation_einfuegen(GL);
	  continue;
	}
      if (factor.Typ()==single_large_prime)
	{
	  // Dynamic Factor
	  TDynamicFactorRelation relation;
	  relation.factor = factor.int_value();
	  pos = Dynamic_Factor_Relations.find(relation); // dynamischen Faktor suchen
	  if (pos == Dynamic_Factor_Relations.end())
	    { // noch nicht bekannt
	      relation.fpos = GL->speichern(DynamicRelations_to_file,relation.factor);
	      relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	      if (!relation.insert_cache(GL)) delete GL; // nicht gecached
	      Dynamic_Factor_Relations.insert(relation);
	      Special_Factoren_durch_Primfaktor_splitten(relation.factor);
	    } 
	  else
	    { // schon bekannt
	      //cout << "Kann dynamische Relation zu statischer kombinieren!" << endl;
              ++Erfolgsaufteilung[1];
              factor.in_mpz_ablegen(x); // factor in mpz-Variable x ablegen
	      mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
	      GL->kombinieren(DynamicRelations_from_file, (*pos).fpos);
	      Relation_einfuegen(GL);
	    }
	  continue;
	}
    
      if (factor.Typ()!=double_large_prime)
        {
	  cerr << "Fehler: es wurde ein unbekannter Relationstyp bertragen!" << endl;
          continue;
        }
      // Special-Factor
      // Specialfactor mit Dynamicfactoren testen:
      /* Anmerkung:
	 Beim verteilten Sieben kann es passieren, da ein (notwendigerweise zusammengesetzter)
	 Special-Factor von einem Client entdeckt wird, der nicht mit dem aktuellen Satz der
	 dynamischen Faktoren siebt. - In diesem Fall, der mit steigender Anzahl der Clients immer
	 wahrscheinlicher wird, knnte es sinnvoll sein, Specialfaktoren auf ihre Teilbarkeit
	 mit den vorliegenden dynamischen Faktoren des Servers zu testen und ggf. zu splitten.
	 Allerdings bringt ein Verzicht auf diesen Test keine weiteren Nachteile mit sich.
	 Wir verzichten daher auf diese - nur beim verteilten Sieben mgliche - Prozedur.
      */
      // if(!Special_Factor_durch_Dynamicfactoren_splitten(factor)) // nicht implementiert (s.Anm.)
      {
	// Echter Specialfactor
	if (SpecialFactorRelation_vermerken(factor,GL))
          {
	    delete GL; // und Speicher der Relation wieder freigeben
	  }
	else
	  {
	    // Special-Factor ist bereits bekannt
	    Relation_einfuegen(GL);
	  }
      }
      //continue;
     } while (s=="Relation!");
     // Relationsblock ist eingelesen
     cout << "Relationsblock mit " << anz_empfangene_Relationen 
          << " Relationen eingelesen." << endl;
     if (!Relationsblock_okay) daten << "ignoriert!" << flush;
     else
      if (s=="Relationsblock_Ende") daten << "empfangen." << flush;
      else
       {
         cerr << "Relationsblock-Endekennung nicht erkannt!" << endl;
         daten << "Relationsblock_korrekt_beendet???" << flush;
       }
     return;
    }
#else
  // LAZY-Variante (neue Implementierung, nur teilweise umsetzbar,
  // ohne die bisherige Implementierung zu stark zu verndern):
  // wenn die empfangenen Relationen Korrekturfaktoren enthalten, werden
  // diese erst dann ausgewertet, wenn eine statische Relation entsteht.
  // Diese Variante wird (zunchst) nur fr das Client-Server-Modell
  // implementiert, da das relativ einfach ist und der Server alle
  // Relationen schnell entgegennehmen soll, was in der bisherigen
  // BUSY-Variante bei greren Gleichungssystemen nicht mehr gewhrleistet
  // werden konnte. Auerdem drfte die LAZY-Variante Plattenplatz sparen.
  // Nachteil: Der Check auf defektierende Relationen wird nun
  // unverhltnismig aufwendig. (Per Induktion gilt zwar: Wenn jede Relation
  // getestet wird, dann reicht es, alle Faktoren (einschlielich der 
  // Korrekturfaktoren) und den quadrierten Restfaktor zusammenzumultiplizieren
  // und das Produkt modulo n auf 1 zu testen. Die Korrektheit der zu den
  // Korrekturfaktoren gehrigen Relationen ergibt sich dann durch einen
  // Existenzcheck der zugehrigen Relation (und der Induktionsannahme, da
  // diese korrekt ist). Dieser Test ist also einfach zu implementieren, aber
  // whrend in der Busy-Variante der Test nicht allzu sehr ins Gewicht fllt,
  // nimmt er hier einen groen Teil der Verarbeitungszeit ein, da die
  // Relation nun nicht mehr als Zeichenkette verarbeitet werden kann, sondern
  // tatschlich ausgewertet werden mu. Gerade aber die Auswertung sollte
  // durch die LAZY-Methode auf das notwendige Ma beschrnkt werden.
  //
  // Vorgehensweise der LAZY-Variante:
  //  1. Relation als String empfangen
  //  2. Relation analysieren (factor.Typ ermitteln)
  //  3. Falls statische Relation -> wie gehabt verfahren (BUSY)
  //  4. Analyse-Routinen wie gehabt ausfhren
  //  5. String in das zugehrige File ablegen.
  if (s == "Relation!")
    { 
      bool Relationsblock_okay = true;
      int anz_empfangene_Relationen = 0;
      do // Block von Relationen empfangen
        {
          static int counter = 100;
          if (--counter==0) { counter=100; Statusmeldung(); }
          //cout << "Empfange bertragene Relation." << endl;
          anz_empfangene_Relationen++;

          ostrstream os; // hierin wird die Gleichungszeile eingelesen
          // WICHTIG: strstreams haben freeze-Problematik bei char* !!!!
          // um Speicherfreigabe auf jeden Fall sicherzustellen, sei daher
          // fr diesen Block folgende Konvention vereinbart:
          // Am Ende des Blockes sowie vor *jedem* continue wird os "entfrozen".
          // Werden (dauerhafte) Kopien von os.str() bentigt, dann mssen sie
          // explizit angefertigt werden!

          //cout << "scanne Gleichungszeile vom File" << endl;
          daten >> s; // Startzeichen
          if (s != "G")
           { 
             cerr << "Error: Falsche Position beim Einlesen der Gleichungszeile" << endl;
             exit(1);
           }
          daten >> factor; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
          os << "G " << setprecision(20) << factor; // << " ";
          while (daten.peek()!='\n')
           { 
             char line[1024];
             daten.get(line,1024,'\n'); os << line;
           }
          { char c; daten.get(c); os << c << ends; }
          //cout << os.str();
      
          //daten.ignore(1,'\n');
          daten >> s;
          if (factor.Typ()==static_prime)
	   {
	     //cout << "Habe statische Relation empfangen!" << endl;
             istrstream is(os.str()); // eingelesenen String aus Eingabe zur Verfgung stellen
             GL = new CGleichungszeile();
             GL->multi_kombinieren_init();
             factor = GL->multi_kombinieren_main(is);
	     ++Erfolgsaufteilung[0];
             //cout << "Fge Relation ein." << endl;
	     Relation_einfuegen(GL,ohne_multi_kombinieren_init);
	     os.freeze(false);
             continue;
  	   }
          if (factor.Typ()==single_large_prime)
	   {
	    // Dynamic Factor
	    TDynamicFactorRelation relation;
	    relation.factor = factor.int_value();
	    pos = Dynamic_Factor_Relations.find(relation); // dynamischen Faktor suchen
	    if (pos == Dynamic_Factor_Relations.end())
	      { // noch nicht bekannt
	        //relation.fpos = GL->speichern(DynamicRelations_to_file,relation.factor);
                relation.fpos=DynamicRelations_to_file.tellp();
                DynamicRelations_to_file << os.str(); // schnelles, weil direktes speichern
	        relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	        //if (!relation.insert_cache(GL)) delete GL; // nicht gecached
	        Dynamic_Factor_Relations.insert(relation);
	        Special_Factoren_durch_Primfaktor_splitten_LAZY(relation.factor);
	      } 
	    else
	      { // schon bekannt
	        //cout << "Kann dynamische Relation zu statischer kombinieren!" << endl;
                istrstream is(os.str()); // eingelesenen String aus Eingabe zur Verfgung stellen
                GL = new CGleichungszeile();
                GL->multi_kombinieren_init();
                factor = GL->multi_kombinieren_main(is);
                ++Erfolgsaufteilung[1];
                factor.in_mpz_ablegen(x); // factor in mpz-Variable x ablegen
	        mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
	        GL->multi_kombinieren_main(DynamicRelations_from_file, (*pos).fpos);
                //GL->multi_kombinieren_exit();
	        Relation_einfuegen(GL,ohne_multi_kombinieren_init);
	      }
            os.freeze(false);
	    continue;
	   }
          if (factor.Typ()!=double_large_prime)
           {
	     cerr << "Fehler: es wurde ein unbekannter Relationstyp bertragen!" << endl;
             os.freeze(false); 
             continue;
           }
          // Special-Factor
          // Specialfactor mit Dynamicfactoren testen:
          /* Anmerkung:
	     Beim verteilten Sieben kann es passieren, da ein (notwendigerweise zusammengesetzter)
	     Special-Factor von einem Client entdeckt wird, der nicht mit dem aktuellen Satz der
	     dynamischen Faktoren siebt. - In diesem Fall, der mit steigender Anzahl der Clients immer
	     wahrscheinlicher wird, knnte es sinnvoll sein, Specialfaktoren auf ihre Teilbarkeit
	     mit den vorliegenden dynamischen Faktoren des Servers zu testen und ggf. zu splitten.
	     Allerdings bringt ein Verzicht auf diesen Test keine weiteren Nachteile mit sich.
	     Wir verzichten daher auf diese - nur beim verteilten Sieben mgliche - Prozedur.
          */
         // if(!Special_Factor_durch_Dynamicfactoren_splitten(factor)) // nicht implementiert (s.Anm.)
          {
	    // Echter Specialfactor
	    SpecialFactorRelation_LAZY_vermerken(factor,os.str());
          }
         os.freeze(false);
        } while (s=="Relation!");
      // Relationsblock ist eingelesen
      cout << "Relationsblock mit " << anz_empfangene_Relationen 
           << " Relationen eingelesen." << endl;
      if (!Relationsblock_okay) daten << "ignoriert!" << flush;
      else
       if (s=="Relationsblock_Ende") daten << "empfangen." << flush;
       else
        {
          cerr << "Relationsblock-Endekennung nicht erkannt!" << endl;
          daten << "Relationsblock_korrekt_beendet???" << flush;
        }
      return;
    }
#endif // LAZY-Variante
  
  cerr << "Falsche Anfrage \"" << s << "\"" << endl;
}

#endif

#ifdef IS_CLIENT
#ifdef USE_NETWORK

//void get_sigchild(int sig)
//{
//  int status;
//
//  wait(&status);
//  if (WIFEXITED(status))
//      {
//	if (WEXITSTATUS(status)==0) Sende_Relationen=true;
//	else exit(1);
//      }
//  signal (SIGCHLD, get_sigchild);
//}

string get_filename(string servername)
{
  string name;
  struct utsname buf;
  pid_t pid;
  char pidbuf[10];

  pid = getpid();
  sprintf(pidbuf, "%i", pid);
  uname(&buf);
  name = "/tmp/qsieve-";
  name += buf.nodename;
  name += "-" + servername + "_" + pidbuf;

  return name;
}

void cleanup_files(void)
{
  string name;

  if (childpid==0) return;

  name = file_name+"A";
  unlink(name.c_str());
  name = file_name+"B";
  unlink(name.c_str());
}

void * uebertrage_Relationen(void * arg)
{
  char c;
  string s, name;
  ifstream file_stream;
  char line[80];

  cout << "Sende Daten" << endl;
  name = file_name + (char*)arg;
  file_stream.open(name.c_str(), ios::in);
  while (file_stream.peek()!=EOF)
   {
     unix_io_client communication_stream2(communication_name, server_port);
     int count=50+1; // maximimal count-1 Relationen am Block bertragen
     while (file_stream.peek()!=EOF && --count)
      {
        communication_stream2 << "Relation! ";
	while (file_stream.peek()!='\n')
	 {
	   file_stream.get(line,80,'\n');
	   communication_stream2 << line;
	 }
	file_stream.get(c);
	communication_stream2 << endl;
      }
     communication_stream2 << "Relationsblock_Ende " << flush;
     communication_stream2 >> s;
     if (s!="empfangen.")
       cerr << "Empfang der Relation nicht besttigt!" << endl;
     if (s=="ignoriert!")
      {
        cerr << "Oops! - Relationen passen dem Server nicht!" << endl;
        exit(1);
      }
   }
  Sende_Relationen = true; // neuen Thread fr bertragung anstoen...
  return NULL;
}

#endif
#endif

#ifdef IS_SERVER

void catch_sigint(int i)
{
  exit(1);
}

void close_server(void)
{
  close(server_socket);

#ifdef CLIENT_AUTOSTART
  // aufgerufen durch atexit, d.h. wir beenden uns eh gleich
  signal(SIGTERM, SIG_IGN);
  // kille alle prozesse meiner gruppe (auch mich aber ich ignoriere, s.o.)
  kill(0,SIGTERM);
#endif
}
#endif

int main(int argc, char* argv[])
{
  bool recover=false;

  cout << " --------------------------------------" << endl;
  cout << "-     Faktorisierungsprogramm V2.60     -" << endl;
  cout << "-  written 02/2000 by Thorsten Reinecke -" << endl;
  cout << "-         (reinecke@heydernet.de)       -" << endl;
  cout << " --------------------------------------" << endl;
#ifdef __GNUG__
  {
    string s=__VERSION__;
    cout << "(" << __DATE__ << ", " << __TIME__ << " using ";
    if (s[0]<='9') cout << "gcc ";
    cout << __VERSION__ << ")" << endl;
  }
#endif
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc>2)
    {
      cerr << "Zu faktorisierende Zahl als Parameter erwartet!" << endl;
      exit(0);
    }
  
  recover = (argc==1); // ohne Argument: Recovery-Modus 

  mpz_init(n); /* Variable fr zu faktorisierende Zahl bereitstellen */
  mpz_init(kN);
  mpz_init(r);
  mpz_init(x);
  mpz_init(y);

#ifndef IS_CLIENT
  if (!recover) // Nicht im Recovery-Modus -> Argumente auswerten
    { 
      // bergebene Zahl holen (und ggf. ausrechnen)
      char* neuer_str = strdup(argv[1]);
      char* str = neuer_str; 
      if (!get_Zahl(n, str)) 
	{
	  printf("Falsche Eingabe ab: '%s'\n", str);
	  exit(1);
	}
      else
	if (str[0]!='\0')
	  {    
	    printf("Klammerungsstruktur falsch!\n");
	    exit(1);
	  }
	else
	  if (mpz_cmp_ui(n,0)<=0)
	    {
	      printf("Eingegebene Zahl sollte positiv sein!\n");
	      exit(1);
	    }
      delete [] neuer_str;
    }

  if (recover)
    {
      // Recovery: Fortsetzung einer abgebrochenen Faktorisierung
      // Streams ffnen
      StaticRelations_buffer.open(StaticRelationsFile,ios::in|ios::out);
      SpecialRelations_buffer.open(SpecialRelationsFile ,ios::in|ios::out);
      DynamicRelations_buffer.open(DynamicRelationsFile,ios::in|ios::out);
      Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::ate);

      // offenbar wird trotz ios::ate nicht (immer) ans Ende positioniert
      // deshalb wird die Testabfrage modifiziert:
      if (Recovery_from_file) Recovery_from_file.seekg(0,ios::end);
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()<1) )
	{
	  cerr << "Recovery nicht mglich!" << endl;
	  exit(1);
	}
      
      Recovery_from_file.seekg(0,ios::beg);
      Recovery_to_file.seekp(0,ios::beg);
      Recovery_from_file >> n; // Zahl holen
      Recovery_from_file.ignore(1,'\n');

      cout << "Setzte Faktorisierung fort fr" << endl;
      cout << n << endl;
      // auerdem Recovery-Modus im Factorization-File vermerken, denn es
      // 1. knnte gelscht sein
      // 2. knnte der Recovery-Modus in seltenen Fllen Faktoren unterdrcken
      //    (wenn vorher das Schreiben unterbrochen wurde)
      // 3. knnte der Recovery-Modus zu doppelten Faktor-Ausgaben fhren
      //    (falls das quadratische Sieb einen zusammengesezten Faktor
      //    gefunden hat und dann noch unterbrochen wird)
      // 4. knnte das Multiplikationssymbol inkorrekt hufig auftreten
      //    (wenn vorher kein Faktor gefunden wurde)
      Factorization_to_file << " [RECOVERY] "; // ohne flush, falls vor Finden weiterer Faktoren erneut abgebrochen wird... 
    }
  else
    {
      // Neustart: Starten einer neuen Faktorisierung
      // Streams ffnen
      StaticRelations_buffer.open(StaticRelationsFile,ios::out|ios::trunc);
      SpecialRelations_buffer.open(SpecialRelationsFile ,ios::in|ios::out|ios::trunc);
      DynamicRelations_buffer.open(DynamicRelationsFile,ios::in|ios::out|ios::trunc);
      Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::trunc);
      
      cout << "Starte Faktorisierung fr" << endl;
      cout << n << endl;
      
      Factorization_to_file << endl << "Factorization of " << argv[1] << ":" << endl;
      Factorization_to_file << n << " = " << endl;
      
      easy_factor(); // "einfache" Faktoren rausstreichen
      
      cout << "Starte Faktorisierung mit ECM und Siebverfahren fr" << endl;
      cout << n << endl;
      Recovery_to_file << n << endl;           
    }

#ifdef IS_SERVER
  {
    int i=0;
    while ((server_socket = open_internet_port(server_port,10))<0
	   && ++i < server_port_tries)
      sleep(1);  
  }
  atexit(close_server);
  signal(SIGINT, catch_sigint);
  if (server_socket<0) {
    cerr << "Kann Port " << server_port << " nicht oeffnen" << endl;
    exit(1);
  }
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning
  cout << "Waiting for net-clients (ECM)..." << endl;
#ifdef CLIENT_AUTOSTART
  {
    string s="autostart_clients.sh ";
    struct utsname buf;
    setpgid(0,0);
    uname(&buf);
    s += buf.nodename;
    cout << "Calling script to start clients..." << flush;
    cout << "(exitstatus=" << system(s.c_str()) << ")" <<endl;
  }
#endif
  bearbeite_daten_stream_ecm(server_socket); // erstmal ECM verteilt starten...
#endif

#ifdef IS_STANDALONE
  // Auch die Standalone-Version sollte elliptische Kurven benutzen.
  bearbeite_ecm(); // erstmal ECM verteilt starten...
#endif

  // nun wird alles fr das Quadratische Sieb vorbereitet...

  bestes_kN_ermitteln(); /* optimaler Wert, fr den gesiebt werden kann */

  /* Streitfall:
     Soll man nach Ermittlung des bestem Multipliers die Siebparameter
     noch tunen?
     Ja   -> denn die Zahl kN hat ja die entsprechende Gre...
     Nein -> denn fr die Wahl des Multipliers wurden diese Gren nicht
             bercksichtigt und entscheidend sollte die Gre der
             zu faktorisierenden Zahl sein. Nur so ist eine Vergleichbarkeit
             der Multiplier unter gleichen Voraussetzungen zu gewhrleisten.
    Wie auch immer:
     Eine "optimale" Ermittlung der Siebgre und der Faktorbasis kann nur
     mittelbar von der Gre der Zahl abhngen. Die Laufzeit der Siebphase
     hngt davon ab, wie viele Relationen zu ermitteln sind *und* wie
     schnell die Relationen ermittelt (=gesiebt) werden knnen!
     Letzterer Wert ist funktional von der Gre der Zahl und der Qualitt
     der Faktorbasis abhngig. Diese Qualitt ist zwar durch Vorgabe eines
     Multipliers (nach Annahme) maximiert worden; aber das heit noch nicht,
     da dadurch die funktionale Abhngigkeit beseitigt wre.
     Mit anderen Worten: Wenn wir die Bestimmung der Faktorbasisgre und
     der Siebgre *nur* von der Gre der Zahl abhngig machen, so gehen
     wir idealisierend von einer einheitlichen Qualitt der Faktorbasis aus.
     Da wir diese einheitliche Qualitt der Faktorbasis nicht definiert haben,
     besitzen wir auch keine Entscheidungsgrundlage darber, ob fr unsere
     Zahl (n bzw. kN) eine unterdurchschnittliche, durchschnittliche oder
     berdurchschnittliche Faktorbasisqualitt besteht. Aus dieser Sichtweise
     erscheint der obige Streitfall nicht entscheidbar zu sein.
  */

#if 0   
  tune_parameters(mpz_sizeinbase(kN,10)); // Parametertuning fr kN
#else
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning fr n
#endif
  /* Hinweis:
      Beim Tuning der Parameter ist in der verteilten Version auf Konsistenz
      zu achten, d.h. ein Client sollte keine Parameter tunen!
      Unterschiedliche cfg-Files auf verschiedenen Rechnern wrden sonst
      zu Problemen fhren!
  */

#else
  if (recover)
    {
      cerr << "Fehlender Parameter" << endl;
      exit(1);
    }
  communication_name = argv[1];
#ifdef USE_NETWORK
  // existiert der rechner?
  struct hostent *hp;
  file_name = get_filename(communication_name);
  hp = gethostbyname(communication_name);
  if (hp == NULL)
    {
      cerr << "Unbekannter Rechner " << communication_name << endl;
      exit (1);
    }

  { // zunchst elliptische Kurven abarbeiten (net-client):
    int ecm_sigma = -1, runde=0;
    while (true)
     {
       unix_io_client communication_stream(communication_name, server_port);
       communication_stream << "ECM?" << endl;
       communication_stream >> ecm_sigma;
       if (ecm_sigma<0) break; // sigma<0 -> keine Kurven mehr!
       communication_stream >> elcu_Phase1 >> elcu_Phase2;
       communication_stream >> n; // Zahl holen
       cout << "Elliptische Kurve Nr. " << ++runde << endl;
       elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
       elliptic_curve.go(ecm_sigma,elcu_Phase1,elcu_Phase2); /* Faktorisierung mit elliptischer Kurve versuchen */
     }
  }
  
  {
    unix_io_client communication_stream(communication_name, server_port);
    communication_stream << "Hunger!" << endl;
    communication_stream >> n;
    communication_stream >> Primbasis_size;
    communication_stream >> Factor_Threshold;
    communication_stream >> M;
    cout << "n=" << n << endl;
    cout << "Gre der statischen Faktorbasis: " << Primbasis_size << endl;
    cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
    cout << "Siebintervall pro Polynom: [" << -M << "," << M << "]" << endl;
  }
  {
    string name;
    name = file_name+"A";
    communication_stream.open(name.c_str(), ios::out|ios::trunc);
  }
  //signal (SIGCHLD, get_sigchild);
  atexit(cleanup_files);
#else
  // existiert das file?
  ifstream communication_stream2(communication_name);
  if (!communication_stream2)
    {
      cerr << "Kann File " << communication_name << " nicht oeffnen" << endl;
      exit (1);
    }
  char *tmp = communication_name;
  communication_name = new char [strlen(tmp)+5];
  strcpy(communication_name, tmp);
  strcat(communication_name, ".out");
  
  communication_stream2 >> n; 
  communication_stream2 >> Primbasis_size;
  communication_stream2 >> Factor_Threshold;
  communication_stream2 >> M;
  communication_stream2.ignore(1,'\n');
  cout << "n=" << n << endl;
  cout << "Gre der statischen Faktorbasis: " << Primbasis_size << endl;
  cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
  cout << "Siebintervall pro Polynom: [" << -M << "," << M << "]" << endl;
  communication_stream.open(communication_name, ios::out|ios::trunc);
#endif
  bestes_kN_ermitteln(); /* optimaler Wert, fr den gesiebt werden kann */
#endif  
  
  if (mpz_cmp_ui(r,Siebgroesse)<0)
    {
      cerr << "Sieb ist berdimensioniert (ggf. verkleinern)!" << endl;
      exit(1);
    }

  // Fr Double-Large-Primes einen Schwellwert setzen,
  // dieser ist das Quadrat des maximalen Specialfactors...
  mpz_init(DLP_Threshold);
  mpz_set_ui(DLP_Threshold,Specialfactor_Sieb_Threshold);
  mpz_mul(DLP_Threshold,DLP_Threshold,DLP_Threshold);
  
  Statische_Faktorbasis_ermitteln();
  cout << "Die ersten 20 Elemente der statischen FB lauten" << endl;
  for (int i=0; i<20; i++) cout << Primzahlen[i] << " "; cout << endl;
  cout << "Grte Primzahl in statischer Faktorbasis: " << biggest_Prime_in_Factorbase << endl;
  cout << "Anzahl Primzahlquadrate in FB: " << FB_maxQuadrate << endl;
  cout << "Anzahl Primzahlpotenzen in FB: " << anz_weitere_Primzahlpotenzen << endl;

  SiebSchwellwert_kalkulieren(); // Schwellwert zur Erkennung der Relationen fr Siebvorgang bestimmen
  Sieb=new TSiebElement [Siebgroesse]; /* Speicher fr das Sieb allokieren */

  Polynom.erstes_Polynom_ermitteln(kN,M); // erstes Polynom ermitteln
  
#ifndef IS_CLIENT
  if (recover)
    {
      // Static Relations einlesen
      cout << "Lese Static Relations ein..." << endl;
      StaticRelations_from_file.seekg(0,ios::beg);
      while (StaticRelations_from_file.peek()=='G') 
	{
	  CGleichungszeile* GL = new CGleichungszeile();
	  GL->kombinieren(StaticRelations_from_file);
          GL->optisize(); // Relation optimieren (dense<->sparse, etc.)
          GLS[GL->relevanter_Faktor]=GL; Fuellung_GLS++; // Relation ins Gleichungssystem einfgen
	  StaticRelations_from_file.ignore(1,'\n');
	}
      StaticRelations_from_file.clear();
      cout << "Recover: " << Fuellung_GLS << " Static Relations eingelesen" << endl;

      // Dynamic Relations einlesen
      cout << "Lese Dynamic Relations ein..." << endl;
      DynamicRelations_from_file.seekg(0,ios::beg);
      while (DynamicRelations_from_file.peek()=='G') 
	{
	  TDynamicFactorRelation relation;
	  relation.fpos = DynamicRelations_from_file.tellg();
	  string s; DynamicRelations_from_file >> s;
	  DynamicRelations_from_file >> relation.factor;
	  relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
#ifndef LAZY_EVALUATION
          // bei busy-evaluation Faktoren ggf. cachen,
	  if (relation.factor<Dynamicfactor_Cache_Threshold)
	    {
              CGleichungszeile *GL; // Zeiger auf GL
	      GL = new CGleichungszeile();
	      GL->kombinieren(DynamicRelations_from_file,relation.fpos); // GL einlesen
              if (!relation.insert_cache(GL)) delete GL; // nicht gecached
	    }
#endif // LAZY_EVALUATION
	  DynamicRelations_from_file.ignore(1000000,'\n');
	  Dynamic_Factor_Relations.insert(relation);
	}
      DynamicRelations_from_file.clear();
      DynamicRelations_to_file.seekp(0, ios::end); // sicherheitshalber...
      cout << "Recover: " << Dynamic_Factor_Relations.size()
	   << " Dynamic Relations eingelesen" << endl;
      {
        // Special Relations einlesen
        // zustzlich Speicherkompaktifizierung der special relations
        // auf dem File durchfhren, d.h. ungltige (=als redundant
        // markierte) Relationen streichen.
        cout << "Lese Special Relations ein und kompaktifiziere File..." << endl;
        unsigned int u_hits = 0;
        streampos rpos=0, wpos=0;
        // Special Relations einlesen
        SpecialRelations_from_file.seekg(0,ios::beg);
        SpecialRelations_to_file.seekp(0,ios::beg);
        while (true) 
	  {
            SpecialRelations_from_file.seekg(rpos,ios::beg);
	    while (SpecialRelations_from_file.peek()=='U')
              {
	        SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
                u_hits++;
              }
	    if (SpecialRelations_from_file.peek()!='G') break;
	    TSpecialFactorRelation relation;
	    relation.fpos = wpos;
	    string s; SpecialRelations_from_file >> s;
	    SpecialRelations_from_file >> relation.factor;
	    Special_Factor_Relations.insert(relation);
	    relation.factor.swap();
	    Special_Factor_Relations.insert(relation);
            relation.factor.swap();
            if (u_hits>0)
            {
              char buffer[2048];
              rpos=SpecialRelations_from_file.tellg();
              SpecialRelations_to_file.seekp(wpos);
              SpecialRelations_to_file << "G " << relation.factor;
              wpos=SpecialRelations_to_file.tellp();
              SpecialRelations_from_file.seekg(rpos);
              while (SpecialRelations_from_file.peek() != '$')
                {
                  SpecialRelations_from_file.get(buffer, sizeof(buffer), '$');
                  rpos=SpecialRelations_from_file.tellg();
                  SpecialRelations_to_file.seekp(wpos);
                  SpecialRelations_to_file << buffer;
                  wpos=SpecialRelations_to_file.tellp();
		  SpecialRelations_from_file.seekg(rpos);
                }
              rpos=SpecialRelations_from_file.tellg();
              SpecialRelations_to_file.seekp(wpos);
              SpecialRelations_to_file << "$" << endl;
              wpos=SpecialRelations_to_file.tellp();
              SpecialRelations_from_file.seekg(rpos);
              SpecialRelations_from_file.ignore(10,'\n');
              rpos=SpecialRelations_from_file.tellg();
            }
            else
            {
              SpecialRelations_from_file.ignore(1000000,'\n');
              //SpecialRelations_to_file.seekp(SpecialRelations_from_file.tellg());
              wpos=rpos=SpecialRelations_from_file.tellg(); //Schreibposition=Leseposition
            }
	  }

        // mangels truncate-Funktion den Rest mit U's auffllen
        {
          streampos pos, ende;
          pos = wpos;
          SpecialRelations_to_file.seekp(0, ios::end);
          ende = SpecialRelations_to_file.tellp();
	  SpecialRelations_to_file.seekp(wpos);
          while (SpecialRelations_to_file.tellp() != ende)
            {
              SpecialRelations_to_file << "U";
            }
          SpecialRelations_to_file << flush;
          SpecialRelations_to_file.seekp(pos);
        }
        SpecialRelations_from_file.clear();
        cout << "Recover: " << Special_Factor_Relations.size()
	     << " Special Relations eingelesen" << endl
             << u_hits << " ungltige DLP-Relations eliminiert" << endl;
      }
      SpecialFactors_Zyklensuche();
      
      
#ifdef IS_SERVER
      // beim verteilten Sieben knnten den Clients einige potentielle dynamische Faktoren
      // durch die Lappen gegangen sein. Dies geschieht, falls ein Client einen Special-Factor
      // findet, der nicht mit dem vollen Satz der aktuellen (und daher nur dem Server bekannten)
      // dynamischen Faktoren gesiebt wurde.
      // Anstelle eines Tests zur Laufzeit werden diese im Normalfall eher seltenen Flle
      // nun hier beim Neu-Hochfahren des Servers bercksichtigt:
      // WICHTIG: erst mssen alle dynamischen Faktoren eingelesen sein! Ansonsten knnen
      //          doppelt auftretende dynamische Faktoren nicht zu einer Relation reduziert werden!
      // folgende Anweisungen wrden das bewerkstelligen, aber es lohnt sich kaum und dauert (zu) lange:
#if 1 /* 0: wir lassen es, 1: wir machen es (trotzdem) */
      cout << "Teste Special-Factoren auf unentdeckt gebliebene Dynamische Faktoren." << endl;
      for (TDynamic_Factor_Relations::iterator pos = Dynamic_Factor_Relations.begin();
	   pos!=Dynamic_Factor_Relations.end(); ++pos)
	Special_Factoren_durch_Primfaktor_splitten((*pos).factor);
      cout << "entdeckte dynamische Faktoren: " << Special_to_dynamic_Factor_hit << endl;
#endif
#endif
      
    }
#else // is client:
#ifndef USE_NETWORK 
  TDynamicFactorRelation relation;
  relation.fpos = 0; relation.factor=0;
  int diff; // wir lesen die Abstnde zwischen den Faktoren ein
  communication_stream2 >> diff; // filepos ueberlesen
  while (true)
    {
      communication_stream2 >> diff;
      relation.factor+=diff;
      // cout << "*" << relation.factor << "*" << flush;
      if (diff<=0) break;
      relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
      Dynamic_Factor_Relations.insert(relation);
    }
  //cout << endl;
#endif
#endif  
  
  
  
#ifndef IS_CLIENT
  if (recover)
    {
      streampos fpos = Recovery_from_file.tellg();
      Polynom.Daten_laden(Recovery_from_file); // aktuelles Polynom (zum Fortsetzen) ermitteln
      Recovery_to_file.seekp(fpos);
    }
#endif

Statuslegende_ausgeben();

#ifdef IS_SERVER
  cout << "Waiting for clients..." << endl;

  while (true)
    {    
      
      int communication_socket = accept(server_socket, NULL, NULL);
      if (communication_socket > 0)
	{
	  unix_buffer buffer(communication_socket);
	  iostream communication_stream(&buffer);
	  
	  bearbeite_daten_stream(communication_stream);
	  communication_stream << flush;
	  close (communication_socket);
	}
      
      // collecting_phase_beendet= ( Fuellung_GLS > 0.6*Primbasis_size );
      /* Durch das Flag "collecting_phase_beendet" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, da dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
      
    }

#endif // Server


#ifdef IS_CLIENT
  mpz_t Obergrenze_D;
  mpz_init(Obergrenze_D);
    
#ifdef USE_NETWORK

 ClientMainLoop: // Hauptschleife: Polynomintervall holen und diese durchsieben
  
  {
    unix_io_client communication_stream2(communication_name, server_port);
    cout << "Hole dynamische Faktoren..." << endl;
    static streampos dynfac_fpos = 0;
    communication_stream2 << "DynamicFactors?_ab_fpos " << dynfac_fpos << endl;
    communication_stream2 >> dynfac_fpos; // nchste Startposition holen
    TDynamicFactorRelation relation;
    relation.fpos=0; relation.factor=0;
    int diff; // wir empfangen die Abstnde zwischen den Faktoren
    while (true) // dynamische Relationen holen
      {
	communication_stream2 >> diff;
        relation.factor+=diff;
	// cout << relation.factor << "\r" << flush;
	if (diff<=0) break;
	relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	Dynamic_Factor_Relations.insert(relation);
      }
    cout << "geholt..." << endl;
  }
  {
    unix_io_client communication_stream2(communication_name, server_port);
    cout << "Hole Polynomintervall..." << endl;
    communication_stream2 << "Polynom? 1000" << endl;
    
    Polynom.Daten_laden(communication_stream2);
    communication_stream2 >> Obergrenze_D;
  }
#else // kein Netzwerk
  Polynom.Daten_laden(communication_stream2);
  communication_stream2 >> Obergrenze_D;
#endif // if Netzwerk else keinNetzwerk
#endif // Client
  
  /* Quadratischen Reste mit Primzahlen aus der Primzahlbasis aussieben */
  do /* solange Relationen aussieben, bis mehr Relationen gefunden sind
	als Primzahlen in der Primzahlbasis sind (oder eine Faktorisierung
	gefunden ist). */
    {
#ifndef IS_CLIENT
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.Daten_sichern(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
#endif
      
      Sieboffset=-M;
      alle_Deltas_neu_berechnen();
      while (Sieboffset<M)
	{
	  Sieb_initialisieren();
	  schnelles_aussieben(); // mit Primzahlen der statischen Faktorbasis
	  aussieben_mit_Quadraten(); // (grere) Quadrate von statischen Faktoren sieben
	  aussieben_mit_Dynamic_Factoren(); // ermglicht weitere Relationen...
	  schnelles_auslesen_des_Siebs();	  
	  Sieboffset+=Siebgroesse;
	}
      Statusmeldung(); // Ausgabe von Informationen

#ifdef IS_CLIENT
#ifdef USE_NETWORK

      pthread_t thread_Relationen_versenden;
 	
      if (Sende_Relationen)
	{
	  // Relationen sollen bertragen werden:
          // a) -> Flag setzen, da bertragung angestoen wurde
          //       (d.h. Sende_Relationen auf false setzen)
          // b) Thread zum bertragen erzeugen
          // c) dieser Thread setzt das Flag bei Beendigung
          
	  string name;
	  Sende_Relationen = false;
	  
	  // close file
	  communication_stream.close();

          int retcode;
	  void * retval;

          static bool first_time_first_love = true;

          if (first_time_first_love)
            { // vor dem ersten Thread nicht joinen, ...
              first_time_first_love=false;
              retcode = pthread_create(&thread_Relationen_versenden, NULL,
                           uebertrage_Relationen, (void *)(Sende_Nr?"B":"A"));
            }
	  else
	    { // ... danach schon...
              retcode = pthread_join(thread_Relationen_versenden, &retval);
              if (retcode != 0)
               {
                 cerr << "Join fehlgeschlagen!" << endl;
                 exit(1);
               }
              retcode = pthread_create(&thread_Relationen_versenden, NULL,
                           uebertrage_Relationen, (void *)(Sende_Nr?"B":"A"));
            }

          if (retcode != 0)
           {
             cerr << "Thread zum bertragen der Relationen fehlgeschlagen!" << endl;
             exit(1);
           }

	  Sende_Nr = 1 - Sende_Nr;
	  // open new file
	  name = file_name + (Sende_Nr?"B":"A");
	  communication_stream.open(name.c_str() ,ios::out|ios::trunc);
	}
#endif  // Network
#endif  // Client

      Polynom.naechstes_Polynom_ermitteln();
      
      // collecting_phase_beendet= ( Fuellung_GLS > 0.6*Primbasis_size );
      /* Durch das Flag "collecting_phase_beendet" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, da dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
    }
#ifndef IS_CLIENT
  while (true);
#else
  while (Polynom < Obergrenze_D);
#ifdef USE_NETWORK
  goto ClientMainLoop; // Hauptschleife: neues Polynomintervall holen und fortsetzen
#endif
  mpz_clear (Obergrenze_D);
#endif
  
  delete [] Sieb; /* Sieb lschen, d.h. Speicher wieder freigeben */ 
  
  cout << endl << "Fertig..." << endl;
  return 0;
  
}
